/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstream>

#include <networkmodels/rackswitch.h>

#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>
#include <sstmac/hardware/network/switches/circuit/circuitswitch.h>

#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/switchtopology.h>

#include <sstmac/common/factories/topology_factory.h>

namespace sstmac
{
  namespace omnetpp
  {

    using namespace sstmac;
    using namespace sstmac::hw;

    debug rackswitch::dbg_;
    topology::ptr rackswitch::topology_;

    /*
     map<cChannel*, rackswitch*> rackswitch::channel_map1_;
     map<cChannel*, rackswitch*> rackswitch::channel_map2_;

     map<cChannel*, int> rackswitch::channel_vals1_;
     map<cChannel*, int> rackswitch::channel_vals2_;
     */

    Define_Module( sstmac::omnetpp::rackswitch);

    rackswitch::~rackswitch()
    {

    }

    void
    rackswitch::initialize()
    {

      params_ = omnetpp_params::construct(this);

      if (params_->get_param("network_name") != "switch")
        {
          throw ssterror(
              "rackswitch::initialize -- network name isn't -switch-, make sure your network is set up right");
        }

      if (!topology_)
        {
          topology_ = SSTTopologyFactory::get_topology(
              params_->get_param("topology_name"), params_);
        }

      switchtopology::ptr switchtop = boost::dynamic_pointer_cast<
          switchtopology>(topology_);
      if (!switchtop)
        {
          sst_throw(ssterror, "switchinterconnect::ctor: topolgy %s doesn not support switch model",
                    params_->get_param("topology_name").c_str());
        }

      networkswitch::ptr switchtemp;
      std::string switchtype = params_->get_param("network_switchtype");
      int nps = params_->get_int_param("network_nodes_per_switch");

      for (int i = 0; i < nps; i++)
        {
          gates_[gate("ports$i", i)] = -(i+1);
        }

      int radix = par("numRackConnections");
      radix -= nps;
      for (int i = 0; i < radix; i++)
        {
          gates_[gate("ports$i", i + nps)] = i;
        }

      SSTMAC_DEBUG << "switchinterconnect: using switch model: " << switchtype
            << "\n";
      if (switchtype == "packet")
        {
          switchtemp = packetswitch::construct();

        }
      else if (switchtype == "cycleaccurate")
        {
          switchtemp = cycleaccurateswitch::construct();
        }
      else if (switchtype == "circuit")
        {
          switchtemp = circuitswitch::construct();
        }

      int rid = par("rackid");
      nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(rid);
      switch_
          = switchtemp->clone(switchtop->create_router(addr, params_), addr);

      switch_->set_eventmanager(fakeeventmanager::construct(this));


    }

    void
    rackswitch::handleMessage(cMessage* msg)
    {
      if (msg->isSelfMessage())
        {
          EventWrapperMsg* emsg = check_and_cast<EventWrapperMsg*> (msg);
          emsg->getEvent().event->execute();
        }
      else if (gates_[msg->getArrivalGate()] >= 0)
        {
          SSTMAC_DEBUG << "switchinterconnect: message arrived from switch: "
                << gates_[msg->getArrivalGate()] << "\n";
          SSTMessageMsg* m = check_and_cast<SSTMessageMsg*> (msg);
          switch_->handle(m->getMsgcontent().load);
        }
      else if (gates_[msg->getArrivalGate()] < 0)
        {
          SSTMAC_DEBUG << "switchinterconnect: message arrived from node: "
                        << gates_[msg->getArrivalGate()] << "\n";

          SSTMessageMsg* m = check_and_cast<SSTMessageMsg*> (msg);
          switch_->get_injector(params_, nodesPerRack)->handle(
              m->getMsgcontent().load);
        }
    }

    void
    rackswitch::finish()
    {

      /*type<ptr_type::ptr>::list::iterator it, end = finishes_.end();
      for (it = finishes_.begin(); it != end; it++)
        {
          eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(*it);
          mod->finish();
        }

      fem_->finish_stats();
*/
    }

  /*
   void
   rackswitch::register_channel(cChannel* ch, rackswitch* rs)
   {
   if (displayConnections)
   {
   if (channel_map1_.find(ch) != channel_map1_.end())
   {
   channel_map2_[ch] = rs;
   channel_vals2_[ch] = 0;
   }
   else
   {
   channel_map1_[ch] = rs;
   channel_vals1_[ch] = 0;
   }
   }
   }

   void
   rackswitch::update_channel_display(cChannel* ch, rackswitch* rs, double frac)
   {

   if (displayConnections)
   {
   cDisplayString& connDispStr = ch->getDisplayString();

   if (channel_map1_[ch] == rs)
   {

   std::stringstream ss;
   int linewidth = (1 + frac * 8);

   ss << "ls=#";

   int val = 255 * (frac > 1 ? 1.0 : frac);
   channel_vals1_[ch] = val; //save it so the other guy can use it

   ss << hex << std::setw(2) << channel_vals1_[ch]; //red

   ss << "00"; //green

   ss << hex << std::setw(2) << channel_vals2_[ch]; //blue

   ss << "," << dec << linewidth;

   std::string dis = ss.str();

   replace(dis.begin(), dis.end(), ' ', '0');


   //cout << "connection display: " << dis;
   connDispStr.parse(dis.c_str());
   }
   else
   {
   std::stringstream ss;
   int linewidth = (1 + frac * 8);

   ss << "ls=#";
   int val = 255 * (frac > 1 ? 1.0 : frac);
   channel_vals2_[ch] = val; //save it so the other guy can use it

   ss << hex << std::setw(2) << channel_vals1_[ch];

   ss << "00"; //green

   ss << hex << std::setw(2) << channel_vals2_[ch];

   ss << "," << dec << linewidth;

   std::string dis = ss.str();

   replace(dis.begin(), dis.end(), ' ', '0');

   //cout << "connection display: " << dis;
   connDispStr.parse(dis.c_str());
   }
   }
   }*/
  }
}
