/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "abstractnodeprocessor.h"

#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/node/degradednode.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/software/services/launch/instantlaunch.h>

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <sstmac/backends/omnetpp/omnetpp_params.h>

#include <sstmac/hardware/network/topology/hdtorus.h>

#include <sstmac/common/factories/nic_factory.h>
#include <sstmac/common/factories/topology_factory.h>

namespace sstmac
{
  namespace omnetpp
  {

    Define_Module( sstmac::omnetpp::abstractnodeprocessor);

    using namespace sstmac::hw;

    debug abstractnodeprocessor::dbg_;

    abstractnodeprocessor::abstractnodeprocessor()
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> omnetnode");
          logger::register_logger(&dbg_);
        }

    }

    abstractnodeprocessor::~abstractnodeprocessor()
    {

    }

    void
    abstractnodeprocessor::initialize()
    {

      int c = 0;

      int p = 0;

      int n = par("nodeid");

      int r = par("rackid");

      myid = omnetpp_address::construct(c, p, n, r);

      omnetpp_params::ptr params = omnetpp_params::construct(this);

      coresPerProc = par("node_cores");
      numOfNodes = par("node_num");
      procsPerNode = par("node_procs");
      nodesPerRack = par("network_nodes_per_switch");
      numRacks = par("rack_num");

      omnetpp_address::set_system_params(coresPerProc, procsPerNode,
          nodesPerRack, numRacks);

      outgate = gate("toNetwork$o");
      ingate = gate("toNetwork$i");

      fem_ = fakeeventmanager::construct(this);

      std::string temp = par("logger_params").stringValue();
      bool color = par("debug_coloring");
      logger::set_debug_coloring(color);
      logger::set_user_param(temp);
      logger::timer_ = fem_;

      // ---- get parameters and such
      std::string app_p_fname = par("app_paramfile").stringValue();
      sim_parameters::ptr appparams = file_parameters::construct(app_p_fname);
      params->add_param_set(appparams);

      sstmac_env::params = params;

      std::string man = par("manager").stringValue();

      if (man == "simple")
        {
          node_ = sstmac::hw::simplenode::construct(params, myid);
        }
      else if (man == "degraded")
        {
          bool found = false;
          std::string deg = par("manager_degraded_nodes").stringValue();

          if (deg == "evens")
            {
              if (myid->unique_id() % 2 == 0)
                {
                  found = true;
                  node_ = sstmac::hw::degradednode::construct(params, myid);
                }
            }
          else if (deg == "odds")
            {
              if (myid->unique_id() % 2 == 1)
                {
                  found = true;
                  node_ = sstmac::hw::degradednode::construct(params, myid);
                }
            }
          else
            {
              bool error = false;
              std::vector<int> degraded = get_intvec(deg.c_str(), error);

              for (size_t i = 0; i < degraded.size(); i++)
                {
                  if (degraded[i] == myid->unique_id())
                    {
                      found = true;
                      node_ = sstmac::hw::degradednode::construct(params, myid);
                      break;
                    }
                }

            }

          if (!found)
            {
              node_ = sstmac::hw::simplenode::construct(params, myid);
            }
        }
      else
        {
          opp_error("unknown manager name");
        }

      interconnect::ptr inter = fakeinterconnect::construct(this, SSTTopologyFactory::get_topology(params->get_param("topology_name"), params));
      networkinterface::ptr nic = SSTNicFactory::get_nic(
          params->get_param("nic_name"), params, node_, inter);
      node_->set_nic(nic);
      nic->set_eventmanager(fem_);

      node_->initialize(fem_);

    }

    //sstmac::hw::topology::ptr
    //abstractnodeprocessor::fakeinterconnect::get_topology()
   // {
    //  throw ssterror(
     //     "abstractnodeprocessor::fakeinterconnect::get_topology - shouldn't be calling this");
   // }

    int64_t
    abstractnodeprocessor::fakeinterconnect::slots() const
    {
      return parent_->numOfNodes;
    }

    void
    abstractnodeprocessor::fakeinterconnect::send(const timestamp &start_time,
        const sst_message::ptr &msg)
    {
      SSTMessageMsg* m = new SSTMessageMsg();
      m->setMsgcontent(msg);

      parent_->sendDelayed(m, start_time.sec(), parent_->outgate);
    }

    void
    abstractnodeprocessor::finish()
    {
      std::list::iterator<ptr_type::ptr>:: it, end = finishes_.end();
      for (it = finishes_.begin(); it != end; it++)
        {
          eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(*it);
          mod->finish();
        }

      fem_->finish_stats();
    }

    void
    abstractnodeprocessor::handleMessage(cMessage* cmsg)
    {

      if (cmsg->isSelfMessage())
        {

          EventWrapperMsg* emsg = check_and_cast<EventWrapperMsg*> (cmsg);
          eventpod epod = emsg->getEvent();
          epod.event->execute();
          delete emsg;

        }
      else if (cmsg->getArrivalGate() == ingate)
        { //message arriving in from the NoC
          SSTMessageMsg *m = check_and_cast<SSTMessageMsg*> (cmsg);

          sst_message::ptr msg = m->getMsgcontent().load;
          node_->handle(msg);

          delete m;
        }
      else
        {
          opp_error("abstractnodeprocessor: unexpected message");
        }
    }

    timestamp
    abstractnodeprocessor::now()
    {
      timestamp t(simTime().dbl());
      return t;
    }

  // ------------------------------------ //

  }
}
