/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_DRIVER_UTIL_H
#define SSTMAC_DRIVER_UTIL_H

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

#include <sstmac/software/services/launch/indexing/indexstrategy.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/network/topology/topology.h>

#include <sstmac/common/sim_parameters.h>

#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/units.h>
#include <ostream>
#include <boost/foreach.hpp>

#include <sstmac/common/vis/vis.h>

namespace sstmac
{

  namespace sw
  {
    class launcher;
    class launchinfo;
  }


  class command_parameters : public sstmac::sim_parameters
  {
  protected:
    maptype params_;

  protected:
    command_parameters();

    command_parameters(const maptype& p);

  public:
    typedef boost::intrusive_ptr<command_parameters> ptr;

    static ptr
    construct();

    virtual sim_parameters::ptr
    clone();

    sim_parameters::ptr
    subspace_clone()
    {
        return new command_parameters;
    }

    virtual std::string
    get_param(const std::string& key) const;

    virtual bool
    has_param(const std::string& key) const;

    virtual void
    add_param(const std::string& key, const std::string& val);

    virtual void
    add_param_override(const std::string& key, const std::string& val);

    virtual void
    combine_into(sim_parameters::ptr sp);

    virtual std::string
    to_string() const;

    virtual
    ~command_parameters();

    virtual void print_params(std::ostream& os = std::cerr) const {
        sim_parameters::print_params(params_, os);
    }

  };

  class file_parameters : public sstmac::sim_parameters
  {
  protected:
    maptype params_;

  protected:
    file_parameters();

    file_parameters(const maptype& p);

  public:
    typedef boost::intrusive_ptr<file_parameters> ptr;

    void
    parse_file(const std::string& fname);

    static ptr
    construct(const std::string& filename);

    sim_parameters::ptr
    subspace_clone()
    {
        return new file_parameters;
    }

    virtual sim_parameters::ptr
    clone();

    void
    append_params(const std::string& filename);

    virtual std::string
    get_param(const std::string& key) const;

    virtual bool
    has_param(const std::string& key) const;

    virtual void
    add_param(const std::string& key, const std::string& val);

    virtual void
    add_param_override(const std::string& key, const std::string& val);

    virtual void
    combine_into(sim_parameters::ptr sp);

    virtual std::string
    to_string() const;

    virtual
    ~file_parameters();

    virtual void print_params(std::ostream& os = std::cerr) const {
        sim_parameters::print_params(params_, os);
    }

  };

  sstmac::vis::vis_engine::ptr
  construct_vis_engine(sstmac::sim_parameters::ptr params);

  double
  get_positive_float(const char *value, bool &errorflag);

  int
  get_int(const char *value, bool &errorflag);

  int
  get_positive_int(const char *value, bool &errorflag);

  void
  get_intvec(const char *value, bool &errorflag, std::vector<int>& retval);

  std::vector<int>
  get_intvec(const char *value, bool &errorflag);

}
#endif
