/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMACRO_COMMON_FACTORIES_FACTORY_H_INCLUDED
#define SSTMACRO_COMMON_FACTORIES_FACTORY_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  class factory_type :
    virtual public ptr_type
  {
    protected:
      std::string factory_name_;

    public:
      typedef boost::intrusive_ptr<factory_type> ptr;

      void set_factory_name(const std::string& name){
          factory_name_ = name;
      }

      std::string get_factory_name() const {
          return factory_name_;
      }
  };


  class SSTDesc_base
  {
  public:

    virtual factory_type::ptr
    build(const sim_parameters::ptr& params) = 0;

    virtual void clear();

  };

  class SSTFactory_base
  {
  public:
    static void
    add_to_map(const std::string& namestr, SSTDesc_base* desc,
        std::map<std::string, SSTDesc_base*> *m);

  };

  template <class T>
  class SSTFactoryTemplateBase
    : public SSTFactory_base
  {
  protected:
    typedef std::map<std::string, SSTDesc_base*> descr_map;
    static descr_map* object_map_;

  public:
    static void
    register_name(const std::string& name, SSTDesc_base* descr){
        if (!object_map_)
            object_map_ = new descr_map;
        add_to_map(name, descr, object_map_);
    }
  };

  template <class T>
  class SSTFactory :
    public SSTFactoryTemplateBase<T>
  {

  public:
      typedef boost::intrusive_ptr<T> ptr;
      using SSTFactoryTemplateBase<T>::object_map_;

      static ptr
      get_value(const std::string& valname, const sim_parameters::ptr& params){
          SSTDesc_base* descr = (*object_map_)[valname];
          if (!descr){
              sst_throw(valueerror, "could not find factory name %s", valname.c_str());
          }
          factory_type::ptr p = descr->build(params);
          p->set_factory_name(valname);
          ptr cast = boost::dynamic_pointer_cast<T>(p);
          return cast;
      }

    static ptr
    get_param(const std::string& param_name, const sim_parameters::ptr& params){
        std::string valname = params->get_param(param_name);
        return get_value(valname, params);
    }

    static ptr
    get_optional_param(const std::string& param_name, const std::string& defval,
                 const sim_parameters::ptr& params)
    {
        std::string valname = params->get_optional_param(param_name, defval);
        return get_value(valname, params);
    }

  };


  template <class T, class U>
  class SSTFactory1InitParam : public SSTFactory<T>
  {
    private:
        typedef SSTFactory<T> parent;

    public:
        typedef boost::intrusive_ptr<T> ptr;
        typedef boost::intrusive_ptr<U> param1ptr;

      static ptr
      get_value(const std::string& valname,
        const sim_parameters::ptr& params,
        const param1ptr& param1)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1);
        return p;
      }

        static ptr
        get_param(const std::string& param_name,
          const sim_parameters::ptr& params,
          const param1ptr& param1)
        {
          ptr p = parent::get_param(param_name, params);
          p->init(param1);
          return p;
        }

        static ptr
        get_optional_param(const std::string& param_name,
          const std::string& defval,
          const sim_parameters::ptr& params,
          const param1ptr& param1)
        {
          ptr p = parent::get_optional_param(param_name, defval, params);
          p->init(param1);
          return p;
        }
  };

  template <class T, class U, class V>
  class SSTFactory2InitParams : public SSTFactory<T>
  {
    private:
        typedef SSTFactory<T> parent;

    public:
        typedef boost::intrusive_ptr<T> ptr;
        typedef boost::intrusive_ptr<U> param1ptr;
        typedef boost::intrusive_ptr<V> param2ptr;

      static ptr
      get_value(const std::string& valname,
        const sim_parameters::ptr& params,
        const param1ptr& param1,
        const param2ptr& param2)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1, param2);
        return p;
      }

        static ptr
        get_param(const std::string& param_name,
          const sim_parameters::ptr& params,
          const param1ptr& param1,
          const param2ptr& param2)
        {
          ptr p = parent::get_param(param_name, params);
          p->init(param1, param2);
          return p;
        }

        static ptr
        get_optional_param(const std::string& param_name,
          const std::string& defval,
          const sim_parameters::ptr& params,
          const param1ptr& param1,
          const param2ptr& param2)
        {
          ptr p = parent::get_optional_param(param_name, defval, params);
          p->init(param1, param2);
          return p;
        }
  };

  template <class T, class U, class V, class W>
  class SSTFactory3InitParams : public SSTFactory<T>
  {
    private:
        typedef SSTFactory<T> parent;

    public:
        typedef boost::intrusive_ptr<T> ptr;
        typedef boost::intrusive_ptr<U> param1ptr;
        typedef boost::intrusive_ptr<V> param2ptr;
        typedef boost::intrusive_ptr<W> param3ptr;

      static ptr
      get_value(const std::string& valname,
        const sim_parameters::ptr& params,
        const param1ptr& param1,
        const param2ptr& param2,
        const param3ptr& param3)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1, param2, param3);
        return p;
      }

        static ptr
        get_param(const std::string& param_name,
          const sim_parameters::ptr& params,
          const param1ptr& param1,
          const param2ptr& param2,
          const param3ptr& param3)
        {
          ptr p = parent::get_param(param_name, params);
          p->init(param1, param2, param3);
          return p;
        }

        static ptr
        get_optional_param(const std::string& param_name,
          const std::string& defval,
          const sim_parameters::ptr& params,
          const param1ptr& param1,
          const param2ptr& param2,
          const param3ptr& param3)
        {
          ptr p = parent::get_optional_param(param_name, defval, params);
          p->init(param1, param2, param3);
          return p;
        }
  };

  template<class Parent,class Child>
    class SSTFactory_desc : public SSTDesc_base
    {

    public:
      SSTFactory_desc(const char* name)
      {
        std::string n(name);
        /** This should only ever be declared within
         the static initialization routine. There is therefore
         nothing unsafe about passing "this" out.  This object
         will be persisent throughout the entire simulation */
        SSTFactory<Parent>::register_name(n, this);
      }

      factory_type::ptr
      build(const sim_parameters::ptr& params)
      {
        return Child::construct(params);
      }

    };

}

#define DeclareFactory(type_name) \
    typedef SSTFactory<type_name> type_name##_factory;

#define DeclareFactory1Param(type_name, param1_name) \
    typedef SSTFactory1Param<type_name, param1_name> type_name##_factory;

#define DeclareFactory1InitParam(type_name, param1_name) \
    typedef SSTFactory1InitParam<type_name, param1_name> type_name##_factory;

#define DeclareFactory2InitParams(type_name, param1_name, param2_name) \
    typedef SSTFactory2InitParams<type_name, param1_name, param2_name> type_name##_factory;

#define DeclareFactory3InitParams(type_name, param1_name, param2_name, param3_name) \
    typedef SSTFactory3InitParams<type_name, param1_name, param2_name, param3_name> type_name##_factory;

#define ImplementFactory(type_name) \
    template<> std::map<std::string, SSTDesc_base*>* SSTFactoryTemplateBase<type_name>::object_map_ = 0;

#define SSTRegister(appstr, parent_app, child_app, ...) \
    static sstmac::SSTFactory_desc<parent_app,child_app> child_app##_cd(appstr)


#endif
