/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstream>

#include <sstmac/common/logger.h>

#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/eventmanager.h>
#include <cstdarg>
#include <sstmac/common/types.h>

namespace sstmac
{

  static char sstprintf_str[2048];


  std::string
  sstprintf(const char *fmt, ...)
  {
    va_list args;
    va_start(args, fmt);
    vsprintf(sstprintf_str, fmt, args);
    va_end(args);

    std::string strobj = sstprintf_str;
    return strobj;
  }

  //string logger::user_param = "";
  std::deque<std::string> logger::user_params;
  std::list<logger*> logger::loggers_;
  bool logger::timer_on_ = false;
  ptr_type::ptr logger::timer_;

  const char* logger::black = "\033[0;30m";
  const char* logger::red = "\033[0;31m";
  const char* logger::green = "\033[0;32m";
  const char* logger::yellow = "\033[0;33m";
  const char* logger::blue = "\033[0;34m";
  const char* logger::magenta = "\033[0;35m";
  const char* logger::cyan = "\033[0;36m";
  const char* logger::white = "\033[0;37m";

  bool logger::debug_coloring_ = false;

  logger::logger(const std::string& namesig, std::ostream *os, bool flush,
      const char* color) :
    active_(0), outstream_(os), my_color_(color), always_flush_(flush),
        is_real_(true)
  {
    size_t left = namesig.find("<");
    size_t right = namesig.find(">");
    if (left != std::string::npos && right != std::string::npos)
      {
        category_ = namesig.substr(left + 1, right - left - 1);
        signature_ = namesig.substr(right + 1);
      }
    else
      {
        category_ = "";
        signature_ = namesig;
      }

    //  std::cout << "cat: " << category_ << std::endl;

    // loggers_.push_back(this);
    update_active();
  }

  void
  logger::update_active()
  {
      /** I am already active */
      if (active_)
          return;

    //  std::cout << "-------- logger: UPDATE ACTIVE: " << " ------- " << std::endl;
    std::deque<std::string> tok;

    std::string space = "|";
    pst::BasicStringTokenizer::tokenize(signature_, tok, space);

    std::deque<std::string>::iterator it, end = tok.end();
    std::deque<std::string>::iterator paramit, paramend = user_params.end();

    // std::cout << "user param: " << user_param << std::endl;

    for (it = tok.begin(); it != end; it++)
      {
        std::string temp = *it;

        temp = trim_str(temp);

        for (paramit = user_params.begin(); paramit != paramend; paramit++)
          {
            std::string t2 = trim_str(*paramit);

            if (t2 == "timestamps")
              {
                timer_on_ = true;
              }
            else
              {

                size_t left = t2.find("<");
                size_t right = t2.find(">");
                std::string user_cat = "";
                std::string user_name = t2;
                if (left != std::string::npos && right != std::string::npos)
                  {
                    user_cat = t2.substr(left + 1, right - left - 1);
                    user_name = t2.substr(right + 1);
                  }

                user_name = trim_str(user_name);
                int level = 1;
                if(user_name.find("(") != std::string::npos){ // setting a level also
                    std::string levstr = user_name.substr(user_name.find("(") + 1, user_name.find(")"));
                    level = atoi(levstr.c_str());
                    user_name = user_name.substr(0, user_name.find("("));
                }

                // std::cout << "user tok: " << t2 << std::endl;
                if (user_cat == "" || user_cat == category_)
                  {
                    if (user_name == temp || user_name == "all")
                      {
                        active_ = level;
                        return;
                      }
                  }
              }
          }

      }

    //std::cout << "done with update_active()" << std::endl;
  }

  std::string
  logger::to_string() const
  {
      return "logger";
  }

  void
  logger::update_all_active()
  {
    std::list<logger*>::iterator it, end = loggers_.end();
    for (it = loggers_.begin(); it != end; it++)
      {
        (*it)->update_active();
      }
    //std::cout << "done with update_all_active()" << std::endl;
  }

  logger::~logger()
  {

    //std::cout << "logger destructor: " << signature_ << std::endl;
    //loggers_.remove(this);
  }

  void
  logger::set_user_param(const std::string& s)
  {
    std::string tosep = s;

    timer_on_ = false;

    std::deque<std::string> tok;

    std::string space = "|";
    pst::BasicStringTokenizer::tokenize(tosep, tok, space);
    user_params = tok;

    update_all_active();
  }

  void
  logger::clear()
  {
    loggers_.clear();
  }

  std::string
  logger::time() const {
    if(timer_on_ && timer_){
        //stringstream ss;
        eventmanager::ptr evman = boost::dynamic_pointer_cast<eventmanager>(
            timer_);
        timestamp t = evman->now();
        return sstprintf("[t=%12.4e] ", t.sec());
       // ss << "[t=" << t.sec() << "] ";
        //return ss.str();
      }
    else
      return "";
  }

  logger&
  logger::operator<<(int i)
  {

    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(long i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(unsigned int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(long long int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(unsigned long int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(unsigned long long int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(bool i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(const char* i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(double i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(long double i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(const std::string& i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(const std::pair<int, int>& p)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << "(" << p.first << ", " << p.second << ")"
          << black;
    else
      (*outstream_) << "(" << p.first << ", " << p.second << ")";

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(timestamp i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  void
  logger::clear_statics()
  {
      loggers_.clear();
      user_params.clear();
      timer_ = 0;
  }
} // end of namespace sstmac
