/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_RESOURCE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_RESOURCE_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/eventmanager.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/process/os_api.h>


namespace sstmac
{

    class resource_event;
    class resource;

    class resource :
        virtual public eventmodule
    {

        public:
            typedef boost::intrusive_ptr<resource> ptr;

            static const int no_limit = -1;

            virtual ~resource();

            void release();

            static ptr
            construct();

            static ptr
            construct(int nmax);

            int use();

            int nmax() const;

            void add_event(const eventcallback::ptr& event);

            virtual void handle(const sst_message::ptr &msg);

        protected:

            long long ndelays_;

            timestamp total_delay_;

            resource(int nmax = no_limit);

            static std::list<resource::ptr> all_resources_;

            static void register_resource(const resource::ptr& res);

            std::list<eventcallback::ptr> pending_events_;

            static debug dbg_;

            int nactive_;

            int nmax_;

    };

    template <
        class Cls,
        typename Fxn,
        typename Arg1>
    void use_resource(
        const boost::intrusive_ptr<Cls>& obj,
        Fxn fxn,
        const Arg1& arg1
    )
    {
        int nactive = obj->use();
        int nmax = obj->nmax();
        if (nmax == resource::no_limit || nactive <= nmax){ //we're good, run it
            (obj.get()->*fxn)(arg1);
        }
        else{ //resources are tied up right now
            eventcallback::ptr cb = new eventcallback_tmpl_1_args<Cls, Fxn, Arg1>
                    (obj, fxn, arg1);
            obj->add_event(cb);
        }
    }

    template <
        class Cls,
        typename Fxn,
        typename Arg1,
        typename Arg2>
    void use_resource(
        const boost::intrusive_ptr<Cls>& obj,
        Fxn fxn,
        const Arg1& arg1,
        const Arg2& arg2
    )
    {
        int nactive = obj->use();
        int nmax = obj->nmax();
        if (nmax == resource::no_limit || nactive <= nmax){ //we're good, run it
            (obj.get()->*fxn)(arg1, arg2);
        }
        else{ //resources are tied up right now
            eventcallback::ptr cb = new eventcallback_tmpl_2_args<Cls, Fxn, Arg1, Arg2>
                    (obj, fxn, arg1, arg2);
            obj->add_event(cb);
        }
    }

    template <
        class Cls,
        typename Fxn,
        typename Arg1,
        typename Arg2,
        typename Arg3>
    void use_resource(
        const boost::intrusive_ptr<Cls>& obj,
        Fxn fxn,
        const Arg1& arg1,
        const Arg2& arg2,
        const Arg3& arg3
    )
    {
        int nactive = obj->use();
        int nmax = obj->nmax();
        if (nmax == resource::no_limit || nactive <= nmax){ //we're good, run it
            (obj.get()->*fxn)(arg1, arg2, arg3);
        }
        else{ //resources are tied up right now
            eventcallback::ptr cb = new eventcallback_tmpl_3_args<Cls, Fxn, Arg1, Arg2, Arg3>
                    (obj, fxn, arg1, arg2, arg3);
            obj->add_event(cb);
        }
    }
} //end of namespace sstmac

#endif
