/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_STATS_STATS_COMMON_H_INCLUDED
#define SSTMAC_COMMON_STATS_STATS_COMMON_H_INCLUDED

#include <sstmac/common/errors.h>
#include <sstmac/common/stats/stat_logger.h>
#include <iostream>
#include <fstream>
#include <sstmac/common/timestamp.h>

namespace sstmac
{

  /**
   * this stat_logger class just adds up all the records given to it,
   * and outputs it to file when destructed.  values can be associated with
   * std::string keys for storing multiple things
   */
  class stat_sum : public stat_logger
  {
  public:
    typedef boost::intrusive_ptr<stat_sum> ptr;

  protected:

    stat_sum(const std::string& namesig, const std::string& filename) :
      stat_logger(namesig, filename)
    {
    }

  public:
    static ptr
    construct(const std::string& namesig, const std::string& filename)
    {
      return new stat_sum(namesig, filename);
    }

    virtual void
    simulation_finished(timestamp end)
    {
      if (is_active() && check_open())
        {
          myfile << "stat_sum(" << signature_ << ") \n";
          std::map<std::string, double>::iterator it, end = val_.end();
          for (it = val_.begin(); it != end; it++)
            {
              myfile << it->first << " ," << it->second << "\n";
            }
          myfile.close();
        }
    }

    virtual
    ~stat_sum()
    {
    }

    virtual void
    record(double i, std::string key = "default")
    {
      if (val_.find(key) == val_.end())
        val_[key] = 0;
      val_[key] += i;
    }

  protected:

    std::map<std::string, double> val_;

  };

  /**
   * this stat_logger class keeps a spy plot
   */
  class stat_spyplot : public stat_logger
  {
  public:
    typedef boost::intrusive_ptr<stat_spyplot> ptr;

  protected:

    stat_spyplot(const std::string& namesig, const std::string& filename) :
      stat_logger(namesig, filename)
    {
    }

  public:
    static ptr
    construct(const std::string& namesig, const std::string& filename)
    {
      return new stat_spyplot(namesig, filename);
    }

    virtual std::string
    to_string() const
    {
      return "stat_spyplot";
    }

    virtual void
    simulation_finished(timestamp end);

    virtual
    ~stat_spyplot()
    {
    }

    virtual void
    add_one(long source, long dest);

    virtual void
    add(long source, long dest, long num);

  protected:

    nested_type<long, long, long>::map vals_;

  };

  /**
   * this stat_logger class keeps a histogram
   */
  template<typename BucketType>
    class stat_histogram : public stat_logger
    {
    public:
      typedef boost::intrusive_ptr<stat_histogram> ptr;

    protected:

      stat_histogram(const std::string& namesig, const std::string& filename) :
        stat_logger(namesig, filename)
      {
        INTERVAL = 1000;
        MAX = 1<<30;
      }

      struct histobucket
      {
        BucketType min_;
        BucketType max_;

        long count_;
      };

      typename std::map<BucketType, histobucket> buckets_;

    public:
      static ptr
      construct(const std::string& namesig, const std::string& filename)
      {
        return new stat_histogram(namesig, filename);
      }

      virtual std::string
      to_string() const
      {
        return "stat_histogram";
      }

      virtual void
      simulation_finished(timestamp endtime) {
        if (is_active() && check_open())
          {
            //type<BucketType, histobucket>::map::iterator it, end = buckets_.end();

            BucketType start = buckets_.begin()->first;
            long cnt = 0;
            for (start = buckets_.begin()->first; cnt < (int) buckets_.size(); start += INTERVAL)
              {
                if(buckets_.find(start) == buckets_.end()){
                    myfile << start << ", 0 \n";
                }else{
                    myfile << start << ", " << buckets_[start].count_ << "\n";
                    cnt++;
                }

              }

            myfile.close();
          }
      }

      BucketType
      get_bucket_start(BucketType b){
        long div = b / INTERVAL;
        return div * INTERVAL;
      }

      virtual
      ~stat_histogram()
      {
      }

      virtual void
      add_one(BucketType b)
      {
        if (is_active() && b <= MAX)
          {
            BucketType st = get_bucket_start(b);
            if (buckets_.find(st) == buckets_.end())
              {
                histobucket hb;
                hb.min_ = st;
                hb.max_ = st + INTERVAL;
                hb.count_ = 0;
              }

            buckets_[st].count_++;
          }
      }

      virtual void
      add(BucketType b, long num){
        if (is_active() && b <= MAX)
          {
            BucketType st = get_bucket_start(b);
            if (buckets_.find(st) == buckets_.end())
              {
                histobucket hb;
                hb.min_ = st;
                hb.max_ = st + INTERVAL;
                hb.count_ = 0;
              }

            buckets_[st].count_ += num;
          }
      }

      BucketType INTERVAL;
      BucketType MAX;

    protected:



    };

}

#endif
