/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/nic/networkinterface.h>
namespace sstmac
{
  namespace hw
  {
    ImplementFactory(interconnect);

    debug interconnect::dbg_;

    bool interconnect::registered_stats_ = false;
    stat_sum::ptr interconnect::pwrstat_;
    stat_time_window_sum_duration::ptr interconnect::energystat_;
    stat_time_window_sum::ptr interconnect::bandwidthstat_;
    stat_time_window_sum::ptr interconnect::msg_count_;

    interconnect::interconnect(const topology::ptr& t) :
      topology_(t)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> interconnect | network");
          logger::register_logger(&dbg_);
        }

      if (!pwrstat_)
        {
          const std::string fname = "stats_interconnect_pwr.csv";
          pwrstat_ = stat_sum::construct("<stats> interconnect | network",
              fname);
          register_stat(pwrstat_);
        }
      if (!energystat_)
        {
          const std::string fname = "stats_interconnect_energy.csv";
          energystat_ = stat_time_window_sum_duration::construct(
              "<stats> interconnect | network", fname);
          register_stat(energystat_);
        }
      if (!bandwidthstat_)
        {
          const std::string fname = "stats_interconnect_bandwidth.csv";
          bandwidthstat_ = stat_time_window_sum::construct(
              "<stats> interconnect | network", fname);
          register_stat(bandwidthstat_);
        }
      if (!msg_count_)
        {
          const std::string fname = "stats_interconnect_msg_count.csv";
          msg_count_ = stat_time_window_sum::construct(
              "<stats> interconnect | network", fname);
          register_stat(msg_count_);
        }
    }

    void
    interconnect::set_endpoint(const nodeaddress::ptr& addr, const boost::intrusive_ptr<eventhandler>& nic)
    {
      endpoints_[addr] = nic;
    }

    eventhandler::ptr
    interconnect::get_endpoint(nodeaddress::ptr addr) const {
      map_type::const_iterator it = endpoints_.find(addr);
      if(it == endpoints_.end()){
          throw ssterror("interconnect: can't find endpoint at addr " + addr->to_string());
      }
      return it->second;
    }

    eventcallback::ptr
    interconnect::get_injection_callback(const sst_message::ptr& payload){
        eventcallback::ptr cb = new_callback(interconnect::ptr(this), &interconnect::send_now, payload);
        return cb;
    }

    void
    interconnect::send_now(const sst_message::ptr &msg){
        send(now(), msg);
    }

    interconnect::~interconnect()
    {
    }

    std::string
    interconnect::label(long node_id) const
    {
        sst_throw(unimplementederror, "interconnect::label for %s",
                  to_string().c_str());
    }

    void
    interconnect::finish()
    {
        eventmodule::finish();
        endpoints_.clear();
    }

    void
    interconnect::clear_statics()
    {
        pwrstat_ = 0;
        energystat_ = 0;
        bandwidthstat_ = 0;
        msg_count_ = 0;
    }

  }
}
