/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/switchinterconnect.h>
#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>
#include <sstmac/hardware/network/switches/circuit/circuitswitch.h>
#include <sstmac/hardware/network/topology/switchtopology.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("switch", interconnect, switchinterconnect,
                "Models the network interconnect as a topology of connected switches");

    switchinterconnect::switchinterconnect(const topology::ptr& top,
        const sim_parameters::ptr& params, const vis::vis_engine::ptr& veng) :
      interconnect(top), vis_display("interconnect", veng)
    {
      switchtop_ = boost::dynamic_pointer_cast<switchtopology>(top);
      if (!switchtop_)
        {
          sst_throw(ssterror, "switchinterconnect::ctor: topology %s does not support switch model",
                    params->reread_param("topology_name").c_str());
        }

      networkswitch::ptr switchtemp = networkswitch_factory::get_param("network_switchtype", params);

      /** sstkeyword {
            docstring = The number of compute nodes connected to each network switch.  For example,
                        the Cray Gemini network has two nodes per switch;
            gui=2;
      } */
      int nps = params->get_int_param("network_nodes_per_switch");

      // HACK? is top->n() * nps really correct?
      for (long i = 0; i < top->n() * nps; i++)
        {
          nodeaddress::ptr nodeaddr =
              nodeaddress::convert_from_uniqueid(i);
          nodeaddress::ptr netaddr =
              nodeaddress::convert_from_uniqueid(i / nps);

          if (switches_.find(netaddr) == switches_.end())
            {
              networkswitch::ptr sw = switchtemp->clone(
                  switchtop_->create_router(netaddr, params), netaddr);
              switches_[netaddr] = sw;

            }

          injector_[nodeaddr] = switches_[netaddr]->get_injector(params,
              i % nps);

        }

      SSTMAC_DEBUG << "switchinterconnect: created " << switches_.size()
            << " gateways switches for " << top->n() << " nodes \n";

      /**
        sstkeyword {
            gui=4.0GB/s;
            docstring=The bandwidth per physical link in the interconnect;
        }
      */
      double linkbw = params->get_bandwidth_param("network_bandwidth_link");

      std::vector<int> red;
      /**
        sstkeyword { gui=1 1 1;
            docstring=The number of redundant physical links that make up
                      each link[edge] in the topology. These vector dimensions
                      should exactly match those in topology_geometry.
                      The total available bandwidth in each topology link is:ENDL
                      BW=Redundant*Link_BW;
         }
      */
      if (params->has_param("topology_redundant"))
      {
        params->get_vector_param("topology_redundant", red);
      }
      else
      {
        for (int i = 0; i < (int) switchtop_->ndimensions(); i++)
        {
          red.push_back(1);
        }
      }


      switchtop_->connect_switches(switches_, red, linkbw);

      //topology must set ejectors for the case of indirect networks
      switchtop_->set_ejectors(top->n(), nps, switches_, ejector_, params);

      //now they're connected up, initialize them.  Note that
      //  connect_switches could very well have added some switches
      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->initialize(params, nps);
        }

    }

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    void
    switchinterconnect::send(const timestamp &start_time,
        const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "switchinterconnect: sending to "
            << msg->toaddr()->to_string() << " from "
            << msg->fromaddr()->to_string() << " using injector "
            << injector_[msg->fromaddr()]->to_string() << " and ejector "
            << ejector_[msg->toaddr()]->to_string() << "\n";

      // injector_[msg->fromaddr_]->set_sendhandler(endpoints_[msg->fromaddr_], msg->fromaddr_);
      // ejector_[msg->toaddr_]->set_recvhandler(endpoints_[msg->toaddr_], msg->toaddr_);
      schedule(start_time, injector_[msg->fromaddr()], msg);
    }

    switchinterconnect::ptr
    switchinterconnect::construct(const sim_parameters::ptr& params)
    {
        sstmac::hw::topology::ptr top = topology_factory::get_param("topology_name", params);
        return new switchinterconnect(top, params, construct_vis_engine(params));
    }

    eventcallback::ptr
    switchinterconnect::get_injection_callback(const sst_message::ptr& msg)
    {
        event_step_callback::ptr cb = new_event_step_callback(injector_[msg->fromaddr()], &injector::send, msg);
        return cb;
    }

    void
    switchinterconnect::set_endpoint(const nodeaddress::ptr& addr, const eventhandler::ptr& nic)
    {
      interconnect::set_endpoint(addr, nic);
      SSTMAC_DEBUG << "switchinterconnect: setting eventhandler for addr "
            << addr->to_string() << "\n";
      injector_[addr]->set_sendhandler(nic, addr);
      ejector_[addr]->set_recvhandler(nic, addr);
    }

    void
    switchinterconnect::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);

      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->set_eventmanager(m);
        }
    }

    void
    switchinterconnect::finish()
    {
        if (finished_)
            return;

        interconnect::finish();

      { switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->finish();
        } }

      { switchtopology::injmaptype::iterator it = injector_.begin(), end = injector_.end();
      for ( ; it != end; ++it){
          it->second->finish();
      } }

        { switchtopology::ejmaptype::iterator it = ejector_.begin(), end = ejector_.end();
        for ( ; it != end; ++it){
            it->second->finish();
        } }

      injector_.clear();
      ejector_.clear();
    }

    vis::vis_topology::vismaptype
    switchinterconnect::vis_get_nodes()
    {
      vis::vis_topology::vismaptype vismap;
      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          vismap[it->first] = it->second;
        }
      return vismap;
    }

    vis::vis_topology::ptr
    switchinterconnect::vis_get_topology()
    {
      return switchtop_;
    }

  }
}
