/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/traininterconnect.h>
#include <sstmac/hardware/nic/train_nic.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("train", interconnect, traininterconnect,
        "Models an interconnect that sends messages as a packet train");

    void
    traininterconnect::init(const sim_parameters::ptr& params)
    {

      switchtop_ = boost::dynamic_pointer_cast<switchtopology>(topology_);
      if (!switchtop_)
      {
        sst_throw(ssterror, "traininterconnect::ctor: topology %s does not support switch model",
            params->reread_param("topology_name").c_str());
      }

      /** sstkeyword {
            docstring = The number of compute nodes connected to each network switch.  For example,
                        the Cray Gemini network has two nodes per switch;
            gui=2;
      } */
      nps_ = params->get_int_param("network_nodes_per_switch");

      /**
        sstkeyword {
            gui=train;
            docstring=Specify the switch congestion model used in arbitrating
                      for switch access;
        }
      */
      trainswitch::ptr switchtemp = trainswitch_factory::get_param("network_switchtype", params);

      for (long i = 0; i < topology_->n(); i++)
      {
        //nodeaddress::ptr nodeaddr = nodeaddress::convert_from_uniqueid(i);
        nodeaddress::ptr netaddr = nodeaddress::convert_from_uniqueid(i);
        router::ptr router = switchtop_->create_router(netaddr, params);
        trainswitch::ptr sw = switchtemp->clone(router, netaddr);
        SSTMAC_DEBUG << "Adding train switch at node address "
              << netaddr->to_string() << "\n";
        switches_[netaddr] = sw;
        sw->vis_parent_ = ptr(this);
      }

      SSTMAC_DEBUG << "traininterconnect: created " << switches_.size()
            << " gateway switches for " << topology_->n() << " nodes \n";

      /**
        sstkeyword { gui=1.5; docstring=$bytes_per_cycle; }
      */
      double bytes_per_cycle = params->get_double_param(
          "network_train_link_bytes_per_cycle");
      /** sstkeyword { gui=3ghz; docstring=$train_freq; } */
      double freq_hz = params->get_freq_param("network_train_link_freq");
      timestamp link_latency = params->get_time_param(
          "network_train_link_latency");
      /**
        sstkeyword { gui=0.6; docstring=$payload_fraction; }
      */
      double payload_fraction = params->get_optional_double_param(
          "network_train_payload_fraction", 1.0);
      /**
        sstkeyword { docstring=$ignore_congestion; }
      */
      bool ignore_congestion = params->get_optional_bool_param(
          "network_train_link_ignore_congestion", false);
      /**
        sstkeyword { docstring=$enable_multipath; }
      */
      bool enable_multipath = params->get_optional_bool_param(
          "network_train_link_enable_multipath", false);

      /**
        sstkeyword { docstring=$train_arbitrator; }
      */
      output_bw_arbitrator::ptr arb_template = output_bw_arbitrator_factory
              ::get_optional_param("network_train_link_arbitrator", "infinite_buffer", params);

      /**
        sstkeyword { gui=1 1 1;
            docstring=The number of redundant physical links that make up
                      each link[edge] in the topology. These vector dimensions
                      should exactly match those in topology_geometry.
                      The total available bandwidth is:ENDL
                      BW=Redundant*Bytes_Per_Cycle*Freq;
        }
      */
      std::vector<int> red;
      if (params->has_param("topology_redundant"))
      {
        params->get_vector_param("topology_redundant", red);
      }
      else
      {
        for (int i = 0; i < (int) switchtop_->ndimensions(); i++)
        {
          red.push_back(1);
        }
      }

      switchtop_->connect_train_tracks(switches_,
                                       red,
                                       bytes_per_cycle,
                                       freq_hz,
                                       link_latency,
                                       payload_fraction,
                                       ignore_congestion,
                                       enable_multipath,
                                       arb_template);
    }

    traininterconnect::ptr
    traininterconnect::construct(const sim_parameters::ptr& params)
    {
        /** sstkeyword {
                gui=torus;
                docstring=Specify the basic topology type used for the
                          network interconnect;
        }
        */
      sstmac::hw::topology::ptr top = topology_factory::get_param(
          "topology_name", params);
      traininterconnect::ptr ret = new traininterconnect(top,  construct_vis_engine(params));
      ret->init( params);

      // fix some parameters
      if (params->has_param("network_train_injection_arbitrator")){
          std::string arbtype = params->reread_param("network_train_injection_arbitrator");
          params->add_param_override("network_train_rdma_arbitrator", arbtype);
          params->add_param_override("network_train_udp_arbitrator", arbtype);
      }
      if (params->has_param("network_train_injection_latency")){
          std::string lat = params->reread_param("network_train_injection_latency");
          params->add_param_override("network_train_rdma_injection_latency", lat);
          params->add_param_override("network_train_udp_injection_latency", lat);
      }
      if (params->has_param("network_train_injection_bw")){
          std::string bw = params->reread_param("network_train_injection_bw");
          params->add_param_override("network_train_rdma_injection_bw", bw);
          params->add_param_override("network_train_udp_injection_bw", bw);
      }

      return ret;
    }

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    void
    traininterconnect::send(const timestamp &start_time,
        const sst_message::ptr &msg)
    {
      sst_throw(illformederror, "traininterconnect::send shouldn't be used to send messages");
    }

    std::string
    traininterconnect::label(long node_id) const
    {
      long net_id = node_id / nps_;
      long sw_id = node_id % nps_;
      std::string net_label = switchtop_->label(net_id);
      std::string node_label = sstprintf("node(%6ld) %s[%2ld]", node_id,
          net_label.c_str(), sw_id);
      return node_label;
    }

    void
    traininterconnect::set_endpoint(const nodeaddress::ptr& addr,
        const eventhandler::ptr& handler)
    {
      train_nic::ptr nic = safe_cast<train_nic> (handler);
      interconnect::set_endpoint(addr, nic);
      nodeaddress::ptr netaddr = nodeaddress::convert_from_uniqueid(
          addr->unique_id() / nps_);
      SSTMAC_DEBUG
            << sstprintf(
                "traininterconnect::set_endpoint: adding node %s in network at %s\n",
                addr->to_string().c_str(), netaddr->to_string().c_str());
      trainswitch::ptr sw = switches_[netaddr];

      nic->set_switch(sw);
    }

    void
    traininterconnect::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);

      switchtopology::trainmap::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        it->second->set_eventmanager(m);
      }
    }

    traininterconnect::~traininterconnect()
    {
    }

    void
    traininterconnect::finish()
    {
      if (finished_)
        return;

      interconnect::finish();

      switchtopology::trainmap::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        it->second->finish();
      }
      switches_.clear();

    }

    vis::vis_topology::vismaptype
    traininterconnect::vis_get_nodes()
    {
      vis::vis_topology::vismaptype vismap;
      switchtopology::trainmap::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        vismap[it->first] = it->second;
      }
      return vismap;
    }

    vis::vis_topology::ptr
    traininterconnect::vis_get_topology()
    {
      return switchtop_;
    }

  }
}

