/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>
#include <sstmac/hardware/network/topology/router.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("packet", networkswitch, packetswitch);

    stat_time_window_avg::ptr packetswitch::stat_Q_wait_ = NULL;
    stat_time_window_sum::ptr packetswitch::stat_bytes_ = NULL;
    stat_time_window_sum_duration::ptr packetswitch::stat_pwr_ = NULL;

    packetswitch::packetswitch()
    {
    }

    networkswitch::ptr
    packetswitch::clone(router::ptr route, nodeaddress::ptr addr)
    {
      ptr ret = new packetswitch;
      ret->init(addr, route);
      return ret;
    }

    injector::ptr
    packetswitch::get_new_injector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetinjector::construct(params, self);

    }

    ejector::ptr
    packetswitch::get_new_ejector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetejector::construct(params, self, my_addr_);
    }

    packetswitch::~packetswitch()
    {
    }

    void
    packetswitch::initialize(sim_parameters::ptr params, int nps)
    {
      if (!stat_Q_wait_)
        stat_Q_wait_ = stat_time_window_avg::construct("<stats> network",
            "stats_packet_Qwait.csv");

      if (!stat_bytes_)
        stat_bytes_ = stat_time_window_sum::construct("<stats> network",
            "stats_packet_bytes.csv");

      if (!stat_pwr_)
        stat_pwr_ = stat_time_window_sum_duration::construct("<stats> network",
            "stats_power_network.csv");

      lat_r2r_ = params->get_time_param("packetswitch_latency_r2r");

      double bw = params->get_bandwidth_param("packetswitch_bandwidth_n2r");

      nps_ = nps;

      // HACK i=0 originally
      // for (int i = router::EJECT; i < nps; i++)
      for (int i = 0; i < nps; i++)
      {
          portbw_[EJECT_PORT][i] = bw;
          //make singly redundant
          finishtimes_[EJECT_PORT][i] = packet_finish(1);
      }

      redundRNG_ = RNG::SimpleCombo::construct(time(NULL));
      srand(time(NULL));


      maxtime_ = timestamp(0);
    }

    void
    packetswitch::finish()
    {
        networkswitch::finish();
        stat_Q_wait_ = 0;
        stat_pwr_ = 0;
        stat_bytes_ = 0;
    }

    void
    packetswitch::handle(const sst_message::ptr &msg)
    {

      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
        {
          throw ssterror("packetswitch::handle() - can't cast to packetmessage");
        }

      int dim;
      int dir;

      pack->rinfo()->set_congestion(finishtimes_, maxtime_, now());
      router_->route(pack->rinfo(), dim, dir);

      if (redund_.find(dim) == redund_.end())
        {
          redund_[dim] = boost::unordered_map<int, int>();
        }

      if (redund_[dim].find(dir) == redund_[dim].end())
        {
          redund_[dim][dir] = 1;
        }

      if(redund_[dim][dir] < 1){
          redund_[dim][dir] = 1;
      }

      SSTMAC_DEBUG << "dim: " << dim << ", dir: " << dir 
                   << ", redund: " << redund_[dim][dir] 
                   << ", portbw: " << portbw_[dim][dir] << "\n";

      //int r = rand() % redund_[dim][dir];
      packet_finish& finish = finishtimes_[dim][dir];
      int r = finish.round_robin_index;
      //update the round robin index for the next packet
      finish.increment_index();
      timestamp txFinishTime = finish.times[r];

      if (txFinishTime < now())
        txFinishTime = now();
      timestamp delay = lat_r2r_ + timestamp(
          pack->get_byte_length() / portbw_[dim][dir]);

      SSTMAC_DEBUG << "message going to dim " << dim << " and dir " << dir
            << " which is switch " << outports_[dim][dir]->to_string()
            << " with delay " << delay << "\n";
      schedule(txFinishTime + delay, outports_[dim][dir], msg);

      finish.times[r] = txFinishTime + delay;

      if(finish.times[r] - now() > maxtime_){
          maxtime_ = finish.times[r] - now();
      }

      double wait = (txFinishTime.sec() - now().sec());
      stat_Q_wait_->record(txFinishTime, wait);

      stat_bytes_->record(txFinishTime, pack->get_byte_length());

    }

    void
    packetswitch::packetinjector::finish()
    {
        parent_ = 0;
        injector::finish();
    }

    void
    packetswitch::packetinjector::send(long nbytes, long byte_offset, const sst_message::ptr& msg)
    {
        timestamp delay(0);
        long bytes_sent = byte_offset;
        long bytes_total = msg->get_byte_length();
        bool tail_packet = false; //assume false for now
        if (MTU_ > 0)
          {
            SSTMAC_DEBUG << "packetinjector: injecting packets: "
                         << msg->to_string()
                         << ", bytes-to-send=" << nbytes
                         << ", bytes-offset=" << byte_offset
                         << ", total-msg-size=" << msg->get_byte_length()
                         << ", MTU=" << MTU_
                         << "\n";

            long bytes_left = std::min(nbytes, bytes_total - byte_offset);
            while (bytes_left > 0)
              {
                long bytes_to_send = std::min(MTU_, bytes_left);
                bytes_sent += bytes_to_send;
                bytes_left -= bytes_to_send;
                tail_packet = bytes_sent >= bytes_total;
                packetmessage::ptr pack = packetmessage::construct(msg,
                    bytes_to_send, tail_packet);
                parent_->schedule(parent_->now() + delay, parent_, pack);
                delay = delay + timestamp(bytes_to_send / n2r_);
              }
          }
        else
          {
            delay = timestamp(nbytes / n2r_);
            SSTMAC_DEBUG << "packetinjector: injecting whole message: "
                         << msg->to_string() << " with delay " << delay
                         << ", bytes-to-send="
                         << ", bytes-offset=" << byte_offset
                         << ", total-msg-size=" << msg->get_byte_length()
                         << ", bw=" << n2r_
                         <<  "\n";

            bytes_sent = byte_offset + nbytes;
            tail_packet = bytes_sent >= bytes_total;
            parent_->schedule(parent_->now() + delay, parent_,
                packetmessage::construct(msg, nbytes, tail_packet));
          }


        if (msg->get_needs_ack() && tail_packet)
          {
            sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
            parent_->schedule(parent_->now() + delay,
                get_sendhandler(msg->fromaddr()), sendack);
          }
    }

    void
    packetswitch::packetejector::finish()
    {
        parent_ = 0;
        ejector::finish();
    }

    void
    packetswitch::connect(int dim, int dir, networkswitch::ptr sw, double bw, int redund)
    {
        SSTMAC_DEBUG << "packetswitch::connect dim:" << dim 
                     << " dir:" << dir << "\n";
        networkswitch::connect(dim, dir, sw, bw, redund);
        finishtimes_[dim][dir] = packet_finish(redund);
    }

    void
    packetswitch::packetejector::handle(const sst_message::ptr &msg)
    {
      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
        {
          throw ssterror(
              "packetswitch::packetejector::handle() - can't cast to packetmessage");
        }

      /** send all packets to the nic now */
      bool send_to_nic = true; //pack->is_tail()

      if (send_to_nic)
        {
          if (recvhandles_.find(pack->orig()->toaddr()) == recvhandles_.end())
            {
              std::cerr << "my addr: " << addr_->to_string() << "\n";
              std::cerr << "message headed to: "
                  << pack->orig()->toaddr()->to_string() << "\n";
              std::cerr << "message came from: "
                  << pack->orig()->fromaddr()->to_string() << "\n";
              throw ssterror(
                  "packetswitch::packetejector::handle() -- can't find recv handler, that's a problem");
            }
          SSTMAC_DEBUG << "packetejector: ejecting message to "
                << pack->orig()->toaddr()->to_string() << ": "
                << msg->to_string() << "\n";
          parent_->schedule(parent_->now(),
              recvhandles_[pack->orig()->toaddr()], pack);
        }
    }

  }
}
