/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// butterfly.cc: Implementation of butterfly networks.
//
// Based on butterfly.c by Jim Schutt
// Adapted by Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstmac/hardware/network/topology/butterfly.h>
#include <sstmac/hardware/network/topology/butterflyrouter.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("butterfly | bfly", topology, butterfly);

    butterfly::butterfly(int n, int r, int nps) :
      switchtopology(nps),
      n_(n), r_(r)
    {
      stages_ = std::log(n_) / std::log(r_);
    }

    butterfly::ptr
    butterfly::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      return butterfly::ptr(new butterfly(args[0], args[1], nps));
    }

    void
    butterfly::connect_switches(switchtopology::maptype &switches,
                                const std::vector<int>& red,
                                double linkbw)
    {

      /*  int numperstage = switches.size();

       networkswitch::ptr templ = switches.begin()->second;
       std::vector<networkswitch::ptr> prev(switches.size());
       for (int i = 0; i < switches.size(); i++)
       prev[i] = switches[nodeaddress::convert_from_uniqueid(i)];

       for (int s = 1; s < stages_; s++)
       {
       std::vector<networkswitch::ptr> sw(numperstage);
       SSTMAC_DEBUG << "butterfly: stage " << s << "\n";
       for (int n = 0; n < numperstage; n++)
       {
       nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(
       s * numperstage + n);
       SSTMAC_DEBUG << "butterfly: switch " << n << " with address "
       << addr->to_string() << "\n";

       networkswitch::ptr newsw = templ->clone(
       create_router(addr, params), addr);
       sw[n] = newsw;

       }

       //connect the previous stage
       long jump = pow(r_, s - 1);
       for (int n = 0; n < numperstage; n++)
       {
       nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(
       (s-1) * numperstage + n);
       for (int r = 0; r < r_; r++)
       {
       int ind = pow(r_, s - 1) * r + (n % r_);

       SSTMAC_DEBUG << "butterfly: connected "
       << prev[ind]->get_addr()->to_string() << " to "
       << sw[n]->get_addr()->to_string() << " at dir " << r
       << "\n";
       prev[ind]->connect(0, r, sw[n]);

       }
       }
       prev = sw;
       }*/

      long edgeswitches = switches.size();
      long swid = edgeswitches;
      switchtopology::maptype prevlevel = switches;
      long thislevelstartid;
      long prevlevelstartid = 0;
      long crossfactor = 1;
      for (int l = 1; l < stages_; l++)
        {
          thislevelstartid = swid;
          switchtopology::maptype thislevel;

          int portstofill = prevlevel.size() * r_;
          int portsdown;

          portsdown = r_;

          int numswitches = portstofill / portsdown;

          SSTMAC_DEBUG << "butterfly: creating " << numswitches
                << " for level " << l << "\n";
          networkswitch::ptr templ = switches.begin()->second;
          for (int i = 0; i < numswitches; i++)
            {
              nodeaddress::ptr swaddr =
                  nodeaddress::convert_from_uniqueid(swid);
              networkswitch::ptr sw = templ->clone(
                  create_router(swaddr, sstmac_env::params), swaddr);
              thislevel[swaddr] = sw;
              switches[swaddr] = sw;
              swid++;
            }

          long thisid = thislevelstartid;
          long logicalid = 0;
          long groupsize = pow(r_, l);
          long upgroupstart = 0;

          //track to the right switch in this level

          for (long previd = prevlevelstartid;
               previd < int(prevlevelstartid + prevlevel.size());
               previd++)
            {
              nodeaddress::ptr prevaddr = nodeaddress::convert_from_uniqueid(
                  previd);

              int div = logicalid / groupsize;
              upgroupstart = thislevelstartid + div * groupsize;

              thisid = upgroupstart + logicalid % groupsize;

              for (int p = 0; p < r_; p++)
                {
                  nodeaddress::ptr thisaddr =
                      nodeaddress::convert_from_uniqueid(thisid);

                  SSTMAC_DEBUG << "connecting up from "
                        << prevaddr->to_string() << " to "
                        << thisaddr->to_string() << " on up port " << p
                        << " and down port " << logicalid / crossfactor << "\n";
                  prevlevel[prevaddr]->connect(0, p, thislevel[thisaddr],
                      linkbw); // up
                  //  thislevel[thisaddr]->connect(1, logicalid / crossfactor,
                  //      prevlevel[prevaddr]); //down
                  // SSTMAC_DEBUG << "thisid: " << thisid << ", crossfactor: "
                  //                      << crossfactor << ", upgroupstart: " << upgroupstart
                  //                     << ", groupsize: " << groupsize << "\n";

                  if (thisid + crossfactor >= upgroupstart + groupsize)
                    thisid = upgroupstart
                        + ((thisid + crossfactor) % groupsize);
                  else if (thisid + crossfactor - thislevelstartid >= (int) thislevel.size())
                    {
                      thisid = thislevelstartid + (thisid + crossfactor)
                          % thislevel.size();
                    }
                  else
                    thisid += crossfactor;

                  // SSTMAC_DEBUG << "thisid now: " << thisid << "\n";


                }
              logicalid++;

            }

          prevlevel = thislevel;
          prevlevelstartid = thislevelstartid;
          crossfactor *= r_;
        }

    }

    void
    butterfly::set_ejectors(long num, int nps, const maptype& switches,
        ejmaptype &ejectors, const sim_parameters::ptr& params)
    {
      int sub = (num / r_);
      for (long i = 0; i < num * nps; i++)
        {

          nodeaddress::ptr nodeaddr = nodeaddress::convert_from_uniqueid(i);

          nodeaddress::ptr swaddr = nodeaddress::convert_from_uniqueid(
              switches.size() - sub + (i / r_));

          const networkswitch::ptr& swtch = mapget(switches, swaddr, "switch for butterfly");
          SSTMAC_DEBUG << "butterfly: setting ejector for node "
                << nodeaddr->to_string() << " to eject from switch "
                << swaddr->to_string() << " with ejector "
                << swtch->get_ejector(params, i)->to_string()
                << "\n";
          ejectors[nodeaddr] = swtch->get_ejector(params, i % nps);

        }
    }

  }
} //end of namespace sstmac
