/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// crossbar.cc: Implementation of crossbar networks.
//
// Author: Curtis Janssen <cljanss@sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/crossbar.h>
#include <sstmac/hardware/network/topology/crossbarrouter.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("crossbar | xbar", topology, crossbar);

    crossbar::ptr
    crossbar::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      return crossbar::ptr(new crossbar(args[0], nps));
    }

    crossbar::crossbar(long nnode, int nps) :
      switchtopology(nps), nps_(nps), size_(nnode)

    {
      nodes_.clear();
      for (int i = 0; i < nnode; i++)
        {
          std::vector<int> node(1);
          node[0] = i;
          nodes_.insert(node);
        }
      init();
    }

    int64_t
    crossbar::diameter() const
    {
      return 2;
    }

    int64_t
    crossbar::topology_size() const
    {
      return size_;
    }

    std::string
    crossbar::name() const
    {
      std::ostringstream ostr;
      ostr << "crossbar(" << n() << ")";
      return ostr.str();
    }

    void
    crossbar::connect_switches(switchtopology::maptype &switches,
                               const std::vector<int>& red,
                               double linkbw)
    {
      for (uli i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

          for (uli j = 0; j < switches.size(); j++)
            {
              nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(j);
              if (i != j)
                {
                  switches[me]->connect(0, j, switches[them], linkbw);
                }

            }

        }

    }

    void
    crossbar::connect_train_tracks(trainmap& switches,
                                   const std::vector<int>& red,
                                   double bytes_per_cycle,
                                   double freq_hz,
                                   timestamp link_latency,
                                   double payload_fraction,
                                   bool ignore_congestion,
                                   bool enable_multipath,
                                   const output_bw_arbitrator::ptr& arb_template)

    {
      packettrain::init_statics();

      for (uli i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

          for (uli j = 0; j < switches.size(); j++)
            {
              nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(j);
              if (i != j)
                {
                  std::stringstream ss;
                  ss << i << " -> " << j;
                  switches[me]->connect(0, j, switches[them], link_latency,
                      bytes_per_cycle, freq_hz, payload_fraction,
                      enable_multipath, ignore_congestion,
                      arb_template->clone(), "network",
                      ss.str());

                }

            }

        }

    }

  }
} //end of namespace sstmac
