/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// crossbar.h: Interface for crossbar networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CROSSBAR_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CROSSBAR_H_INCLUDED

#include <sstmac/hardware/network/topology/switchtopology.h>

namespace sstmac
{
  namespace hw
  {

    /** The crossbar network generates a network which connects
     all nodes with only two hops: those to and from the crossbar.
     The inputs for this network can specify the following attributes:

     <table border="1">

     <tr><td>Keyword<td>Type<td>Default<td>Description

     <tr><td><tt>n</tt><td>int<td>256<td>The radix of the crossbar.

     </table>
     */

    class crossbar : public switchtopology
    {
    private:
      int nps_;
      long size_;

    protected:
      crossbar(long nnode, int nps);

    public:
      typedef boost::intrusive_ptr<crossbar> ptr;

      static crossbar::ptr
      construct(sim_parameters::ptr params);

      virtual
      ~crossbar()
      {
      }

      virtual int
      radix() const
      {
        return n();
      }

      int64_t
      diameter() const;

      std::string
      name() const;

      virtual int ndimensions() const {
          return 1;
      }

      virtual void
      connect_switches(switchtopology::maptype &switches,
                       const std::vector<int>& red,
                       double linkbw);

      virtual void
      connect_train_tracks(trainmap& switches,
                           const std::vector<int>& red,
                           double bytes_per_cycle,
                           double freq_hz,
                           timestamp link_latency,
                           double payload_fraction,
                           bool ignore_congestion,
                           bool enable_multipath,
                           const output_bw_arbitrator::ptr& arb_template);


      virtual std::string
      to_string() const
      {
        return "crossbartopology";
      }

      virtual int64_t
      topology_size() const;
    };

  }
} //end of namespace sstmac

#endif
