/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
#include <sstmac/hardware/network/topology/dragonfly.h>
#include <sstmac/hardware/network/topology/dragonflyrouter.h>
#include <math.h>
#include <sstream>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    using namespace std;
    SSTRegister("dragonfly | dfly", topology, dragonfly);

    dragonfly::dragonfly(long x, long y, long g, int gc, int nps) :
      switchtopology(nps), x_(x), y_(y), g_(g), group_con_(gc)
    {

    }

    dragonfly::ptr
    dragonfly::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      int group_con = params->get_int_param("topology_group_connections");
      return ptr(new dragonfly(args[0], args[1], args[2], group_con, nps));
    }

    void
    dragonfly::get_coords(long uid, int &x, int &y, int &g)
    {
      x = uid % x_;
      y = (uid / x_) % y_;
      g = uid / (x_ * y_);
    }

    std::string
    dragonfly::coord_string(int x, int y, int g)
    {
      std::stringstream ss;
      ss << "(" << x << ", " << y << ", " << g << ")";

      return ss.str();
    }

    long
    dragonfly::get_uid(int x, int y, int g)
    {
      return g * (x_ * y_) + y * (x_) + x;
    }

    /// coordinates go x,y - group
    void
    dragonfly::connect_switches(switchtopology::maptype &switches,
        const std::vector<int>& red, double linkbw)
    {

      bool outputgraph = sstmac_env::params->get_optional_bool_param(
          "topology_output_graph", false);

      for (long i = 0; i < switches.size(); i++)
      {
        std::vector<int> connected;
        std::vector<int> connected_red;

        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
        int myx;
        int myy;
        int myg;
        get_coords(i, myx, myy, myg);

        bool errorflag = false;

        double opt_linkbw = get_bandwidth(
            sstmac_env::params->get_param("optical_link_bandwidth").c_str());

        for (int x = 0; x < x_; x++)
        {
          if (x != myx)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(x, myy, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(x, myy, myg) << "\n";
            switches[me]->connect(0, x, switches[them], linkbw, red[0]);

            if (outputgraph)
            {
              connected.push_back(them->unique_id());
              connected_red.push_back(red[0]);

            }
          }
        }

        for (int y = 0; y < y_; y++)
        {
          if (y != myy)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(myx, y, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, y, myg) << "\n";
            switches[me]->connect(1, y, switches[them], linkbw, red[1]);

            if (outputgraph)
            {

              connected.push_back(them->unique_id());
              connected_red.push_back(red[1]);
            }
          }
        }

        //each switch connects to a max of 5 different groups

        int gspace = max(1, g_ / group_con_);
        int myid = myx + myy * x_;
        int gset = myid % g_;
        for (int g = 0; g < group_con_; g++)
        {

          int gid = (gset + gspace * g) % g_;
          if (gid != myg)
          {
            long uid = get_uid(myx, myy, gid);
            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(uid);

            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, myy, gid) << "\n";
            switches[me]->connect(2, gid, switches[them], opt_linkbw, red[2]);

            if (outputgraph)
            {

              connected.push_back(them->unique_id());
              connected_red.push_back(red[2]);
            }
          }

        }

        if (outputgraph)
        {
          for (int k = 0; k < nps_; k++)
          {
            std::cout << (i * nps_ + k);
            for (int l = 0; l < connected.size(); l++)
            {
              for (int m = 0; m < nps_; m++)
              {
                std::cout << " " << (connected[l] * nps_ + m) << ":"
                    << connected_red[l];

              }
            }

            for (int m = 0; m < nps_; m++)
            {
              if (m != k)
              {
                std::cout << " " << (i * nps_ + m) << ":" << 100;
              }
            }
            std::cout << "\n";
          }

        }

      }

    }

    static std::string
    get_link_name(const std::vector<int>& src, const std::vector<int>& dst)
    {
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) src.size(); ++i)
      {
        sstr << sstprintf(" %2d", src[i]);
      }
      sstr << " ]->[";
      for (int i = 0; i < (int) dst.size(); ++i)
      {
        sstr << sstprintf(" %2d", dst[i]);
      }
      sstr << "]";
      return sstr.str();
    }

    std::string
    dragonfly::label(long id)
    {
      int x, y, g;

      get_coords(id, x, y, g);
      std::stringstream sstr;
      sstr << "[";
      sstr << sstprintf(" %2d", x);
      sstr << sstprintf(" %2d", y);
      sstr << sstprintf(" %2d", g);

      sstr << " ]";
      return sstr.str();
    }

    void
    dragonfly::connect_train_tracks(trainmap& switches,
        const std::vector<int>& red, double bytes_per_cycle, double freq_hz,
        timestamp link_latency, double payload_fraction,
        bool ignore_congestion, bool enable_multipath,
        const output_bw_arbitrator::ptr& arb_template)
    {
      packettrain::init_statics();

      double opt_linkbw = sstmac_env::params->get_optional_double_param(
          "network_train_optical_freq_ghz", freq_hz / 1e9);

      opt_linkbw = opt_linkbw * 1e9;

      bool outputgraph = sstmac_env::params->get_optional_bool_param(
          "topology_output_graph", false);

      bool errorflag = false;

      for (long i = 0; i < switches.size(); i++)
      {
        std::vector<int> connected;
        std::vector<int> connected_red;

        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
        int myx;
        int myy;
        int myg;
        get_coords(i, myx, myy, myg);

        for (int x = 0; x < x_; x++)
        {
          if (x != myx)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(x, myy, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(x, myy, myg) << "\n";

            trainmap::iterator it1 = switches.find(them);

            if (it1 == switches.end())
            {
              sst_throw(
                  ssterror,
                  "cray_dragon::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                  get_uid(x, myy, myg), them->to_string().c_str());
            }

            std::vector<int> mycoords(3), theircoords(3);
            mycoords[0] = myx;
            mycoords[1] = myy;
            mycoords[2] = myg;

            theircoords[0] = x;
            theircoords[1] = myy;
            theircoords[2] = myg;

            std::string link_name = get_link_name(mycoords, theircoords);

            switches[me]->connect(0, x, it1->second, link_latency,
                bytes_per_cycle * red[0], freq_hz, payload_fraction,
                enable_multipath, ignore_congestion, arb_template->clone(), "network",
                link_name);

            if (outputgraph)
            {
              connected.push_back(them->unique_id());
              connected_red.push_back(red[0]);

            }
          }
        }

        for (int y = 0; y < y_; y++)
        {
          if (y != myy)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(myx, y, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, y, myg) << "\n";

            trainmap::iterator it1 = switches.find(them);

            if (it1 == switches.end())
            {
              sst_throw(
                  sstmac::ssterror,
                  "cray_dragon::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                  get_uid(myx, y, myg), them->to_string().c_str());
            }

            std::vector<int> mycoords(3), theircoords(3);
            mycoords[0] = myx;
            mycoords[1] = myy;
            mycoords[2] = myg;

            theircoords[0] = myx;
            theircoords[1] = y;
            theircoords[2] = myg;

            std::string link_name = get_link_name(mycoords, theircoords);

            switches[me]->connect(1, y, it1->second, link_latency,
                bytes_per_cycle * red[1], freq_hz, payload_fraction,
                enable_multipath, ignore_congestion, arb_template->clone(),
                "network", link_name);

            if (outputgraph)
            {

              connected.push_back(them->unique_id());
              connected_red.push_back(red[1]);
            }
          }
        }

        //each switch connects to a max of 5 different groups
        int gspace = max(1, g_ / group_con_);
        int myid = myx + myy * x_;
        int gset = myid % g_;
        for (int g = 0; g < group_con_; g++)
        {

          int gid = (gset + gspace * g) % g_;
          if (gid != myg)
          {
            long uid = get_uid(myx, myy, gid);
            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(uid);

            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, myy, gid) << "\n";

            trainmap::iterator it1 = switches.find(them);

            if (it1 == switches.end())
            {
              sst_throw(
                  sstmac::ssterror,
                  "cray_dragon::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                  uid, them->to_string().c_str());
            }

            std::vector<int> mycoords(3), theircoords(3);
            mycoords[0] = myx;
            mycoords[1] = myy;
            mycoords[2] = myg;

            theircoords[0] = myx;
            theircoords[1] = myy;
            theircoords[2] = gid;

            std::string link_name = get_link_name(mycoords, theircoords);

            switches[me]->connect(2, gid, it1->second, link_latency,
                bytes_per_cycle * red[2], opt_linkbw, payload_fraction,
                enable_multipath, ignore_congestion, arb_template->clone(), "network",
                link_name);

            if (outputgraph)
            {

              connected.push_back(them->unique_id());
              connected_red.push_back(red[2]);
            }
          }

        }

        if (outputgraph)
        {
          for (int k = 0; k < nps_; k++)
          {
            std::cout << (i * nps_ + k);
            for (int l = 0; l < connected.size(); l++)
            {
              for (int m = 0; m < nps_; m++)
              {
                std::cout << " " << (connected[l] * nps_ + m) << ":"
                    << connected_red[l];

              }
            }

            for (int m = 0; m < nps_; m++)
            {
              if (m != k)
              {
                std::cout << " " << (i * nps_ + m) << ":" << 100;
              }
            }
            std::cout << "\n";
          }

        }

      }

    }

  }
} //end of namespace sstmac


