/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/dragonflyrouter.h>
#include <math.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {

    using namespace std;

    SSTRegister("dragonfly | dfly", router, dragonflyrouter);

    RNG::Combo::ptr dragonflyrouter::rng_;
    int dragonflyrouter::global_routing_ = -1;

    void
    dragonflyrouter::route_group(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      int hisx, hisy, hisg;
      get_coords(netaddr->unique_id(), hisx, hisy, hisg);

      global_routing_info::ptr gto = boost::dynamic_pointer_cast<
          global_routing_info>(to);

      if (!gto)
      {
        sst_throw(ssterror, "cray_dragon_router::route_group - info didn't cast");
      }

      if (global_routing_ == VALIANT && gto->global_min())
      {

        std::vector<int> connected;
        int gspace = std::max(1, numG_ / group_con_);
        int myid = myX_ + myY_ * numX_;
        int gset = myid % numG_;

        for (int g = 0; g < group_con_; g++)
        {
          int gid = (gset + gspace * g) % numG_;
          if (gid != myG_)
            connected.push_back(gid);
        }
        hisg = connected[rng_->value() % connected.size()];
        dim = 2;
        dir = hisg;
        gto->set_global_min(false);
        SSTMAC_DEBUG << "routing globally valiant to group " << dir << "\n";
        return;
      }

      bool firsty = true;
      std::vector<std::pair<int, int> > valid;
      std::vector<std::pair<int, int> > twohops;

      for (int y = myY_; firsty || y != myY_; y = (y + 1) % numY_)
      {
        firsty = false;
        bool firstx = true;
        for (int x = myX_; firstx || x != myX_; x = (x + 1) % numX_)
        {
          firstx = false;
          int gspace = std::max(1, numG_ / group_con_);
          int myid = x + y * numX_;
          int gset = myid % numG_;

          for (int g = 0; g < group_con_; g++)
          {

            int gid = (gset + gspace * g) % numG_;
            if (gid == hisg)
            {
              if (x == myX_ && y == myY_)
              {
                //std::pair<int, int> p = make_pair<int, int> (2, gid);
                //valid.push_back(p);
                dim = 2;
                dir = gid;
                SSTMAC_DEBUG << "going to group " << hisg
                      << " which is connected to me \n";
                return;
              }
              else if (y != myY_ && x == myX_)
              {
                std::pair<int, int> p = make_pair<int, int> (1, y);
                valid.push_back(p);
                // dim = 1;
                // dir = y;
                //SSTMAC_DEBUG << "going to group " << hisg
                //      << " which is at y pos " << y << " \n";
              }
              else if (x != myX_ && y == myY_)
              {
                std::pair<int, int> p = make_pair<int, int> (0, x);
                valid.push_back(p);
                // dim = 0;
                // dir = x;
                //SSTMAC_DEBUG << "going to group " << hisg
                //      << " which is at x pos " << x << " \n";
              }
              else
              {
                std::pair<int, int> p = make_pair<int, int> (0, x);
                twohops.push_back(p);
              }
              // return;
            }

          }

        }

      }

      if (valid.size() == 0 && twohops.size() == 0)
      {
        sst_throw(ssterror, "dragonrouter: no valid connections to group");
      }
      if (valid.size() > 0)
      {
        int ind = rng_->value() % valid.size();
        dim = valid[ind].first;
        dir = valid[ind].second;

      }
      else
      {
        int ind = rng_->value() % twohops.size();
        dim = twohops[ind].first;
        dir = twohops[ind].second;
      }

      SSTMAC_DEBUG << "going to group " << hisg << " which is at dim " << dim
            << " and dir " << dir << " \n";

      //if we get here, we didn't find a group connection here
      //sst_throw(ssterror, "cray_dragon_router - didn't find a connection to group %i", hisg);
      //dim = 0;
      //dir = (myX_ + 1) % numX_;
    }

    void
    dragonflyrouter::route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      int hisx, hisy, hisg;
      get_coords(netaddr->unique_id(), hisx, hisy, hisg);
      SSTMAC_DEBUG << "cray_dragon_router(" << me_->to_string()
            << "): heading to " << to->to()->to_string() << " at net addr "
            << netaddr->to_string() << "...";

      if (hisg != myG_)
      { //route to the right group first
        route_group(to, dim, dir);

      }
      else if (hisx != myX_)
      {
        dim = 0;
        dir = hisx;
        SSTMAC_DEBUG << "in group, headed to x pos " << hisx << "\n";
      }
      else if (hisy != myY_)
      {
        dim = 1;
        dir = hisy;
        SSTMAC_DEBUG << "in group, headed to y pos " << hisy << "\n";
      }
      else
      {
        dim = networkswitch::EJECT_PORT;
        dir = to->to()->unique_id() % conc_;
        SSTMAC_DEBUG << "i'm here, ejecting \n";
      }
    }

    int
    dragonflyrouter::convert_to_outport(int dim, int dir)
    {
      if (dim == networkswitch::EJECT_PORT)
      {
        return -(1 + dir);
      }
      if (dim == 0)
      {
        return dir;
      }
      else if (dim == 1)
      {
        return numX_ + dir;
      }
      else
        return numX_ + numY_ + dir;
    }

    void
    dragonflyrouter::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
      {
        dim = networkswitch::EJECT_PORT;
        dir = (-outport) - 1;
      }
      else
      {
        if (outport < numX_)
        {
          dim = 0;
          dir = outport;
        }
        else if (outport >= numX_ && outport < numX_ + numY_)
        {
          dim = 1;
          dir = outport - numX_;
        }
        else
        {
          dim = 2;
          dir = outport - (numX_ + numY_);
        }

      }
    }

    int
    dragonflyrouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(from);
      int hisx, hisy, hisg;
      get_coords(netaddr->unique_id(), hisx, hisy, hisg);
      int thedir;
      if (dim == 0)
      {
        thedir = hisx;
      }
      else if (dim == 1)
      {
        thedir = hisy;
      }
      else
      {
        thedir = hisg;
      }
      return convert_to_outport(dim, thedir);

    }
  }

}

