/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_ROUTING_DRAGONFLYROUTER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_ROUTING_DRAGONFLYROUTER_H_INCLUDED

#include <sstmac/hardware/network/topology/router.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/hardware/network/topology/dragonfly.h>

namespace sstmac
{
  namespace hw
  {

    class global_routing_info : public sstmac::hw::routing_info
    {
    protected:
      global_routing_info(sstmac::nodeaddress::ptr to) :
        routing_info(to), global_min_(true)
      {

      }

      bool global_min_;
    public:
      typedef boost::intrusive_ptr<global_routing_info> ptr;

      virtual
      ~global_routing_info()
      {
      }

      virtual std::string
      to_string() const
      {
        return "global_routing_info";
      }

      void
      set_global_min(bool b)
      {
        global_min_ = b;
      }

      bool
      global_min() const
      {
        return global_min_;
      }

      virtual routing_info::ptr
      clone(sstmac::nodeaddress::ptr to)
      {
        return ptr(new global_routing_info(to));
      }

      static ptr
      make_new()
      {
        return ptr(new global_routing_info(sstmac::nodeaddress::ptr()));
      }

    };

    class dragonflyrouter : public router
    {

    protected:

    public:
      typedef boost::intrusive_ptr<dragonflyrouter> ptr;

      static ptr
      construct(sim_parameters::ptr params)
      {
        return new dragonflyrouter;
      }

      virtual void
      init(const nodeaddress::ptr& me, const topology::ptr& top)
      {
        dragonfly::ptr dfly = safe_cast<dragonfly> (top);
        router::init(dfly->nps(), me);
        numX_ = dfly->numX();
        numY_ = dfly->numY();
        numG_ = dfly->numG();
        group_con_ = dfly->group_con();
        get_coords(me->unique_id(), myX_, myY_, myG_);
      }

      virtual void
      route(routing_info::ptr to, int &dim, int &dir);

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      virtual int
      get_inport(nodeaddress::ptr from, int dim, int dir);

    protected:

      void
      get_coords(long uid, int &x, int &y, int &g)
      {
        x = uid % numX_;
        y = (uid / numX_) % numY_;
        g = uid / (numX_ * numY_);
      }

      int numX_;
      int numY_;
      int numG_;

      int myX_;
      int myY_;
      int myG_;

      int group_con_;

      enum GLBL_ROUTE_TYPE { MINIMAL = 0, VALIANT} ;

      static int global_routing_;
      static RNG::Combo::ptr rng_;

    protected:
      dragonflyrouter()
       {
         if (!rng_)
         {
           long seed = time(NULL);
           if (sstmac::sstmac_env::params->has_param("router_seed"))
             seed = sstmac::sstmac_env::params->get_long_param("router_seed");

           std::vector<RNG::rngint_t> seeds;
           seeds.push_back(seed);
           rng_ = RNG::Combo::construct(seeds);

         }

         std::string glbl_name = sstmac::sstmac_env::params->get_optional_param(
             "dragon_global_routing", "minimal");

         if(global_routing_ < 0){ //initialize it
           if(glbl_name == "minimal"){
             global_routing_ = MINIMAL;
           }else if(glbl_name == "valiant"){
             global_routing_ = VALIANT;
           }
         }

         if (!template_init_)
         {
           sstmac::hw::routing_info::set_template(global_routing_info::make_new());
           template_init_ = true;
         }
       }


       virtual void
       route_group(sstmac::hw::routing_info::ptr to, int &dim, int &dir);


    };

  }
}
#endif
