/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// fattree.cc: Implementation of fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/fattree.h>
#include <sstmac/hardware/network/topology/fattreerouter.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fattree | ftree", topology, fattree);

    fattree::ptr
    fattree::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 0);
      if (args.size() == 2)
      {
        return fattree::ptr(new fattree(args[0], args[1], nps, 1));
      }
      else if (args.size() == 3)
      {
        return fattree::ptr(new fattree(args[0], args[1], args[2], 1));
      }
      else if (args.size() == 4)
      {
        return fattree::ptr(new fattree(args[0], args[1], args[2], args[3]));
      }
      else
      {
        throw ssterror("fattree::construct -- give either 2 or 3 parameters");
      }
    }

    std::string
    fattree::label(long id)
    {
        return sstprintf("fat tree %ld", id);
    }

    static bool
    incr_vector(int l, int k, std::vector<int>& v)
    {
      for (int i = l - 1; i >= 0; i--)
      {
        v[i]++;
        if (v[i] == k)
        {
          v[i] = 0;
        }
        else
          return true;
      }
      return false;
    }

    static bool
    incr_vector2(int l, int ki, int kl, std::vector<int>& v)
    {
      int i = l - 1;
      v[i]++;
      if (v[i] == kl)
      {
        v[i] = 0;
      }
      else
        return true;
      i--;
      for (; i >= 0; i--)
      {
        v[i]++;
        if (v[i] == ki)
        {
          v[i] = 0;
        }
        else
          return true;
      }
      return false;
    }

    static void
    compute_nodes(int l, int k, std::set<vertex> &vertices, int nndpersw)
    {
      vertices.clear();

      // The node vertices are all possible l-digit numbers.
      // The first l-1 digits are base k.
      // The last digit is base nndpersw.
      vertex v(l);
      fill(v.v().begin(), v.v().end(), 0);
      do
      {
        vertices.insert(v);
        //std::cerr << "adding node vertex ";
        //for(int p=0; p<v.v().size(); ++p)
        //  std::cerr << v[p] << " ";
        //std::cerr << std::endl;
      }
      while (incr_vector2(l, k, nndpersw, v.v()));
    }

    fattree::fattree(int l, int k, int nndpersw, int widtharg)
        : switchtopology(nndpersw)
    {
      init_fattree(l, k, nndpersw, widtharg);
    }

    static void
    compute_switches(int l, int k, std::set<vertex> &vertices,
        std::vector<std::set<vertex> > &tiers)
    {
      vertices.clear();

      // Switch vertices have an extra digit with the tier and
      // the last significant digit is zero.

      for (int i = 1; i <= l; i++)
      {
        tiers[i].clear();
        std::vector<int> v(l - 1);
        fill(v.begin(), v.end(), 0);
        do
        {
          vertex vert;
          vert.v().resize(l);
          // switches are distigushed from nodes by the have a negative
          // coordinate (the tier number)
          vert.v()[0] = -i;
          for (int j = 1; j < l; j++)
            vert.v()[j] = v[j - 1];
          // For the top tier, the next to last significant digit rounded
          // down to the nearest even number, since the up links are
          // used as down links.
          if (i == l)
            vert.v()[l - 1] &= ~1;
          vertices.insert(vert);
          tiers[i].insert(vert);
          //std::cerr << "adding switch vertex ";
          //for(int p=0; p<vert.v().size(); ++p)
          //  std::cerr << vert[p] << " ";
          //std::cerr << std::endl;
        }
        while (incr_vector(l - 1, k, v));
      }
    }

    static void
    compute_edges(int l, int k, std::vector<std::set<vertex> > &tiers,
                  int nndpersw)
    {
      // Compute the edges between the nodes and tier 1.
      vertex sv(l); // sv is a switch index
      sv[0] = -1;
      // loop through the nodes (tiers[0])
      for (std::set<vertex>::iterator i = tiers[0].begin();
           i != tiers[0].end(); i++)
      {
        // Copy the nodes indices into the switch index.  The last node
        // index is ignored: it enumerates the nodes on a single switch.
        for (uint j = 0; j < i->v().size() - 1; j++)
        {
          sv.v()[j + 1] = i->v()[j];
        }
        //std::cerr << "adding edge from ";
        //for(int p=0; p<sv.v().size(); ++p)
        //  std::cerr << sv[p] << " ";
        //std::cerr << "to ";
        //for(int p=0; p<sv.v().size(); ++p)
        //  std::cerr << (*i)[p] << " ";
        //std::cerr << std::endl;

        //std::cerr << "adding edge from ";
        //for(int p=0; p<sv.v().size(); ++p)
        //  std::cerr << (*i)[p] << " ";
        //std::cerr << "to ";
        //for(int p=0; p<sv.v().size(); ++p)
        //  std::cerr << sv[p] << " ";
        //std::cerr << std::endl;
      }

      // Compute the remaining edges.
      for (int il = 1; il < l; il++)
      {
        for (std::set<vertex>::iterator iv = tiers[il].begin(); iv
            != tiers[il].end(); iv++)
        {
          for (int ik = 0; ik < k; ik++)
          {
            vertex up(*iv);
            up.v()[0]--;
            up.v()[l - il] = ik;
            int edge_id = 0;
            if (il + 1 == l)
            {
              // On the top tier, the next to last digit is even.
              // Edges that are folded into a different switch
              // are given an id of 1.
              if (up.v()[l - 1] & 1)
                edge_id = 1;
              up.v()[l - 1] &= ~1;
            }

            //std::cerr << "adding edge from ";
            //for(int p=0; p<up.v().size(); ++p)
            //  std::cerr << (*iv)[p] << " ";
            //std::cerr << "to ";
            //for(int p=0; p<up.v().size(); ++p)
            //  std::cerr << up[p] << " ";
            //std::cerr << std::endl;

            //std::cerr << "adding edge from ";
            //for(int p=0; p<up.v().size(); ++p)
            //  std::cerr << up[p] << " ";
            //std::cerr << "to ";
            //for(int p=0; p<up.v().size(); ++p)
            //  std::cerr << (*iv)[p] << " ";
            //std::cerr << std::endl;
          }
        }
      }
    }

    void
    fattree::init_fattree(int l, int k, int nndpersw, int widtharg)
    {
      width = widtharg;
      l_ = l;
      k_ = k;

      if (l < 2)
        throw std::runtime_error("fattree: must have l > 1");
      if (k & 1)
        throw std::runtime_error("fattree: must have k even");

      std::vector<std::set<vertex> > tiers(l + 1);

      if (nndpersw == 0)
        nndpersw = k;

      compute_nodes(l, k, nodes_, nndpersw);
      tiers[0] = nodes_;

      //std::set<vertex> switches;
      //compute_switches(l, k, switches, tiers);

      //compute_edges(l, k, tiers, nndpersw);

      //std::set<vertex> all_vertices;
      //all_vertices.insert(nodes_.begin(), nodes_.end());
      init();
    }

    size_t
    fattree::size(int l, int k)
    {
      size_t r = 1;
      for (int i = 0; i < l; i++)
      {
        r *= k;
      }
      return r;
    }

    int
    fattree::l() const
    {
      return l_;
    }

    int
    fattree::k() const
    {
      return k_;
    }

    int64_t
    fattree::diameter() const
    {
      return (l() + 1) * 2;
    }

    std::string
    fattree::name() const
    {
      std::ostringstream ostr;
      ostr << "FatTree(" << l_ << "," << k_ << ")";
      return ostr.str();
    }

    /// ports 0->(k-1) of a switch connect down, ports k->(2k-1) connect up
    void
    fattree::connect_switches(switchtopology::maptype &switches,
                              const std::vector<int>& red,
                              double linkbw)
    {
      long edgeswitches = switches.size();
      long swid = edgeswitches;
      switchtopology::maptype prevlevel = switches;
      long thislevelstartid;
      long prevlevelstartid = 0;
      long crossfactor = 1;
      for (int l = 1; l < l_; l++)
      {
        thislevelstartid = swid;
        switchtopology::maptype thislevel;

        int portstofill = prevlevel.size() * k_;
        int portsdown;

        if (l == l_ - 1)
          portsdown = 2 * k_;
        else
          portsdown = k_;

        int numswitches = portstofill / portsdown;
        if(l==1) numleafswitches_ = numswitches;

        SSTMAC_DEBUG << "fattree: creating " << numswitches << " for level "
              << l << "\n";
        networkswitch::ptr templ = switches.begin()->second;
        for (int i = 0; i < numswitches; i++)
        {
          nodeaddress::ptr swaddr = nodeaddress::convert_from_uniqueid(swid);
          networkswitch::ptr sw = templ->clone(create_router(swaddr, sstmac_env::params),
              swaddr);
          thislevel[swaddr] = sw;
          switches[swaddr] = sw;
          if(l==1) leafswitches_[swaddr] = sw;
          swid++;
        }

        long thisid = thislevelstartid;
        long logicalid = 0;
        long groupsize = pow(k_, l);
        long thisgroupsize = (l == l_ - 1) ? pow(k_, l - 1) : pow(k_, l);
        long upgroupstart = 0;
        if (l == l_ - 1 && l != 1)
        {
          crossfactor /= 2;
        }
        for (long previd = prevlevelstartid; previd < prevlevelstartid
            + (long) prevlevel.size(); previd++)
        {
          nodeaddress::ptr prevaddr =
              nodeaddress::convert_from_uniqueid(previd);
          //track to the right switch in this level


          int div = logicalid / groupsize;
          upgroupstart = thislevelstartid + div * groupsize;
          thisid = upgroupstart + (logicalid) % thisgroupsize;
          int connectionsperswitch = (l == l_ - 1) ? std::max(1,
              (k_ / numswitches)) : 1;
          int portdiv = (l == l_ - 1) ? std::max(1, (numswitches / k_)) : pow(
              k_, l - 1);
          int portmod = (l == l_ - 1) ? groupsize * connectionsperswitch : k_;
          int downport = (logicalid / portdiv) * connectionsperswitch
              % (portmod);

          for (int p = 0; p < k_; p++)
          {
            nodeaddress::ptr thisaddr = nodeaddress::convert_from_uniqueid(
                thisid);

            SSTMAC_DEBUG << "connecting up from " << prevaddr->to_string()
                  << " to " << thisaddr->to_string() << " on up port " << p
                  << " and down port " << downport << "\n";
            if (!prevlevel[prevaddr])
            {
              sst_throw(ssterror,
                  "fattree:: can't find previous level switch");
            }
            if (!thislevel[thisaddr])
            {
              sst_throw(ssterror,
                  "fattree:: can't find this level switch");
            }
            prevlevel[prevaddr]->connect(0, p, thislevel[thisaddr], linkbw); // up
            thislevel[thisaddr]->connect(1, downport, prevlevel[prevaddr],
                linkbw); //down

            prevlevel[prevaddr]->get_router()->set_inport(thisaddr, p);
            thislevel[thisaddr]->get_router()->set_inport(prevaddr,
                k_ + downport);

            if (thisid + crossfactor >= upgroupstart + groupsize)
            {
              thisid = upgroupstart + ((thisid + crossfactor) % groupsize);

            }
            else if (thisid + crossfactor - thislevelstartid
                >= (long) thislevel.size())
            {
              thisid = thislevelstartid + (thisid + crossfactor)
                  % thislevel.size();

            }
            else
            {
              thisid += crossfactor;

            }

            if (p == (numswitches / portdiv) - 1)
            {
              downport++;
            }

          }
          logicalid++;

        }

        prevlevel = thislevel;
        prevlevelstartid = thislevelstartid;
        crossfactor *= k_;
      }

    }

    void
    fattree::connect_train_tracks(trainmap& switches,
                                  const std::vector<int>& red,
                                  double bytes_per_cycle,
                                  double freq_hz,
                                  timestamp link_latency,
                                  double payload_fraction,
                                  bool ignore_congestion,
                                  bool enable_multipath,
                                  const output_bw_arbitrator::ptr& arb_template)
    {

      SSTMAC_DEBUG <<
            "fattree: send " << bytes_per_cycle << " bytes at freq " << freq_hz <<
            " hz for bw " << (bytes_per_cycle * freq_hz) << " GB/sec\n";

      if(red.size() != l_){
        sst_throw(ssterror, "fattree::connect_train_tracks: redundant vector doesn't match levels");
      }

      long edgeswitches = switches.size();
      long swid = edgeswitches;
      switchtopology::trainmap prevlevel = switches;
      long thislevelstartid;
      long prevlevelstartid = 0;
      long crossfactor = 1;
      for (int l = 1; l < l_; l++)
      {
        thislevelstartid = swid;
        switchtopology::trainmap thislevel;

        int portstofill = prevlevel.size() * k_;
        int portsdown;

        if (l == l_ - 1)
          portsdown = 2 * k_;
        else
          portsdown = k_;

        int numswitches = portstofill / portsdown;

        SSTMAC_DEBUG << "fattree: creating " << numswitches << " for level "
              << l << "\n";
        trainswitch::ptr templ = switches.begin()->second;
        for (int i = 0; i < numswitches; i++)
        {
          nodeaddress::ptr swaddr = nodeaddress::convert_from_uniqueid(swid);
          trainswitch::ptr sw = templ->clone(create_router(swaddr, sstmac_env::params),
              swaddr);
          thislevel[swaddr] = sw;
          switches[swaddr] = sw;
          swid++;
        }

        long thisid = thislevelstartid;
        long logicalid = 0;
        long groupsize = pow(k_, l);
        long thisgroupsize = (l == l_ - 1) ? pow(k_, l - 1) : pow(k_, l);
        long upgroupstart = 0;
        if (l == l_ - 1 && l != 1)
        {
          crossfactor /= 2;
        }
        for (long previd = prevlevelstartid; previd < prevlevelstartid
            + (long) prevlevel.size(); previd++)
        {
          nodeaddress::ptr prevaddr =
              nodeaddress::convert_from_uniqueid(previd);
          //track to the right switch in this level


          int div = logicalid / groupsize;
          upgroupstart = thislevelstartid + div * groupsize;
          thisid = upgroupstart + (logicalid) % thisgroupsize;
          int connectionsperswitch = (l == l_ - 1) ? std::max(1,
              (k_ / numswitches)) : 1;
          int portdiv = (l == l_ - 1) ? std::max(1, (numswitches / k_)) : pow(
              k_, l - 1);
          int portmod = (l == l_ - 1) ? groupsize * connectionsperswitch : k_;
          int downport = (logicalid / portdiv) * connectionsperswitch
              % (portmod);

          for (int p = 0; p < k_; p++)
          {
            nodeaddress::ptr thisaddr = nodeaddress::convert_from_uniqueid(
                thisid);

            SSTMAC_DEBUG << "connecting up from " << prevaddr->to_string()
                  << " to " << thisaddr->to_string() << " on up port " << p
                  << " and down port " << downport << "\n";
            if (!prevlevel[prevaddr])
            {
              sst_throw(ssterror,
                  "fattree:: can't find previous level switch");
            }
            if (!thislevel[thisaddr])
            {
              sst_throw(ssterror,
                  "fattree:: can't find this level switch");
            }
            prevlevel[prevaddr]->connect(0, p, thislevel[thisaddr],
                link_latency, bytes_per_cycle * red[l], freq_hz,
                payload_fraction, enable_multipath, ignore_congestion,
                 arb_template->clone(), "network"); // up


            thislevel[thisaddr]->connect(1, downport, prevlevel[prevaddr],
                link_latency, bytes_per_cycle * red[l], freq_hz,
                payload_fraction, enable_multipath, ignore_congestion,
                arb_template->clone(), "network"); //down

            prevlevel[prevaddr]->get_router()->set_inport(thisaddr, p);
            thislevel[thisaddr]->get_router()->set_inport(prevaddr,
                k_ + downport);

            if (thisid + crossfactor >= upgroupstart + groupsize)
            {
              thisid = upgroupstart + ((thisid + crossfactor) % groupsize);

            }
            else if (thisid + crossfactor - thislevelstartid
                >= (long) thislevel.size())
            {
              thisid = thislevelstartid + (thisid + crossfactor)
                  % thislevel.size();

            }
            else
            {
              thisid += crossfactor;

            }

            if (p == (numswitches / portdiv) - 1)
            {
              downport++;
            }

          }
          logicalid++;

        }

        prevlevel = thislevel;
        prevlevelstartid = thislevelstartid;
        crossfactor *= k_;
      }
    }

    void
    fattree::set_ejectors(long num, int nps, const maptype& switches,
                          ejmaptype &ejectors, 
                          const sim_parameters::ptr& params)
    {
      for (long i = 0; i < numleafswitches_ * nps; i++)
      {

        nodeaddress::ptr nodeaddr = 
          nodeaddress::convert_from_uniqueid(i);
        nodeaddress::ptr swaddr = 
          nodeaddress::convert_from_uniqueid((i / nps));

        ejectors[nodeaddr] = mapget(switches,swaddr,"switch address")
          ->get_ejector(params, i % nps);       
      }
    }

  }
} //end of namespace sstmac
