/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// fattree.h: Interface for fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED

#include <sstmac/hardware/network/topology/switchtopology.h>

namespace sstmac
{
  namespace hw
  {

    /** The fattree network generates a k-ary fat tree with l tiers.  In the fat
     tree, not all vertices are nodes. The inputs for this network
     can specify the following attributes:

     <table border="1">

     <tr><td>Keyword<td>Type<td>Default<td>Description

     <tr><td><tt>l</tt><td>int<td>2<td>The number of levels in the fat tree.

     <tr><td><tt>k</tt><td>int<td>2<td>Half the number of links a switch
     element has.

     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.

     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.

     </table>
     */

    class fattree : public switchtopology
    {
      int width;
    private:
      int l_, k_, numleafswitches_;
      switchtopology::maptype leafswitches_;
      void
      init_fattree(int l, int k, int nndpersw, int widtharg);

    protected:
      //fattree(TiXmlHandle);
      fattree(int l, int k, int nndpersw, int widtharg);

    public:
      typedef boost::intrusive_ptr<fattree> ptr;

      static fattree::ptr
      construct(sim_parameters::ptr params);

      static ptr
      construct(int l, int k, int nps = 1, int wid = 1)
      {
        return ptr(new fattree(l, k, nps, wid));
      }

      virtual
      ~fattree()
      {
      }

      static size_t
      size(int l, int k);
      int
      l() const;

      int
      k() const;

      int64_t
      diameter() const;

      std::string
      name() const;

      virtual int
      ndimensions() const {
          return l_;
      }

      virtual void
      connect_train_tracks(trainmap& switches,
                           const std::vector<int>& red,
                           double bytes_per_cycle,
                           double freq_hz,
                           timestamp link_latency,
                           double payload_fraction,
                           bool ignore_congestion,
                           bool enable_multipath,
                           const output_bw_arbitrator::ptr& arb_template);

      virtual void
      connect_switches(switchtopology::maptype &switches,
                       const std::vector<int>& red,
                       double linkbw);

      virtual std::string label(long id);

      virtual int
      radix() const
      {
        return 2 * k_;
      }

      virtual int64_t
      topology_size() const
      {
        return pow(k_, l_) / k_;
      }

      virtual std::string
      to_string() const
      {
        return "fattreetopology";
      }

      virtual void
      set_ejectors(long num, int nps, const maptype& switches,
                   ejmaptype &ejectors, const sim_parameters::ptr& params);
    };

  }
} //end of namespace sstmac

#endif
