/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/fattreerouter.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {

    SSTRegister("fattree | ftree", router, fattreerouter);

    void
    fattreerouter::route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());

      SSTMAC_DEBUG << "fattreerouter(" << me_->to_string() << "): heading to "
            << to->to()->to_string() << " at net addr " << netaddr->to_string()
            << "\n";

      if (netaddr->equals(me_))
        {
          dim = networkswitch::EJECT_PORT;
          dir = to->to()->unique_id() % conc_;
          SSTMAC_DEBUG << "ejecting \n";
          return;
        }
      else if (myL_ == 0)
        { // if we're on the first level and we're not there yet, go up
          dim = 0;
          dir = rand() % k_;
          SSTMAC_DEBUG << "routing up from first level \n";
        }
      else
        {
          int numunder = pow(k_, myL_ + 1);

          if (to->to()->unique_id() >= (logicalid_ * numunder) && to->to()->unique_id()
              < ((logicalid_ + 1) * numunder)) // we're routing down
            {

              dim = 1;
              if (myL_ == l_ - 1)
                {
                  int switchesperlevel = pow(k_, l_ - 1) / 2;
                  int connectionsperswitch = std::max(1,
                      (k_ / switchesperlevel));

                  int divsize = std::max(1, (switchesperlevel * 2) / (k_ * 2));
                  dir = (rand() % connectionsperswitch) + connectionsperswitch
                      * (netaddr->unique_id() / divsize);
                }
              else
                dir = (netaddr->unique_id() / (int) pow(k_, myL_ - 1)) % k_;

              SSTMAC_DEBUG << "routing down with dir " << dir << "\n";
            }
          else
            { //route up
              dim = 0;
              dir = rand() % k_;
              SSTMAC_DEBUG << "routing random up with dir " << dir << "\n";
            }
        }

    }

    int
    fattreerouter::convert_to_outport(int dim, int dir)
    {
      if (dim == networkswitch::EJECT_PORT)
        {
          return -(1+dir);
        }
      return dim * k_ + dir;
    }

    void
    fattreerouter::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = networkswitch::EJECT_PORT;
          dir = (-outport) - 1;
        }
      else
        {
          if (myL_ == l_ - 1)
            {
              dim = 1;
              dir = (outport-k_) % (2 * k_);
            }
          else
            {
              dim = outport / k_;
              dir = outport % k_;
            }
        }
    }

    void
    fattreerouter::set_inport(nodeaddress::ptr from, int port)
    {
      inports_[from->unique_id()] = port;
    }

    int
    fattreerouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {
      return inports_[from->unique_id()];
    }

  }
}
