/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/fattreerouter_congestion.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fattree_congestion | ft_congestion", router, fattreerouter_congestion) ;

    void
    fattreerouter_congestion::route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());

      SSTMAC_DEBUG << "ft_congestion_router(" << me_->to_string()
            << "): heading to " << to->to()->to_string() << " at net addr "
            << netaddr->to_string() << "...";

      if (netaddr->equals(me_))
      {
        dim = networkswitch::EJECT_PORT;
        dir = to->to()->unique_id() % conc_;
        SSTMAC_DEBUG << "ejecting \n";
        return;
      }
      else if (myL_ == 0)
      { // if we're on the first level and we're not there yet, go up
        //find the min
        SSTMAC_DEBUG << "routing up from first level ...";
        int d, r;
        int bestd = -1, bestr = -1;
        double min_cong = 1000000;
        boost::unordered_map<int, double>::iterator it, end = cong_.end();
        for (it = cong_.begin(); it != end; it++)
        {
          convert_to_dimdir(it->first, d, r);
          if (d == 0) // up only
          {
            SSTMAC_DEBUG2 << "congestion on port " << it->first << " is "
                  << it->second << "\n";
            if (it->second < min_cong)
            {
              min_cong = it->second;
              bestd = d;
              bestr = r;
            }
          }
        }

        if (bestd < 0)
        {
          //sst_throw(ssterror, "ft_congestion_router::route() - something weird happened while trying to find minimum congestion");
          dim = 0;
          dir = rand() % k_;

          SSTMAC_DEBUG << " using backup for some reason \n ";
        }
        else
        {
          dim = bestd;
          dir = bestr;
          SSTMAC_DEBUG << " using minimum congestion \n";
        }

      }
      else
      {
        int numunder = pow(k_, myL_ + 1);

        if (to->to()->unique_id() >= (logicalid_ * numunder)
            && to->to()->unique_id() < ((logicalid_ + 1) * numunder)) // we're routing down
        {

          dim = 1;
          if (myL_ == l_ - 1)
          {
            int switchesperlevel = pow(k_, l_ - 1) / 2;
            int connectionsperswitch = std::max(1, (k_ / switchesperlevel));

            int divsize = std::max(1, (switchesperlevel * 2) / (k_ * 2));
            dir = (rand() % connectionsperswitch) + connectionsperswitch
                * (netaddr->unique_id() / divsize);
          }
          else
            dir = (netaddr->unique_id() / (int) pow(k_, myL_ - 1)) % k_;

          SSTMAC_DEBUG << "routing down with dir " << dir << "\n";
        }
        else
        { //route up
          SSTMAC_DEBUG << "in the middle routing up ...";
          int d, r;
          int bestd = -1, bestr = -1;
          double min_cong = 1000000;
          boost::unordered_map<int, double>::iterator it, end = cong_.end();
          for (it = cong_.begin(); it != end; it++)
          {
            convert_to_dimdir(it->first, d, r);
            if (d == 0) // up only
            {
              if (it->second < min_cong)
              {
                min_cong = it->second;
                bestd = d;
                bestr = r;
              }
            }
          }

          if (bestd < 0)
          {
            // sst_throw(ssterror, "ft_congestion_router::route() - something weird happened while trying to find minimum congestion");
            dim = 0;
            dir = rand() % k_;
            SSTMAC_DEBUG << " using backup for some reason \n ";
          }
          else
          {

            dim = bestd;
            dir = bestr;
            SSTMAC_DEBUG << " using minimum congestion \n";
          }

        }
      }

    }

  }
}
