/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// flattenedbutterfly.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov> 
// 
#include <sstmac/hardware/network/topology/flattenedbutterfly.h>
#include <sstmac/hardware/network/topology/flattenedbflyrouter.h> 
#include <stdio.h> 
#include <sstream>
#include <sstmac/common/driver_util.h>
#include <sstmac/common/factories/factory.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fbfly | flattenedbfly | flattenedbutterfly", topology, flattenedbutterfly);

    flattenedbutterfly::ptr
    flattenedbutterfly::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);

      return ptr(new flattenedbutterfly(args, nps));

    }

    // --------------------------------------------------------------------------------------------------------

    flattenedbutterfly::flattenedbutterfly(const std::vector<int>& dimensionarray,
        int nndpersw)
        : switchtopology(nndpersw)
    {
      dimensions_ = dimensionarray;
      nps_ = nndpersw;
    }

    // ----------------------------------------------------------------------------------


    std::vector<int>
    flattenedbutterfly::get_coords(long uid)
    {
      long div = 1;
      std::vector<int> ret(dimensions_.size());
      for (int i = 0; i < (int) dimensions_.size(); i++)
        {
          ret[i] = (uid / div) % dimensions_[i];
          div *= dimensions_[i];
        }
      return ret;
    }

    long
    flattenedbutterfly::get_netid(const std::vector<int>& coords)
    {
      long ret = 0;
      long mult = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
        {
          ret += coords[i] * mult;
          mult *= dimensions_[i];
        }
      return ret;
    }

    std::string
    flattenedbutterfly::coord_string(const std::vector<int>& coords)
    {
      std::stringstream ss;
      ss << "(";
      for (int i = 0; i < (int) dimensions_.size(); i++)
        {
          ss << coords[i] << ",";
        }
      ss << ")";
      return ss.str();
    }

    void
    flattenedbutterfly::connect_switches(switchtopology::maptype &switches,
                                         const std::vector<int>& red,
                                         double linkbw)
    {
      SSTMAC_DEBUG << "flattenedbutterfly: link bandwidth is " << linkbw
            << "\n";

      for (long i = 0; i < (int) switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
          std::vector<int> coords = get_coords(i);

          for (int z = 0; z < (int) dimensions_.size(); z++)
            {
              for (int d = 0; d < (int) dimensions_[z]; d++)
                {
                  std::vector<int> connect = coords;
                  if (coords[z] != d)
                    {
                      connect[z] = d;

                      long uid = get_netid(connect);

                      nodeaddress::ptr addr =
                          nodeaddress::convert_from_uniqueid(uid);

                      if (switches.find(addr) == switches.end())
                        {
                          std::cerr << "uid: " << uid << ", addr: "
                              << addr->to_string() << "\n";
                          throw ssterror(
                              "flattenedbutterfly::connect_switches() -- can't find switch when connecting");
                        }

                      SSTMAC_DEBUG << "switch " << me->to_string()
                            << " with coords " << coord_string(coords)
                            << " for dim " << z
                            << " connecting in the pos dir to "
                            << coord_string(connect) << " whic is addr "
                            << addr->to_string() << "\n";

                      switches[me]->connect(z, d, switches[addr], linkbw,
                          red[z]);
                    }
                }

            }
        }

    }

    int64_t
    flattenedbutterfly::topology_size() const
    {
      int64_t ret = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
        {
          ret *= dimensions_[i];
        }
      return ret;
    }

    void
    flattenedbutterfly::display_nodes(vismaptype &switches,
        const vis::vis_engine::ptr& eng, std::list<vis::vis_obj::ptr> &objs)
    {
      /* for (long i = 0; i < switches.size(); i++)
       {
       nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

       std::vector<int> coords = get_coords(i);

       double x = 0, y = 0, z = 0;
       x = coords[0] * 2;
       if (coords.size() > 1)
       y = coords[1] * 2;

       if (coords.size() > 2)
       z = coords[2] * 2;

       vis::vis_obj::ptr vo = eng->create_cube(x, y, z, 1);
       switches[me]->set_vis_obj(vo);
       switches[me]->set_vis_engine(eng);
       objs.push_back(vo);

       for (int z = 0; z < dimensions_.size(); z++)
       {
       double x1 = coords[0] * 2;
       double y1 = coords[1] * 2;
       double z1 = coords[2] * 2;

       double x2 = coords[0] * 2;
       double y2 = coords[1] * 2;
       double z2 = coords[2] * 2;
       int dir;

       bool onegood = (coords[z] != dimensions_[z] - 1);
       bool twogood = (coords[z] != 0);

       if (z == 0)
       {

       x1 += 1.5;
       z1 += 0.5;
       y1 += 0.25;

       x2 -= 0.5;
       z2 += 0.5;
       y2 += 0.75;

       dir = vis::vis_engine::VIS_Z;
       }
       else if (z == 1)
       {
       y1 += 1.5;
       z1 += 0.25;
       x1 += 0.5;

       y2 -= 0.5;
       z2 += 0.75;
       x2 += 0.5;
       dir = vis::vis_engine::VIS_Y;
       }
       else
       {
       z1 += 1.5;
       y1 += 0.25;
       x1 += 0.5;

       z2 -= 0.5;
       y2 += 0.75;
       x2 += 0.5;
       dir = vis::vis_engine::VIS_X;
       }

       if (onegood)
       {
       vis::vis_obj::ptr c1 = eng->create_cylinder(x1, y1, z1,
       0.125, 1, dir);
       objs.push_back(c1);
       switches[me]->set_link_vis(c1, z, router::POS);
       }

       if (twogood)
       {
       vis::vis_obj::ptr c2 = eng->create_cylinder(x2, y2, z2,
       0.125, 1, dir);
       switches[me]->set_link_vis(c2, z, router::NEG);
       objs.push_back(c2);
       }
       }
       }*/
    }

  }
} //end of namespace sstmac


