/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
#include <sstmac/hardware/network/topology/grouped.h>
#include <math.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("grouped", topology, grouped);

    grouped::grouped(const switchtopology::ptr& inner, const switchtopology::ptr& outer,
        long g, int nps) :
      switchtopology(nps),
      inner_(inner), outer_(outer), g_(g), nps_(nps)
    {

    }

    grouped::ptr
    grouped::construct(const sim_parameters::ptr& params)
    {
      bool errorflag = false;
      std::vector<int> args;
      params->get_vector_param("topology_geometry", args);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);

      empty_parameters::ptr inpar = empty_parameters::construct();
      inpar->add_param("topology_name",
          params->get_param("inner_topology_name"));
      inpar->add_param("topology_geometry",
          params->get_param("inner_topology_geometry"));
      inpar->add_param("topology_redundant",
          params->get_param("inner_topology_redundant"));
      inpar->add_param("network_bandwidth_link",
          params->get_param("network_bandwidth_inner_link"));
      inpar->add_param("network_nodes_per_switch", params->get_optional_param("network_nodes_per_switch", "1"));

      topology::ptr in = topology_factory::get_param("inner_topology_name", inpar);
      switchtopology::ptr incast = boost::dynamic_pointer_cast<switchtopology>(in);


      empty_parameters::ptr outpar = empty_parameters::construct();
      outpar->add_param("topology_name",
          params->get_param("outer_topology_name"));
      outpar->add_param("topology_geometry",
          params->get_param("outer_topology_geometry"));
      outpar->add_param("topology_redundant",
          params->get_param("outer_topology_redundant"));
      outpar->add_param("network_bandwidth_link",
          params->get_param("network_bandwidth_outer_link"));
      outpar->add_param("network_nodes_per_switch", "1");

      topology::ptr out = topology_factory::get_param("outer_topology_name", outpar);
      switchtopology::ptr outcast = boost::dynamic_pointer_cast<switchtopology>(out);


      grouped::ptr ret = ptr(new grouped(incast, outcast, (long)args[0], nps));
      ret->inpar_ = inpar;
      ret->outpar_ = outpar;
      return ret;

    }

    void
    grouped::connect_switches(switchtopology::maptype &switches,
                              const std::vector<int>& red,
                              double linkbw)
    {
      std::vector<int> inred;
      if (inner_->ndimensions() != red.size()){
          inred.insert(inred.begin(), inner_->ndimensions(), 1);
      }
      else {
          inred = red;
      }
      std::vector<int> outred;
      if (outer_->ndimensions() != red.size()){
          outred.insert(inred.begin(), outer_->ndimensions(), 1);
      }
      else {
          outred = red;
      }

      switchtopology::maptype outsw;
      for (long g = 0; g < g_; g++)
        {

          switchtopology::maptype insw;
          for (long i = 0; i < inner_->n(); i++)
            {
              nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(
                  g * inner_->n() + i);
              insw[nodeaddress::convert_from_uniqueid(i)] = switches[me];
            }
          inner_->connect_switches(insw, inred, linkbw);

          superswitch::ptr su = superswitch::construct(insw);
          outsw[nodeaddress::convert_from_uniqueid(g)] = su;
        }

      outer_->connect_switches(outsw, outred, linkbw);

    }

    router::ptr
    grouped::create_router(const nodeaddress::ptr& addr, const sim_parameters::ptr& params)
    {

      router::ptr in = inner_->create_router(
          nodeaddress::convert_from_uniqueid(addr->unique_id() % inner_->n()),
          params);
      router::ptr out = outer_->create_router(
          nodeaddress::convert_from_uniqueid(addr->unique_id() / inner_->n()),
          params);
      router::ptr ret = groupedrouter::construct(nps_, in, out,
          inner_->n(), inner_->radix(), g_);

      ret->init(nps_, addr);

      return ret;
    }

    grouped::superswitch::superswitch(switchtopology::maptype sw) :
      mysw_(sw)
    {

    }

    injector::ptr
    grouped::superswitch::get_new_injector(sim_parameters::ptr params)
    {
        sst_throw(illformederror, "superswitch cannot returned injector");
        return injector::ptr();
    }

    ejector::ptr
    grouped::superswitch::get_new_ejector(sim_parameters::ptr params)
    {
        sst_throw(illformederror, "superswitch cannot returned ejector");
        return ejector::ptr();
    }

    void
    grouped::superswitch::connect(int dim, int dir, networkswitch::ptr sw, double bw, int redund)
    {
      nodeaddress::ptr ind = nodeaddress::convert_from_uniqueid(
          dir % mysw_.size());
      superswitch::ptr cast = boost::dynamic_pointer_cast<superswitch>(sw);
      if (!cast)
        {
          throw ssterror(
              "grouped topology :: superswitch - connecting switch didn't cast to superswitch");
        }
      mysw_[ind]->connect(dim + groupedrouter::GROUP_DIM, dir,
          cast->mysw_[ind], bw, redund);
    }

    networkswitch::ptr
    grouped::superswitch::clone(router::ptr route, nodeaddress::ptr addr)
    {
        sst_throw(unimplementederror, "superswitch::clone");
        return networkswitch::ptr();
    }

  }
} //end of namespace sstmac


