/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// hdtorus.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov> 
// 

#include <sstmac/hardware/network/topology/hdtorus.h>
#include <sstmac/hardware/network/topology/hdtorusrouter.h>
#include <stdio.h> 
#include <sstream>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("torus | hdtorus", topology, hdtorus,
                "hdtorus implements a high-dimension torus with an arbitrary number of dimensions");
    hdtorus::ptr
    hdtorus::construct(sim_parameters::ptr params)
    {
      std::vector<int> args;
      /**
        sstkeyword {
            gui=10 5 10;
            docstring=The number of network switches in each X,Y,Z dimension of the torus.
                      An arbitrary number of dimensions can be used.
                      The total number of switches is:ENDL
                      N_Switches=N_X*N_Y*...*N_Z.ENDL
                      The total number of compute nodes is:ENDL
                      N_Comp_Nodes=N_Switches*N_Nodes_Per_switch;
        }
      */
      params->get_vector_param("topology_geometry", args);
      if (args.size() == 0)
      {
        sst_throw(valueerror, "empty topology vector for hdtorus");
      }
      int nps = params->get_optional_int_param("network_nodes_per_switch", 0);

      return ptr(new hdtorus(args, nps, 1));

    }

    static bool
    equals(const std::vector<int>& coords, int x, int y, int z)
    {
      if (coords.size() != 3)
      {
        return false;
      }
      return coords[0] == x && coords[1] == y && coords[2] == z;
    }

    /*  Computes the index array for all vertices  for a high dimensional torus
     It uses dsize : the number of dimensions
     dimensions_:  an array of size dsize that stores each dimension
     */
    void
    hdtorus::compute_vertex_numbers(nested_type<int>::vector &vertices, int P)
    {
      int bsize = P;

      int64_t dsize = dimensions_.size();
      /* Below bsize  denotes how many consecutive  processors will be numbered
       the same in the  d-tg digit.  Note that  the for loop with  index i
       does not necessarily run P  times, since i is incremented each time
       in the innermost loop
       */
      for (int d = 0; d < dsize; d++)
      {
        bsize = bsize / dimensions_[d];
        for (int i = 0; i < P;)
          for (int k = 0; k < dimensions_[d]; k++)
            for (int j = 0; j < bsize; j++)
              vertices[i++][d] = k;
      }
    }

    // Generates  the vertex numbers --and the vertices-- for a given number
    // of processors, P
    void
    hdtorus::compute_vertices(nested_type<int>::vector &vertices,
        std::set<vertex> &vertices_set, int id)
    {
      int64_t i, P;
      int64_t dsize = dimensions_.size();

      if (!vertices.size())
      {
        std::cout << " empty vertices in compute_vertices \n";
        std::cout.flush();
        for (i = 0, P = 1; i < dsize; P *= dimensions_[i++])
          ;
      }
      P = vertices.size();

      if (id < 0) /* single processor at each node */
        for (int i = 0; i < P; i++)
        {
          vertices_set.insert(vertex(vertices[i]));
        }
      else
        /* multiple processors at each node */
        for (int i = 0; i < P; i++)
        {
          vertices[i][dsize] = id;
          vertices_set.insert(vertex(vertices[i]));
        }
    }

    /*  Generates a high dimensional torus network
     Inputs:
     dimensionarray: specifies the number of processors in each dimension
     and whether  the links are wrapped around  (torus) or not (mesh).
     dimensionarray[i] <0 refers to absence of  wrap around edges.
     dimensionarray[i] >0 refers to presence of wrap around edges.
     nndpersw is the number of nodes per switch.
     widtharrg specifies the width between the switches and the  nodes.
     */
    hdtorus::hdtorus(const std::vector<int>& dimensionarray, int nndpersw,
        int widtharg) :
      switchtopology(nndpersw)
    {
      int P, i;

      // Pre-process  the input;
      int64_t dsize = dimensionarray.size();
      nested_type<int>::vector vertices;

      dimensions_.resize(dsize);
      wrapped_.resize(dsize);

      for (i = 0; i < dsize; i++)
      {
        if (dimensionarray[i] == 0)
        {
          throw illformederror(
              "hdtorus::hdtorus: dimension of size 0 not allowed");
        }
        else if (dimensionarray[i] < 0)
        {
          wrapped_[i] = 0;
          dimensions_[i] = -dimensionarray[i];
        }
        else if (dimensionarray[i] > 2)
        {
          dimensions_[i] = dimensionarray[i];
          wrapped_[i] = 1;
        }
        else if (dimensionarray[i] == 2)
        {
          // This handles the case of dimension = 1 or 2, where
          // wrapping is turned off.
          dimensions_[i] = dimensionarray[i];
          wrapped_[i] = 0;
        }
        else
        {
          sst_throw(illformederror, "hdtorus::hdtorus: dimension of size 1 is not allowed");
        }
      }

      direct_ = (nndpersw <= 0);
      nndpersw_ = nndpersw;
      width = widtharg;

      // Generate the  nodes.
      for (i = 0, P = 1; i < dsize; P *= dimensions_[i++])
        ;

      vertices.resize(P);
      if (direct_)
        for (i = 0; i < P; vertices[i++].resize(dsize))
          ;
      else
        for (i = 0; i < P; vertices[i++].resize(dsize + 1))
          ;

      compute_vertex_numbers(vertices, P);

      if (direct_)
      { // each node is connected to the  network directly.
        compute_vertices(vertices, nodes_);
        init(nodes_);
      }
      else
      { // each node is connected to the network  through a switch
        for (int i = 0; i < nndpersw; i++)
          compute_vertices(vertices, nodes_, i);

        std::set<vertex> all(nodes_);
        // insert the switch vertices
        compute_vertices(vertices, all, nndpersw);
        init(all);
      }

    }

    // The following is an internal function that is used when computing a route between two vertices.
    // It computes the difference  along with the direction to travel the shortest distance while
    // taking into account the wrapped edges.  <0 means route by decreasing the  processors number
    // (e.g., go north) and >0 means  increase the processor number (e.g., go south)
    void
    hdtorus::compute_diffvect(const vertex &v1, const vertex &v2,
        std::vector<int> &diffvect) const
    {
      int d1;
      int64_t dsize = dimensions_.size();

      for (int64_t i = 0; i < dsize; i++)
      {
        d1 = v2.v()[i] - v1.v()[i];

        if (d1 < 0)
        {
          if (wrapped_[i] && dimensions_[i] < -2 * d1)
            diffvect[i] = dimensions_[i] + d1;
          else
            diffvect[i] = d1;
        }
        else if (d1 > 0)
        {
          if (wrapped_[i] && dimensions_[i] < 2 * d1)
            diffvect[i] = -dimensions_[i] + d1;
          else
            diffvect[i] = d1;
        }
        else
          diffvect[i] = 0;
      }
    }

    /// Returns a descriptive name of the graph.
    std::string
    hdtorus::name() const
    {
      int64_t dsize = dimensions_.size();
      std::ostringstream ostr;
      ostr << "hdtorus(";
      for (int i = 0; i < dsize; i++)
        ostr << dimensions_[i] << ",";
      ostr << ")";
      return ostr.str();
    }

    std::string
    hdtorus::label(long id)
    {
      std::vector<int> coords = get_coords(id);
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) coords.size(); ++i)
      {
        sstr << sstprintf(" %2d", coords[i]);
      }
      sstr << " ]";
      return sstr.str();
    }

    std::vector<int>
    hdtorus::get_coords(long uid) const
    {
      long div = 1;
      std::vector<int> ret(dimensions_.size());
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        ret[i] = (uid / div) % dimensions_[i];
        div *= dimensions_[i];
      }
      return ret;
    }

    long
    hdtorus::get_netid(const std::vector<int>& coords) const
    {
      long ret = 0;
      long mult = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        ret += coords[i] * mult;
        mult *= dimensions_[i];
      }
      return ret;
    }

    std::string
    hdtorus::coord_string(const std::vector<int>& coords)
    {
      std::stringstream ss;
      ss << "(";
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        ss << coords[i] << ",";
      }
      ss << ")";
      return ss.str();
    }

    int64_t
    hdtorus::topology_size() const
    {
      int64_t ret = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        ret *= dimensions_[i];
      }
      return ret;
    }

    static std::string
    get_link_name(const std::vector<int>& src, const std::vector<int>& dst)
    {
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) src.size(); ++i)
      {
        sstr << sstprintf(" %2d", src[i]);
      }
      sstr << " ]->[";
      for (int i = 0; i < (int) dst.size(); ++i)
      {
        sstr << sstprintf(" %2d", dst[i]);
      }
      sstr << "]";
      return sstr.str();
    }

    void
    hdtorus::connect_train_tracks(trainmap& switches,
                                  const std::vector<int>& red,
                                  double bytes_per_cycle,
                                  double freq_hz,
                                  timestamp link_latency,
                                  double payload_fraction,
                                  bool ignore_congestion,
                                  bool enable_multipath,
                                  const output_bw_arbitrator::ptr& arb_template)
    {
      packettrain::init_statics();

      for (long i = 0; i < (int) switches.size(); i++)
      {
        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
        std::vector<int> coords = get_coords(i);

        for (int z = 0; z < (int) dimensions_.size(); z++)
        {
          std::vector<int> connect1 = coords;
          std::vector<int> connect2 = coords;
          if (coords[z] == 0)
          {
            connect1[z]++;
            connect2[z] = dimensions_[z] - 1;
          }
          else if (coords[z] == (dimensions_[z] - 1))
          {
            connect1[z] = 0;
            connect2[z]--;
          }
          else
          {
            connect1[z]++;
            connect2[z]--;
          }
          long uid1 = get_netid(connect1);
          long uid2 = get_netid(connect2);

          nodeaddress::ptr addr1 = nodeaddress::convert_from_uniqueid(uid1);
          nodeaddress::ptr addr2 = nodeaddress::convert_from_uniqueid(uid2);

          trainmap::iterator it1 = switches.find(addr1);
          trainmap::iterator it2 = switches.find(addr2);

          if (it1 == switches.end())
          {
            sst_throw(ssterror, "hdtorus::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                uid1, addr1->to_string().c_str());
          }
          if (it2 == switches.end())
          {
            sst_throw(ssterror, "hdtorus::connect_train_tracks(): can't find switch 2 when connecting:   uid=%d  addr=%s",
                uid2, addr2->to_string().c_str());
          }

          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the pos dir to " << coord_string(connect1)
                << " which is addr " << addr1->to_string() << "\n";
          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the neg dir to " << coord_string(connect2)
                << " which is addr " << addr2->to_string() << "\n";

          std::string link_name_pos = get_link_name(coords, connect1);
          std::string link_name_neg = get_link_name(coords, connect2);

          trainswitch::ptr myswitch = switches[me];
          myswitch->connect(z, router::POS, it1->second, link_latency,
              bytes_per_cycle * red[z], freq_hz, payload_fraction,
              enable_multipath, ignore_congestion, arb_template->clone(), "network",
              link_name_pos);
          myswitch->connect(z, router::NEG, it2->second, link_latency,
              bytes_per_cycle * red[z], freq_hz, payload_fraction,
              enable_multipath, ignore_congestion, arb_template->clone(), "network",
              link_name_neg);
        }
      }
    }

    void
    hdtorus::connect_switches(switchtopology::maptype &switches,
                              const std::vector<int>& red,
                              double linkbw)
    {
      SSTMAC_DEBUG << "hdtorus: link bandwidth is " << linkbw << "\n";

      bool outputgraph = sstmac_env::params->get_optional_bool_param(
          "topology_output_graph", false);

      for (long i = 0; i < (int) switches.size(); i++)
      {
        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
        std::vector<int> coords = get_coords(i);

        std::vector<int> connected;
        std::vector<int> connected_red;

        for (int z = 0; z < (int) dimensions_.size(); z++)
        {
          std::vector<int> connect1 = coords;
          std::vector<int> connect2 = coords;
          if (coords[z] == 0)
          {
            connect1[z]++;
            connect2[z] = dimensions_[z] - 1;
          }
          else if (coords[z] == (dimensions_[z] - 1))
          {
            connect1[z] = 0;
            connect2[z]--;
          }
          else
          {
            connect1[z]++;
            connect2[z]--;
          }
          long uid1 = get_netid(connect1);
          long uid2 = get_netid(connect2);

          nodeaddress::ptr addr1 = nodeaddress::convert_from_uniqueid(uid1);
          nodeaddress::ptr addr2 = nodeaddress::convert_from_uniqueid(uid2);

          if (switches.find(addr1) == switches.end())
          {
            sst_throw(ssterror, "hdtorus::connect_switches(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                uid1, addr1->to_string().c_str());
          }
          if (switches.find(addr2) == switches.end())
          {
            sst_throw(ssterror, "hdtorus::connect_switches(): can't find switch 2 when connecting:   uid=%d  addr=%s",
                uid2, addr2->to_string().c_str());
          }

          if (outputgraph)
          {
            connected.push_back(addr1->unique_id());
            connected.push_back(addr2->unique_id());
            connected_red.push_back(red[z]);
            connected_red.push_back(red[z]);

          }

          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the pos dir to " << coord_string(connect1)
                << " which is addr " << addr1->to_string() << "\n";
          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the neg dir to " << coord_string(connect2)
                << " which is addr " << addr2->to_string() << "\n";

          switches[me]->connect(z, router::POS, switches[addr1], linkbw, red[z]);
          switches[me]->connect(z, router::NEG, switches[addr2], linkbw, red[z]);
        }

        if (outputgraph)
        {
          for (int k = 0; k < nndpersw_; k++)
          {
            std::cout << (i * nndpersw_ + k);
            for (int l = 0; l < (int) connected.size(); l++)
            {
              for (int m = 0; m < nndpersw_; m++)
              {
                std::cout << " " << (connected[l] * nndpersw_ + m) << ":"
                    << connected_red[l];

              }
            }
            for (int m = 0; m < nndpersw_; m++)
            {
              if (m != k)
              {
                std::cout << " " << (i * nndpersw_ + m) << ":" << 100000;
              }
            }
            std::cout << "\n";
          }

        }
      }

    }

    void
    hdtorus::display_nodes(const vismaptype &switches,
        const vis::vis_engine::ptr& eng, std::list<vis::vis_obj::ptr> &objs)
    {
      for (long i = 0; i < (int) switches.size(); i++)
      {
        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

        std::vector<int> coords = get_coords(i);

        double x = 0, y = 0, z = 0;
        x = coords[0] * 2;
        if (coords.size() > 1)
          y = coords[1] * 2;

        if (coords.size() > 2)
          z = coords[2] * 2;

        vis::vis_obj::ptr vo = eng->create_cube(x, y, z, 1);
        const vismaptype::mapped_type& myswitch = mapget(switches, me,
            "network switch for hdtorus");
        myswitch->set_vis_obj(vo);
        myswitch->set_vis_engine(eng);
        objs.push_back(vo);

        for (int z = 0; z < (int) dimensions_.size(); z++)
        {
          double x1 = coords[0] * 2;
          double y1 = coords[1] * 2;
          double z1 = coords[2] * 2;

          double x2 = coords[0] * 2;
          double y2 = coords[1] * 2;
          double z2 = coords[2] * 2;
          int dir;

          bool onegood = (coords[z] != dimensions_[z] - 1);
          bool twogood = (coords[z] != 0);

          if (z == 0)
          {

            x1 += 1.5;
            z1 += 0.5;
            y1 += 0.25;

            x2 -= 0.5;
            z2 += 0.5;
            y2 += 0.75;

            dir = vis::vis_engine::VIS_Z;
          }
          else if (z == 1)
          {
            y1 += 1.5;
            z1 += 0.25;
            x1 += 0.5;

            y2 -= 0.5;
            z2 += 0.75;
            x2 += 0.5;
            dir = vis::vis_engine::VIS_Y;
          }
          else
          {
            z1 += 1.5;
            y1 += 0.25;
            x1 += 0.5;

            z2 -= 0.5;
            y2 += 0.75;
            x2 += 0.5;
            dir = vis::vis_engine::VIS_X;
          }

          const vismaptype::mapped_type& myswitch = mapget(switches, me,
              "vis switch for hdtorus");
          if (onegood)
          {
            vis::vis_obj::ptr c1 = eng->create_cylinder(x1, y1, z1, 0.125, 1,
                dir);
            objs.push_back(c1);
            SSTMAC_DEBUG << "switch " << i << " setting vis link for dim " << z << " dir "
                  << router::POS << "\n";
            myswitch->set_link_vis(c1, z, router::POS);
          }

          if (twogood)
          {
            vis::vis_obj::ptr c2 = eng->create_cylinder(x2, y2, z2, 0.125, 1,
                dir);
            SSTMAC_DEBUG << "switch " << i << " setting vis link for dim " << z << " dir "
                             << router::NEG << "\n";
            myswitch->set_link_vis(c2, z, router::NEG);
            objs.push_back(c2);
          }
        }
      }
    }

  }
} //end of namespace sstmac


