/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// hdtorus.h: Interface for torus networks.
//
// Author: Ali Pinar <apinar@sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_HDTORUS_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_HDTORUS_H_INCLUDED
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/switchtopology.h>

namespace sstmac
{
  namespace hw
  {

    /** Implements a high Dimensional torus network.

     <b>Configuring the 2HDTorus network</b>

     The inputs for this network can specify the following attributes:

     <table border="1">
     <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>

     <b>Configuring the 3HDTorus network</b>

     The inputs for this network
     can specify the following attributes:

     <table border="1">
     <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>

     <b>Configuring the 4HDTorus network</b>

     The inputs for this network
     can specify the following attributes:

     <table border="1">      <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
     dimension.
     <tr><td><tt>nx</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>
     */

    class hdtorus : public switchtopology
    {
    public:
      int width;

    protected:
      std::string name_;
      bool direct_;
      int nndpersw_;
      std::vector<int> dimensions_;
      std::vector<bool> wrapped_;
      void
      compute_diffvect(const vertex &v1, const vertex &v2,
          std::vector<int> &diffvect) const;

      void
      compute_vertex_numbers(nested_type<int>::vector &vertices, int P);

      void
      compute_vertices(nested_type<int>::vector& vertices,
                       std::set<vertex> &vertices_set, int id = -1);
    protected:
      hdtorus(const std::vector<int>& dimensionarray, int nndpersw, int widtharg);

    public:
      typedef boost::intrusive_ptr<hdtorus> ptr;

      static hdtorus::ptr
      construct(sim_parameters::ptr params);

      std::vector<int>
      get_coords(long i) const;

      long
      get_netid(const std::vector<int>& v) const;

      virtual std::string label(long id);

      virtual int
      radix() const
      {
        return 2 * dimensions_.size();
      }

      std::string
      coord_string(const std::vector<int>& coords);

      virtual
      ~hdtorus()
      {
      }

      std::string
      name() const;

      /// Returns the vector giving each dimension of the torus.
      const std::vector<int> &
      dimensions() const
      {
        return dimensions_;
      }

      virtual int64_t
      topology_size() const;

      virtual void
      display_nodes(const vismaptype &switches, const vis::vis_engine::ptr& eng,
          std::list<vis::vis_obj::ptr> &objs);

      virtual int
      ndimensions() const {
          return dimensions_.size();
      }

      virtual void
      connect_train_tracks(trainmap& switches,
                           const std::vector<int>& red,
                           double bytes_per_cycle,
                           double freq_hz,
                           timestamp link_latency,
                           double payload_fraction,
                           bool ignore_congestion,
                           bool enable_multipath,
                           const output_bw_arbitrator::ptr& arb_template);

      virtual void
      connect_switches(switchtopology::maptype &switches,
                       const std::vector<int>& red,
                       double linkbw);

      virtual std::string
      to_string() const
      {
        return "hdtorustopology";
      }
    };

  }
} //end of namespace sstmac

#endif
