/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/hdtorusrouter.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("torus | hdtorus", router, hdtorusrouter);

    std::vector<int>
    hdtorusrouter::get_coords(long uid)
    {
      long div = 1;
      std::vector<int> ret(dims_.size());
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        ret[i] = (uid / div) % dims_[i];
        div *= dims_[i];
      }
      return ret;
    }

    long
    hdtorusrouter::get_uid(const std::vector<int>& coords)
    {
      long ret = 0;
      long mult = 1;
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        ret += coords[i] * mult;
        mult *= dims_[i];
      }
      return ret;
    }

    int
    hdtorusrouter::get_dir(const std::vector<int>& coords, int i)
    {
      if (coords[i] > coords_[i] && fabs(coords_[i] + (dims_[i] - coords[i]))
          > dims_[i] / 2 || coords[i] < coords_[i] && fabs(
          coords[i] + (dims_[i] - coords_[i])) <= dims_[i] / 2)
      {
        return POS;
      }
      else if (coords[i] < coords_[i] && fabs(
          coords[i] + (dims_[i] - coords_[i])) > dims_[i] / 2 || coords[i]
          > coords_[i] && fabs(coords_[i] + (dims_[i] - coords[i])) <= dims_[i]
          / 2)
      {
        return NEG;
      }
      else
      {
        return EJECT;
      }
    }

    void
    hdtorusrouter::xyz_route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      std::vector<int> coords = get_coords(netaddr->unique_id());
      SSTMAC_DEBUG << "hdtorusrouter(" << me_->to_string() << "): heading to "
            << to->to()->to_string() << " at net addr " << netaddr->to_string()
            << "\n";
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        if (coords[i] != coords_[i])
        {
          dim = i;
        }
        else
        {
          continue;
        }

        if (coords[i] > coords_[i] && fabs(coords_[i] + (dims_[i] - coords[i]))
            > dims_[i] / 2 || coords[i] < coords_[i] && fabs(
            coords[i] + (dims_[i] - coords_[i])) <= dims_[i] / 2)
        {

          dir = POS;
        }
        else if (coords[i] < coords_[i] && fabs(
            coords[i] + (dims_[i] - coords_[i])) > dims_[i] / 2 || coords[i]
            > coords_[i] && fabs(coords_[i] + (dims_[i] - coords[i]))
            <= dims_[i] / 2)
        {

          dir = NEG;

        }
        else
        {
          throw ssterror("hdtorusrouter::route -- i'm confused");
        }
        SSTMAC_DEBUG << "returning dim " << dim << " and dir " << dir << "\n";
        return;
      }

      dim = networkswitch::EJECT_PORT;
      dir = to->to()->unique_id() % conc_;
    }

    hdtorusrouter::ptr
    hdtorusrouter::construct(sim_parameters::ptr params)
    {
      route_model_t model;
      /**
        sstkeyword {
            gui=turn;
            turn=Follows +X,+Y,+Z,-X,-Y,-Z route order;
            xyz=Follows +X,-X,+Y,-Y,+Z,-Z route order;
            seastar=Follows +X,-X,+Y,+Z,-Y,-Z route order;
        }
      */
      std::string route_order = params->get_optional_param(
          "torus_routing_model", "xyz");
      if (route_order == "xyz")
      {
        model = xyz_model;
      }
      else if (route_order == "turn")
      {
        model = turn_model;
      }
      else if (route_order == "disperse")
      {
        model = disperse_model;
      }
      else if (route_order == "seastar")
      {
        model = seastar_model;
      }
      else if (route_order == "congestion")
      {
        model = congestion_model;
      }
      else if (route_order == "dimension")
      {
        model = dimension_order_model;
      }
      else if (route_order == "wrongturn")
      {
        model = wrong_turn_model;
      }
      else
      {
        sst_throw(inputerror,
          "hdtorusrouter::construct: invalid torus routing algorithm %s\n",
          route_order.c_str());
      }
      return new hdtorusrouter(model);
    }

    void
    hdtorusrouter::disperse_model_route(routing_info::ptr to, int &dim,
                                        int &dir)
    {
        nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
        std::vector<int> coords = get_coords(netaddr->unique_id());
        SSTMAC_DEBUG << "hdtorusrouter(" << me_->to_string() << "): heading to "
              << to->to()->to_string() << " at net addr "
              << netaddr->to_string()
              << "\n";

        if (coords.size() != 3){
            sst_throw(illformederror,
              "hdtorusrouter::disperse_model_route: 3D torus only");
        }

        int dir_z = get_dir(coords, 2);
        if (dir_z == POS){
            dim = 2;
            dir = POS;
            return;
        }

        int dir_y = get_dir(coords, 1);
        if (dir_y == POS){
            dim = 1;
            dir = POS;
            return;
        }

        int dir_x = get_dir(coords, 0);
        if (dir_x == POS){
            dim = 0;
            dir = POS;
            return;
        }

        if (dir_z == NEG){
            dim = 2;
            dir = NEG;
            return;
        }

        if (dir_y == NEG){
            dim = 1;
            dir = NEG;
            return;
        }

        if (dir_x == NEG){
            dim = 0;
            dir = NEG;
            return;
        }

        dim = networkswitch::EJECT_PORT;
        dir = to->to()->unique_id() % conc_;
        SSTMAC_DEBUG << "ejecting \n";
    }

    void
    hdtorusrouter::seastar_model_route(routing_info::ptr to, int &dim, int &dir)
    {
        nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
        std::vector<int> coords = get_coords(netaddr->unique_id());
        SSTMAC_DEBUG << "hdtorusrouter(" << me_->to_string() << "): heading to "
              << to->to()->to_string() << " at net addr "
              << netaddr->to_string()
              << "\n";

        if (coords.size() != 3){
            sst_throw(illformederror,
              "hdtorusrouter::disperse_model_route: 3D torus only");
        }

        int dir_x = get_dir(coords, 0);
        if (dir_x == POS){
            dim = 0;
            dir = POS;
            return;
        }
        else if (dir_x == NEG){
            dim = 0;
            dir = NEG;
            return;
        }

        int dir_y = get_dir(coords, 1);
        if (dir_y == POS){
            dim = 1;
            dir = POS;
            return;
        }

        int dir_z = get_dir(coords, 2);
        if (dir_z == POS){
            dim = 2;
            dir = POS;
            return;
        }

        if (dir_y == NEG){
            dim = 1;
            dir = NEG;
            return;
        }

        if (dir_z == NEG){
            dim = 2;
            dir = NEG;
            return;
        }

        dim = networkswitch::EJECT_PORT;
        dir = to->to()->unique_id() % conc_;
        SSTMAC_DEBUG << "ejecting \n";
    }

    void
    hdtorusrouter::wrong_turn_model_route(routing_info::ptr to, int &dim,
                                          int &dir)
    {
        nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
        std::vector<int> coords = get_coords(netaddr->unique_id());
        SSTMAC_DEBUG << "hdtorusrouter(" << me_->to_string() << "): heading to "
            << to->to()->to_string() << " at net addr " << netaddr->to_string()
            << "\n";
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        if (coords[i] > coords_[i])
        {
          dim = i;
        }
        else
        {
          continue;
        }

        dir = get_dir(coords, i);
        return;
      }
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        if (coords[i] < coords_[i])
        {
          dim = i;
        }
        else
        {
          continue;
        }

        dir = get_dir(coords, i);
        return;
      }

      dim = networkswitch::EJECT_PORT;
      dir = to->to()->unique_id() % conc_;
      SSTMAC_DEBUG << "ejecting \n";
    }


    void
    hdtorusrouter::turn_model_route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      std::vector<int> coords = get_coords(netaddr->unique_id());
      std::vector<int> dirs(coords.size(), 0);

      //check for positive shift
      for (int i=0; i < (int) dims_.size(); ++i){
          dir = get_dir(coords, i);
          if (dir == POS){
              dim = i;
              return;
          }
          else {
              dirs[i] = dir;
          }
      }

      //check for negative shift
      for (int i=0; i < (int) dims_.size(); ++i){
          dir = dirs[i];
          if (dir == NEG){
              dim = i;
              return;
          }
      }

      dim = networkswitch::EJECT_PORT;
      dir = to->to()->unique_id() % conc_;
      SSTMAC_DEBUG << "ejecting \n";
    }

    void
    hdtorusrouter::dimension_order_route(routing_info::ptr to, int &dim,
                                         int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      std::vector<int> coords = get_coords(netaddr->unique_id());
      //check for positive shift
      for (int i=0; i < (int) dims_.size(); ++i){
          dir = get_dir(coords, i);
          if (dir != EJECT){
              dim = i;
              return;
          }
      }

      dim = networkswitch::EJECT_PORT;
      dir = to->to()->unique_id() % conc_;
      SSTMAC_DEBUG << "ejecting \n";
    }

    void
    hdtorusrouter::congestion_model_route(routing_info::ptr to, int &dim,
                                          int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      std::vector<int> coords = get_coords(netaddr->unique_id());
      SSTMAC_DEBUG << "hdtorusrouter(" << me_->to_string() << "): heading to "
            << to->to()->to_string() << " at net addr " << netaddr->to_string()
            << "...";

      boost::unordered_map<int, bool> there;
      int therecount = 0;
      int backupd = -1, backupr = -1;
      for (int i = 0; i < (int) dims_.size(); i++)
      {
        if (coords[i] == coords_[i])
        {
          there[i] = true;
          therecount++;
        }
        else
        {
          there[i] = false;
          backupd = i;
          backupr = get_dir(coords, i);
        }

      }

      if (therecount == (int) dims_.size())
      {
        SSTMAC_DEBUG << "ejecting \n";
        dim = networkswitch::EJECT_PORT;
        dir = to->to()->unique_id() % conc_;
        SSTMAC_DEBUG << "ejecting \n";
      }
      else
      {

        //find the min
        int d, r;
        int bestd = -1, bestr = -1;

        double min_cong = 1000000;
        boost::unordered_map<int, double>::iterator it, end = cong_.end();
        for (it = cong_.begin(); it != end; it++)
        {
          convert_to_dimdir(it->first, d, r);
          if (d >= 0 && !there[d] && r == get_dir(coords, d))
          {
            if (it->second < min_cong)
            {
              min_cong = it->second;
              bestd = d;
              bestr = r;
            }
          }
        }

        if (bestd < 0)
        {
          SSTMAC_DEBUG << "reverting to backup for some reason \n";
          dim = backupd;
          dir = backupr;
        }
        else
        {
          SSTMAC_DEBUG << "selecting minimal congestion port \n";
          dim = bestd;
          dir = bestr;
        }
      }

    }

    void
    hdtorusrouter::route(routing_info::ptr to, int &dim, int &dir)
    {
      switch (model_)
      {
      case xyz_model:
        xyz_route(to, dim, dir);
        break;
      case turn_model:
        turn_model_route(to, dim, dir);
        break;
      case wrong_turn_model:
        wrong_turn_model_route(to, dim, dir);
        break;
      case disperse_model:
        disperse_model_route(to, dim, dir);
        break;
      case seastar_model:
        seastar_model_route(to, dim, dir);
        break;
      case dimension_order_model:
        dimension_order_route(to, dim, dir);
        break;
      case congestion_model:
        congestion_model_route(to, dim, dir);
        break;
      }
    }

    int
    hdtorusrouter::convert_to_outport(int dim, int dir)
    {
      if (dim == networkswitch::EJECT_PORT)
      {
        return -(1 + dir);
      }
      return dim * 2 + (dir == POS ? 0 : 1);

    }

    void
    hdtorusrouter::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
      {
        dim = networkswitch::EJECT_PORT;
        dir = (-outport) - 1;
      }
      else
      {
        dim = outport / 2;
        dir = (outport % 2 == 0 ? POS : NEG);
      }
    }

    int
    hdtorusrouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {
      return convert_to_outport(dim, get_opp_dir(dir));
    }

  }
}
