/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWTICHES_ROUTING_TORUSROUTER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWTICHES_ROUTING_TORUSROUTER_H_INCLUDED

#include <sstmac/hardware/network/topology/router.h>
#include <sstmac/hardware/network/topology/hdtorus.h>
#include <sstmac/common/messages/sst_message.h>

namespace sstmac
{
  namespace hw
  {

    class hdtorusrouter : public router
    {
    public:
      typedef boost::intrusive_ptr<hdtorusrouter> ptr;

      typedef enum
      {
        xyz_model, turn_model, congestion_model, disperse_model, seastar_model, dimension_order_model, wrong_turn_model
      } route_model_t;

      static ptr
      construct(sim_parameters::ptr params);

      virtual void
      route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route minimally in the order X,Y,Z regardless of
        +/- direction.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      xyz_route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route minimally in the order +X,+Y,+Z,-X,-Y,-Z
        Here +X means X(dest) > X(src) including wrap around.
        For N(X)=5, 0->4 is considered -X.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      turn_model_route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route minimally in the order +X,+Y,+Z,-X,-Y,-Z.
        Here +X means X(dest) > X(src) neglecting wrap around.
        For N(X)=5, 0->4 is considered +X.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      wrong_turn_model_route(routing_info::ptr to, int &dim, int &dir);


      /**
        Route minimally in the order +Z,+Y,+X,-Z,-Y,-X.
        Here +X means X(dest) > X(src) including wrap around.
        For N(X)=5, 0->4 is considered -X.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      disperse_model_route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route minimally in the order +X,-X,+Y,+Z,-Y,-Z.
        Here +X means X(dest) > X(src) including wrap around.
        For N(X)=5, 0->4 is considered -X.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      seastar_model_route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route along whichever link has the least congestion,
        provided it makes progress towards the destination.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      congestion_model_route(routing_info::ptr to, int &dim, int &dir);

      /**
        Route minimally in the order +X,-X,+Y,-Y,+Z,-Z
        Here +X means X(dest) > X(src) including wrap around.
        For N(X)=5, 0->4 is considered -X.
        @param to The destination node address for a message
        @param[out] The dimension to route on (x,y,z)
        @param[out] The direction to route on (+/-)
      */
      void
      dimension_order_route(routing_info::ptr to, int &dim, int &dir);

      /**
        @return Whether the routing model makes decisions based
                on congestion
      */
      virtual bool
      needs_congestion_info() const
      {
        return model_ == congestion_model;
      }

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      virtual
      ~hdtorusrouter()
      {
      }

      virtual void
      init(const nodeaddress::ptr& me, const topology::ptr& top)
      {
          hdtorus::ptr torustop = safe_cast<hdtorus>(top);
        router::init(torustop->nps(), me);
        dims_ = torustop->dimensions();
        coords_ = get_coords(me->unique_id());
      }

      virtual int
      get_inport(nodeaddress::ptr from, int dim, int dir);

    protected:
      std::vector<int> dims_;
      std::vector<int> coords_;
      route_model_t model_;

      /**
        @param coords The x,y,z destination coordinates
        @param axis The axis(x,y,z...) to check
        @return The direction with the shortest wrap-around path
                to the destination
      */
      int
      get_dir(const std::vector<int>& coords, int axis);

      hdtorusrouter(route_model_t model) :
        model_(model)
      {
      }

      /**
        The inverse operation of #get_uid
        @param A unique node id (network node, not compute node)
        @return The coordinates of the node in the network
      */
      std::vector<int>
      get_coords(long i);

      /**
        The inverse operation of #get_coords
        @param The coordinates of a network node
        @return The unique node id
      */
      long
      get_uid(const std::vector<int>& v);

    };

  }
}
#endif
