/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_ROUTING_ROUTER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_ROUTING_ROUTER_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/network/topology/routing_info.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {

    /**
      @class router
      Class that computes the next step a messag should taken in traversing
      the network.  This performs routing operations only and is not actually
      a 'component' in the network - those are switches.  Switch and router
      are not synonymous in SST/macro.  All switches have routers.
    */
    class router :
        public factory_type
    {
    public:
      typedef boost::intrusive_ptr<router> ptr;

      virtual std::string
      to_string() const
      {
        return "router";
      }

      virtual
      ~router()
      {
      }

      virtual void
      init(int npr, const nodeaddress::ptr& me)
      {
        conc_ = npr;
        me_ = me;
      }

      virtual void init(const nodeaddress::ptr& me, const topology::ptr& top) = 0;

      /**
        @param to   The routing info containing the destination address
        @param[out] The dimension (e.g. x, y, or z) that should be routed on next.
                    Integer value has different meaning depending on type of
                    router (torus/dragonfly/etc)
        @param[out] The direction (e.g. +/-) that should be routed on.
                    Integer value has different physical meaning depending on type
                    of router (torus/dragonly/etc)
      */
      virtual void
      route(routing_info::ptr to, int &dim, int &dir) = 0;

      /**
        @param The dimension (e.g. x, y, or z) that should be routed on next.
                    Integer value has different meaning depending on type of
                    router (torus/dragonfly/etc)
        @param The direction (e.g. +/-) that should be routed on.
                    Integer value has different physical meaning depending on type
                    of router (torus/dragonly/etc)
        @return The port number on a switch that the message should be routed through
                for a given dim/dir
      */
      virtual int
      convert_to_outport(int dim, int dir) = 0;

        /**
        @param The port number on a switch that the message should be routed through
                for a given dim/dir
        @param[out] The dimension (e.g. x, y, or z) that should be routed on next.
                    Integer value has different meaning depending on type of
                    router (torus/dragonfly/etc)
        @param[out] The direction (e.g. +/-) that should be routed on.
                    Integer value has different physical meaning depending on type
                    of router (torus/dragonly/etc)
        */
      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir) = 0;

      /**
        @enum The direction a message should route in a given direction.
              Generally, only POS and NEG are used. The rest only
              really have significance in a 3D torus. Ejection
              indicates the message has arrived at the correct
              address.
      */
      enum DIRECTION
      {
        POS, NEG, UP, DOWN, EAST, WEST, NORTH, SOUTH, EJECT = -1
      };

      /**
        @return  The number of nodes per router (i.e. the concentration)
      */
      int
      conc() const
      {
        return conc_;
      }

      /**
        @return The inport
      */
      virtual int
      get_inport(nodeaddress::ptr from, int dim, int dir) = 0;

      /**
      */
      virtual bool
      is_fixed_routing() const
      {
        return false;
      }

      /**
        @return Whether the router needs congestion info to properly make decisions.
                This is usally used to notify the switch that it should update
                congestion info - i.e. the router makes the routing decisions
                but the switch is what actually knows about congestion.
      */
      virtual bool
      needs_congestion_info() const {
        return false;
      }

      /**
        @param num  The port number to add congestion info for
        @param val  The congestion value. Units will be router-specific.
                    Torus router may use, e.g., the soonest time a link
                    is free as the metric.
      */
      virtual void
      set_congestion(int num, double val){
        SSTMAC_DEBUG2 << "router(" << me_->to_string() << "): setting congestion for port "
            << num  << " to " << val << "\n";
        cong_[num] = val;
      }

      /**
        @param The node the port is connected to
        @param The port number
      */
      virtual void
      set_inport(nodeaddress::ptr from, int port)
      {
      }

    protected:
      nodeaddress::ptr me_;

      int conc_;

      static sstmac::debug dbg_;

      static bool template_init_;

      boost::unordered_map<int, double> cong_;

      router() : conc_(0)
      {

        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> router | network");
            logger::register_logger(&dbg_);
          }

      }

      /**
        @param A compute node address.  This is not necessarily equal to
               the network node address since the there may be
               more than one compute node per network switch.
        @param A network node address.
      */
      nodeaddress::ptr
      convert_to_network_addr(nodeaddress::ptr n)
      {
        long uid = n->unique_id() / conc_;
        return nodeaddress::convert_from_uniqueid(uid);

      }

      /**
        @param dir  The direction (e.g. +/-) to be reversed
        @return The opposite direction (+:- , -:+)
      */
      DIRECTION
      get_opp_dir(int dir)
      {

        switch (dir)
          {
        case POS:
          return NEG;

        case NEG:
          return POS;

        case UP:
          return DOWN;

        case DOWN:
          return UP;

        case EAST:
          return WEST;

        case WEST:
          return EAST;

        case NORTH:
          return SOUTH;

        case SOUTH:
          return NORTH;
          }

        sst_throw(valueerror, "No router opposite direction found for %d\n", dir);
        return POS;
      }
    };

    DeclareFactory2InitParams(router,nodeaddress,topology);

  }
}
#endif
