/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// network.h: Interface for networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOPOLOGY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOPOLOGY_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/network/topology/vertex.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/factories/factory.h>

namespace sstmac { namespace hw {

  /** A network is represented by a graph.  A subset of the graph's vertices
 correspond to nodes. Other vertices may be switching elements. */
  class topology :
          public factory_type
  {
  private:
    typedef std::map<vertex_index, vertex_index> mapvv;
    mapvv node_to_switch_;
  protected:

    /// Nodes are identified using "vertex" labels (rest of graph stuff has
    /// been removed).
    std::set<vertex> nodes_;
    std::vector<vertex> inodes_;
    std::map<vertex, int64_t> nodesi_;

    static debug dbg_;

    topology()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> topology | network");
        logger::register_logger(&dbg_);
      }
    }

    /// The nodes_ set must be initialized before init() is called.
    void
    init()
    {
      int index = 0;
      inodes_.clear();
      for (std::set<vertex>::iterator i = nodes_.begin();
        i != nodes_.end(); i++,index++) {
        inodes_.push_back(*i);
        nodesi_.insert(std::make_pair(*i,index));
      }
    }

    void
    init(const std::set<vertex> &v)
    {
      nodes_.clear();
      for (std::set<vertex>::iterator i = v.begin();
           i != v.end(); i++) {
          nodes_.insert(*i);
      }
   
      init();
    }

  public:

    typedef boost::intrusive_ptr<topology> ptr;

    ~topology() {}

    virtual std::string to_string() const{
      return "topology";
    }

    /// Returns a descriptive name of the graph.
    virtual std::string
    name() const;

    /**
        Given an index input, return a sensible label
        (e.g. hdtorus returns x,y,z coordinates) to
        identify the vertx
    */
    virtual std::string label(long id);

    /// Returns the number of nodes.
    virtual int64_t n() const
    {
      return nodes_.size();
    }

    /// Returns the set of vertices corresponding to the nodes.
    const std::set<vertex>&
    nodes() const
    {
      return nodes_;
    }

    /// Returns the vertex of the node with the given index.
    const vertex &
    index_to_node(int64_t i) const
    {
      return inodes_[i];
    }

    /// Converts a node vertex to its index.
    int64_t node_to_index(const vertex &v) const;

  };

  DeclareFactory(topology);

  } } //end of namespace sstmac

#endif
