/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/circuitnic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/waiter.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("circuitnic | circuit", networkinterface, circuitnic);

    circuitnic::map_type circuitnic::node_map_;

    long circuitnic::got_datas_ = 0;
    long circuitnic::got_teardowns_ = 0;
    long circuitnic::got_path_acks_ = 0;
    long circuitnic::got_setups_ = 0;
    long circuitnic::numsends_ = 0;
    bool circuitnic::printed_ = false;
    long circuitnic::timeouts_ = 0;
    //
    // Hi.
    //
    circuitnic::circuitnic(const sim_parameters::ptr& params) :
      networkinterface()
    {
      bw_ = params->get_bandwidth_param("circuitnic_bandwidth");
      propdelay_ = params->get_time_param("circuitnic_propdelay");
      busy_ = false;
      firstsendset_ = true;
      setup_timeout_ = params->get_optional_time_param(
          "circuitnic_setup_timeout", timestamp(0));

    }

    void
    circuitnic::init(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn)
    {
      networkinterface::init(parent, interconn);
      if (parent->get_id())
        {
          ptr s(this);
          node_map_[parent->get_id()] = s;
        }
    }

    void
    circuitnic::send(const nodeaddress::ptr &recver,
        const sst_message::ptr &payload)
    {
      use_me();
      job j;
      j.payload = payload;
      j.recver = recver;
      j.arrived = now();
      jobs_.push(j);
      check_jobs();
      numsends_++;

      //sender_->send(payload, recver, completion);
    }

    void
    circuitnic::finish()
    {
        if (finished_)
            return;

        networkinterface::finish();
      if (jobs_.size() > 0)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") still has " << jobs_.size() << " jobs \n";
        }

      if (!printed_)
        {
          SSTMAC_DEBUG << "got " << numsends_ << " jobs \n";
          SSTMAC_DEBUG << "got " << got_setups_ << " setups \n";
          SSTMAC_DEBUG << "got " << got_path_acks_ << " path acks \n";
          SSTMAC_DEBUG << "got " << got_datas_ << " datas \n";
          SSTMAC_DEBUG << "got " << got_teardowns_ << " teardowns \n";
          SSTMAC_DEBUG << "timeouts: " << timeouts_ << " \n";
          printed_ = true;
        }

    }

    void
    circuitnic::check_jobs()
    {
      if (!busy_ && jobs_.size() > 0)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") starting job \n";
          sst_message::ptr msg = new_message_callback(circuitnic::ptr(this),
              &circuitnic::send_to_network_link, jobs_.front().payload,
              jobs_.front().recver, jobs_.front().arrived);
          get_parent_node()->get_os()->execute_kernel(sstmac::sw::AMI::AMI_COMP_MEMREAD, msg);
          jobs_.pop();
          busy_ = true;
        }
    }

    //
    // Schedule a send operation to happen at the given time.
    //
    void
    circuitnic::send_to_network_link(const sst_message::ptr& payload,
        const nodeaddress::ptr& recver, const timestamp &arrived)
    {
      SSTMAC_DEBUG << "circuitnic::send(" << recver->to_string() << ", "
            << payload->to_string() << ")\n";

      if (payload == 0)
        {
          std::cerr << "I'm here throwing this payload error" << std::endl;
          throw nullerror("circuitnic::send: null payload pointer.");
        }

      timestamp delayed = now(); //don't model delay for now


      circuitmessage::ptr circ = circuitmessage::construct(payload, 64,
          circuitmessage::SETUP);
      circ->fromaddr() = payload->fromaddr();
      circ->toaddr() = payload->toaddr();
      circ->arrive_ = arrived;

      if (!firstsendset_)
        {

        }
      else
        {
          firstsendset_ = false;
        }
      current_ = circ;
      interconn_->send(delayed, circ);

      if (setup_timeout_ > timestamp(0))
        {
          circ->ttl_ = delayed + setup_timeout_;
          eventhandler::ptr ev =
                  new_callback(ptr(this), &circuitnic::timeout, circ);
          schedule(delayed + setup_timeout_, ev, circ);
        }
      else
        {
          circ->ttl_ = timestamp(timestamp::max_time() - 10);
        }

      spy_->add_one(payload->fromaddr()->unique_id(),
          payload->toaddr()->unique_id());
      spy_bytes_->add(payload->fromaddr()->unique_id(),
          payload->toaddr()->unique_id(), payload->get_byte_length());
    }

    networkinterface::ptr
    circuitnic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      circuitnic::ptr ret = circuitnic::construct( params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    circuitnic::handle(const sst_message::ptr &msg)
    {

      validate_msg_supported(msg);
      recv(msg);

    }

    void
    circuitnic::timeout(sst_message::ptr msg)
    {

      circuitmessage::ptr circ = boost::dynamic_pointer_cast<circuitmessage>(
          msg);
      if (circ == current_) // we're still working on this, redo it
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") timed out, restarting setup \n";

          timeouts_++;

          circuitmessage::ptr teardown = circuitmessage::construct(
              circ->orig(), 64, circuitmessage::TEARDOWN);

          teardown->toaddr() = current_->toaddr();
          teardown->fromaddr() = current_->fromaddr();

          timestamp teartime = now();
          interconn_->send(teartime, teardown);

          circuitmessage::ptr circ2 = circuitmessage::construct(circ->orig(),
              64, circuitmessage::SETUP);
          circ2->fromaddr() = current_->fromaddr();
          circ2->toaddr() = current_->toaddr();

          current_ = circ2;
          circ2->backoff_ = circ->backoff_ + 1;

          timestamp at = now() + (rand() % 10 == 0 ? timestamp(1e-9)
              : timestamp(circ2->backoff_ * 50e-9 + (rand() % 5) * 10e-9));

          circ2->ttl_ = at + setup_timeout_;

          eventhandler::ptr outev = new_callback(ptr(this),
              &circuitnic::send_out_resetup, circ2);
          schedule(at, outev, circ2);

          if (setup_timeout_ > timestamp(0))
            {

              eventhandler::ptr ev = new_callback(ptr(this), &circuitnic::timeout,
                  circ2);
              schedule(at + setup_timeout_, ev, circ2);
            }
          else
            {
              circ2->ttl_ = timestamp(timestamp::max_time() - 10);
            }
        }

    }

    void
    circuitnic::send_out_resetup(const sst_message::ptr &msg)
    {
      if (current_)
        {
          interconn_->send(now(), msg);
        }
    }

    void
    circuitnic::recv(const sst_message::ptr &msg)
    {
      circuitmessage::ptr circ = boost::dynamic_pointer_cast<circuitmessage>(
          msg);

      if (!circ)
        {
          if (msg)
            std::cerr << "msg: " << msg->to_string() << "\n";
          else
            std::cerr << "null msg. \n";

          throw ssterror(
              "message arrived at nic that can't be cast to circuitmessage");
        }

      if (!circ->toaddr()->equals(parent_->get_id()))
        {
          throw ssterror("message arrived at wrong nic");
        }

      if (circ->get_circ_type() == circuitmessage::BLOCKED)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - got BLOCKED while heading for "
                << circ->fromaddr()->to_string() << "\n";
          circ->set_circ_type(circuitmessage::SETUP);
          circ->backoff_++;
          nodeaddress::ptr addr = circ->toaddr();
          circ->toaddr() = circ->fromaddr();
          circ->fromaddr() = addr;

          timestamp delayed = now() + (rand() % 10 == 0 ? timestamp(0)
              : timestamp(circ->backoff_ * 100e-9));

          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - backing off for " << timestamp(circ->backoff_ * 100e-9)
                << "\n";
          interconn_->send(delayed, circ);

        }
      else if (circ->get_circ_type() == circuitmessage::PATH_ACK)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - got PATH_ACK with msg " << circ->to_string()
                << " with orig " << circ->orig()->to_string() << "\n";
          got_path_acks_++;
          if (!circ->orig())
            {
              std::cerr << "msg: " << circ->to_string() << "\n";
              throw ssterror("circuitnic: got path_ack, but there's no orig()");
            }
          circuitmessage::ptr circout = circuitmessage::construct(circ->orig(),
              circ->orig()->get_byte_length(), circuitmessage::DATA);

          if (!circout)
            {
              throw ssterror(
                  "circuitnic: in path_ack, can't cast orig to circuitmessage");
            }
          circout->fromaddr() = circ->toaddr();
          circout->toaddr() = circ->fromaddr();

          if (!current_)
            {
              throw ssterror(
                  "circuitnic: got a path_ack when I'm not waiting for one");
            }

          if (current_->backoff_ != circ->backoff_)
            {
              SSTMAC_DEBUG << "returning \n";
              return;
            }

          SSTMAC_DEBUG << "circuitnic: getting dstnic at "
                << circ->orig()->toaddr()->to_string() << "\n";

          circuitnic::ptr dstnic = node_map_[circ->orig()->toaddr()];

          if (!dstnic)
            {
              throw ssterror(
                  "circuitnic: can't find dstnic at address "
                      + circ->orig()->toaddr()->to_string());
            }
          timestamp at = now() + propdelay_ + timestamp(
              (double) circ->orig()->get_byte_length() / bw_);
          schedule(at, dstnic, circout);

          //schedule a message so we check the jobs again
          eventhandler::ptr ev = new_callback(ptr(this), &circuitnic::newmsg,
              sst_message::ptr());
          schedule(at + timestamp(10e-9), ev, sst_message::ptr());

          double stat = at.sec() - current_->arrive_.sec();
          total_lat_->record(stat);

          if (circ->orig()->get_needs_ack())
            {
              SSTMAC_DEBUG << "sending back send ack \n";
              sst_message::ptr sendack = circ->orig()->clone(
                  sst_message::SEND_ACK);
              schedule(now(), parent_, sendack);
            }

          current_ = circuitmessage::ptr();
          circuitmessage::ptr teardown = circuitmessage::construct(
              circ->orig(), 64, circuitmessage::TEARDOWN);

          teardown->toaddr() = circout->toaddr();
          teardown->fromaddr() = circout->fromaddr();

          timestamp teartime = at - propdelay_;
          interconn_->send(teartime, teardown);

        }
      else if (circ->get_circ_type() == circuitmessage::DATA)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - got DATA \n";
          got_datas_++;


          sst_message::ptr callback_msg;
          if (circ->is_tail()){
              callback_msg = new_message_callback(parent_, &node::handle, circ->orig());
          }
          else {
              callback_msg = circ; //just do the memory access - don't pass off to OS yet
              circ->add_callback(donothingcallback::construct());
          }

          get_parent_node()->get_os()->execute_kernel(
              sstmac::sw::AMI::AMI_COMP_MEMWRITE,
              callback_msg);
        }
      else if (circ->get_circ_type() == circuitmessage::SETUP)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - got SETUP \n";
          got_setups_++;
          circuitmessage::ptr pack = circuitmessage::construct(circ->orig(),
              64, circuitmessage::PATH_ACK);
          nodeaddress::ptr addr = circ->fromaddr();
          pack->fromaddr() = circ->toaddr();
          pack->toaddr() = addr;

          pack->backoff_ = circ->backoff_;

          interconn_->send(now(), pack);
        }
      else if (circ->get_circ_type() == circuitmessage::TEARDOWN)
        {
          SSTMAC_DEBUG << "circuitnic(" << parent_->get_id()->to_string()
                << ") - got TEARDOWN \n";
          got_teardowns_++;
        }
      else
        {
          std::cerr << "type: " << circ->get_circ_type() << "\n";
          throw ssterror(
              "circuitnic::recv -- got a message I don't know what to do with");
        }

    }

  }
} // end of namespace sstmac.
