/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/rdma_nic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("rdma | rdma_nic", networkinterface, rdma_nic,
                "implements a nic that is capable of sending rmda get/put without interrupting CPU");

    // Pointer types.

    using namespace sstmac::hw;

    rdma_nic::rdma_nic() :
      networkinterface()
    {
    }

    rdma_nic::ptr
    rdma_nic::construct(const sim_parameters::ptr& params)
    {
      return rdma_nic::ptr(new rdma_nic);
    }

    //
    // Goodbye.
    //
    rdma_nic::~rdma_nic() throw ()
    {
    }

    networkinterface::ptr
    rdma_nic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      rdma_nic::ptr ret = rdma_nic::construct(params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    rdma_nic::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic: " << get_parent_id()->to_string()
            << "handle message: " << msg->to_string() << "\n";
      sst_message::ptr orig_msg = msg->get_parent_msg();
      switch (orig_msg->get_type())
      {
      case sst_message::RDMA_GET_REQ:
        if (!msg->is_tail())
        {
          sst_throw(illformederror, "RDMA requests shoud always be a single message");
        }
        send_rdma_rsp(orig_msg);
        break;
      case sst_message::RDMA_GET_RSP:
        recv_rdma_rsp(msg);
        break;
      case sst_message::RDMA_PUT_REQ:
      case sst_message::RDMA_PUT_RSP:
        throw unimplementederror("rdma_nic: RDMA PUT not yet allowed");
        break;
      case sst_message::SEND_ACK:
        schedule(now(), parent_, msg);
        break;
      default:
        recv(msg);
        break;
      }
    }

    void
    rdma_nic::set_eventmanager(const eventmanager::ptr &m)
    {
      networkinterface::set_eventmanager(m);
    }

    void
    rdma_nic::send(const nodeaddress::ptr &recver,
        const sst_message::ptr &payload)
    {
      use_me();

      SSTMAC_DEBUG << "rdma_nic::send(" << recver->to_string() << ", "
            << payload->to_string() << ")\n";

      if (payload == 0)
      {
        std::cerr << "I'm here throwing this payload error" << std::endl;
        throw nullerror("simplenic::send: null payload pointer.");
      }

      if (spy_)
      {
        spy_->add_one(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id());
        spy_bytes_->add(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id(), payload->get_byte_length());
      }

      eventcallback::ptr cb = interconn_->get_injection_callback(payload);
      lib_mem_->nic_read(payload->get_byte_length(), payload->source_task(), cb);
    }

    void
    rdma_nic::recv_rdma_rsp(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::recv_rdma_rsp: "
            << get_parent_id()->to_string() << " for msg " << msg->to_string()
            << " for original message " << msg->get_parent_msg()->to_string()
            << (msg->is_tail() ? " -> found tail!" : " -> not yet tail")
            << "\n";
      //perform the memory access
      eventcallback::ptr cb = msg->is_tail() ? new_callback(
          rdma_nic::ptr(this), &rdma_nic::notify_session_manager,
          msg->get_parent_msg()) : donothingcallback::construct();

      //parent_->get_lib_mem()->rdma_write(msg->get_byte_length(), msg->dest_task(), cb);
      lib_mem_->rdma_put(msg->get_byte_length(),
          msg->get_parent_msg()->dest_task(), cb);
    }

    void
    rdma_nic::send_rdma_rsp(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::send_rdma_rsp: "
            << get_parent_id()->to_string() << " " << msg->to_string() << "\n";

      msg->mutate(sst_message::RDMA_GET_REQ_TO_RSP);

      /** pull the data in from memory and fire off to the requesting node */
      eventcallback::ptr cb = interconn_->get_injection_callback(msg);

      if (spy_)
      {
        spy_->add_one(msg->fromaddr()->unique_id(),
            msg->toaddr()->unique_id());
        spy_bytes_->add(msg->fromaddr()->unique_id(),
            msg->toaddr()->unique_id(), msg->get_byte_length());
      }

      lib_mem_->rdma_get(msg->get_byte_length(), msg->source_task(), cb);
    }

    void
    rdma_nic::recv(const sst_message::ptr &msg)
    {
      sst_message::ptr orig_msg = msg->get_parent_msg();

      SSTMAC_DEBUG << "rdma_nic::recv: " << get_parent_id()->to_string()
            << " payload message " << msg->to_string()
            << " for original message " << orig_msg->to_string()
            << (msg->is_tail() ? " -> found tail!" : " -> not yet tail")
            << "\n";

      eventcallback::ptr cb = msg->is_tail() ? new_callback(
          rdma_nic::ptr(this), &rdma_nic::notify_session_manager, orig_msg)
          : donothingcallback::construct();
      lib_mem_->nic_write(msg->get_byte_length(), orig_msg->dest_task(), cb);
    }

    void
    rdma_nic::notify_session_manager(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::notify_session_manager: "
            << get_parent_id()->to_string() << " " << msg->to_string() << "\n";
      get_parent_node()->handle(msg);
    }

    void
    rdma_nic::validate_msg_supported(const sst_message::ptr &msg) const
    {
      switch (msg->get_type())
      {
      case sst_message::MPI:
      case sst_message::HPX:
      case sst_message::SEND_ACK:
      case sst_message::ACTOR:
      case sst_message::NETWORK:
      case sst_message::RDMA_GET_REQ:
      case sst_message::RDMA_GET_RSP:
        return; //valid
      default:
        sst_throw(unimplementederror, "%s: network interface card received invalid message type %s",
            to_string().c_str(), sst_message::tostr(msg->get_type()));
      }
    }

  }
} // end of namespace sstmac.
