/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/simplenic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/waiter.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace hw
  {

    SSTRegister("simplenic | simple", networkinterface, simplenic,
                "simplest possible nic which only does basic read/write operations, more or less corresponding to ethernet + sockets");
    // Hello.
    //
    simplenic::ptr
    simplenic::construct(const sim_parameters::ptr& params)
    {
      return simplenic::ptr(new simplenic());
    }

    //
    // Goodbye.
    //
    simplenic::~simplenic() throw ()
    {
    }

    void
    simplenic::send(const nodeaddress::ptr &recver,
        const sst_message::ptr &payload)
    {
      use_me();

      validate_msg_supported(payload);

      eventcallback::ptr cb = new_callback(simplenic::ptr(this),
          &simplenic::send_to_network_link, payload, recver);

      lib_mem_->nic_read(payload->get_byte_length(), payload->source_task(), cb);
    }

    //
    // Schedule a send operation to happen at the given time.
    //
    void
    simplenic::send_to_network_link(const sst_message::ptr& payload,
        const nodeaddress::ptr& recver)
    {
      SSTMAC_DEBUG << "simplenic::send(" << recver->to_string() << ", "
            << payload->to_string() << ")\n";

      if (payload == 0)
        {
          std::cerr << "I'm here throwing this payload error" << std::endl;
          throw nullerror("simplenic::send: null payload pointer.");
        }

      timestamp delayed = now(); //don't model delay for now


      SSTMAC_DEBUG << "simplenic:send - sending now \n";

      interconn_->send(delayed, payload);

      spy_->add_one(payload->fromaddr()->unique_id(),
          payload->toaddr()->unique_id());
      spy_bytes_->add(payload->fromaddr()->unique_id(),
          payload->toaddr()->unique_id(), payload->get_byte_length());
      hist_msg_size_->add_one(payload->get_byte_length());
    }

    networkinterface::ptr
    simplenic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      simplenic::ptr ret = simplenic::construct(params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    simplenic::handle(const sst_message::ptr &msg)
    {
      validate_msg_supported(msg);

      if (msg->get_type() == sst_message::SEND_ACK)
      {
          schedule(now(), parent_, msg);
      }
      else if (msg->is_tail())
      {
          /** once the mem access is complete, send the parent msg
              on to the OS */
          sst_message::ptr parent_msg = msg->get_parent_msg();
          eventcallback::ptr cb = new_callback(simplenic::ptr(this),
                                        &simplenic::send_to_os, parent_msg);
          /** We now have to transfer the message into memory */
          lib_mem_->nic_write(msg->get_byte_length(), parent_msg->dest_task(), cb);
      }
      else //not yet the tail message
      {
          /** We now have to transfer the message into memory */
          eventcallback::ptr cb = donothingcallback::construct();
           sst_message::ptr parent_msg = msg->get_parent_msg();
          lib_mem_->nic_write(msg->get_byte_length(), parent_msg->dest_task(), cb);
      }
    }

    void
    simplenic::send_to_os(const sst_message::ptr &msg)
    {
      parent_->handle(msg);
    }

  }
} // end of namespace sstmac.
