#include <sstmac/hardware/packettrain/packettrain.h>

namespace sstmac
{
  namespace hw
  {

    debug packettrain::dbg_;
    uint64_t packettrain::min_num_bytes_per_train_ = 0;
    uint64_t packettrain::max_num_trains_ = 0;
    uint64_t packettrain::id_count_ = 0;
    uint64_t packettrain::negligible_size_ = 0;

    std::string
    train_metadata::to_string() const
    {
      return sstprintf("train metadata %d for %lu bytes", msg_id,
          num_bytes_total);
    }

    packettrain::packettrain(const sst_message::ptr &parent) :
      sst_message(parent->get_type()), mdata_(0), input_bw_(uninitialized_bw),
          num_bytes_(0), byte_offset_(0), next_pending_(0),
        t_arrive_(0), vc_allocated_(false), congestion_delay_us_(0)
    {
      needs_ack_ = false;
      if (min_num_bytes_per_train_ == 0)
      {
        init_statics();
      }

      bool tail_needs_ack = false;
      mdata_ = new train_metadata;
      mdata_->num_bytes_total = parent->get_byte_length();
      num_bytes_ = mdata_->num_bytes_total;
      mdata_->msg_id = id_count_++;
      mdata_->parent = parent;
      //mdata_->rinfo = 0;
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> train");
      }

      if (parent->is_network_message())
      {
        rinfo_ = routing_info::construct(parent->toaddr());
        tail_needs_ack = parent->get_needs_ack();
      }

      // now determine if we need to split this up into multiple trains
      uint64_t num_trains =
          divide_and_round_up(num_bytes_, min_num_bytes_per_train_);
      if (num_trains == 1)
      {
        needs_ack_ = tail_needs_ack;
        return;
      }

      num_trains = std::min(num_trains, max_num_trains_);
      uint64_t num_bytes_per_train = num_bytes_ / num_trains;
      uint64_t num_heavy_trains = num_bytes_ % num_trains;
      uint64_t num_light_trains = num_trains - num_heavy_trains;

      num_bytes_ = num_bytes_per_train;
      uint64_t byte_offset = num_bytes_;
      packettrain* prev = this;
      for (uint64_t i = 1; i < num_light_trains; ++i, byte_offset
          += num_bytes_per_train)
      {
        packettrain* next = new packettrain(mdata_, num_bytes_per_train,
            byte_offset, input_bw_);
        prev->next_pending_ = next;
        prev = next;
      }

      num_bytes_per_train++;
      for (uint64_t i = 0; i < num_heavy_trains; ++i, byte_offset
          += num_bytes_per_train)
      {
        packettrain* next = new packettrain(mdata_, num_bytes_per_train,
            byte_offset, input_bw_);
        prev->next_pending_ = next;
        prev = next;
      }

      packettrain* tail = prev;
      tail->set_needs_ack(tail_needs_ack);
    }

    void
    packettrain::throttle_bw(double bw)
    {
      set_initial_bw(bw);
      input_bw_ = std::min(bw, input_bw_);
      if (next_pending_)
        next_pending_->throttle_bw(bw);
    }

    void
    packettrain::add_next(const packettrain::ptr &train)
    {
      if (train.get() == this)
      {
        std::cerr << to_string() << std::endl;
        sst_throw(illformederror, "packettrain::add_next: adding self");
      }
      if (next_pending_)
        next_pending_->add_next(train);
      else
        next_pending_ = train;
    }

    void
    packettrain::set_input_bw(double bw)
    {
      //if (input_bw_ > 0 && bw > input_bw_){
      //    sst_throw(valueerror, "cannot reconfigure bw=%8.4e to a higher bandwidth %8.4e",
      //              input_bw_, bw);
      //}
      input_bw_ = bw;
    }

    void
    packettrain::init_statics()
    {
      //initialize the static variables
      min_num_bytes_per_train_ = sstmac_env::params->get_optional_int_param(
          "train_min_num_bytes", 256);
      max_num_trains_ = sstmac_env::params->get_optional_int_param(
          "train_max_num_messages", 10);
      uint64_t negligible_size = sstmac_env::params->get_optional_long_param("train_negligible_size", 512);
      negligible_size_ = std::min(negligible_size, min_num_bytes_per_train_);
    }

    packettrain::packettrain(const train_metadata::ptr& mdata,
        uint64_t num_bytes, uint64_t offset, double bw) :
      sst_message(mdata->parent->get_type()), mdata_(mdata), input_bw_(bw),
          num_bytes_(num_bytes), byte_offset_(offset), next_pending_(0),
        vc_allocated_(false), congestion_delay_us_(0)
    {
      needs_ack_ = false;

      if (mdata->parent->is_network_message())
      {
        rinfo_ = routing_info::construct(mdata->parent->toaddr());
      }
    }

    void
    packettrain::finalize_congestion_delay()
    {
        mdata_->parent->add_delay(timestamp(1e-6*congestion_delay_us_));
    }

    packettrain::ptr
    packettrain::construct(const sst_message::ptr &parent)
    {
      return new packettrain(parent);
    }

    packettrain::ptr
    packettrain::pop_next()
    {
      packettrain::ptr ret = next_pending_;
      next_pending_ = 0;
      return ret;
    }

    packettrain::~packettrain()
    {
    }

    std::string
    packettrain::to_string() const
    {
      return sstprintf("train %d, %lu->%lu:%lu at bw %8.4e next=%p",
          msg_id(), byte_offset_, byte_offset_ + num_bytes_,
          num_bytes_total(), input_bw_, next_pending_.get());
    }


  }
}
