#ifndef TRAINSWITCH_CC
#define TRAINSWITCH_CC

/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>
#include <sstmac/hardware/nic/train_nic.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>

namespace sstmac
{
  namespace hw
  {
    ImplementFactory(trainswitch);
    SSTRegister("train | trainswitch", trainswitch, trainswitch);

    trainswitch::trainswitch(double freq_hz, uint64_t switch_latency_cycles,
        double bytes_per_cycle, double payload_fraction, int num_tiles,
        bool enable_multipath, bool ignore_congestion, bool vis) :
          traincongestionpoint(bytes_per_cycle * num_tiles, freq_hz,
              payload_fraction, "switch", ""),
          latency_(switch_latency_cycles / freq_hz),
          bytes_per_cycle_(bytes_per_cycle),
          switch_latency_cycles_(switch_latency_cycles), num_tiles_(num_tiles),
          tile_bw_(freq_hz * bytes_per_cycle * payload_fraction), vis_(vis)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> trainswitch | network");
        logger::register_logger(&dbg_);
      }
      arbiter_ = new throttle_bw_arbitrator();
      arbiter_->set_bw(bw_bytes_);
      vc_alloc_ = new uncongested_vc_allocator;

      enable_multipath_ = enable_multipath;
      ignore_congestion_ = ignore_congestion;
    }

    trainswitch::ptr
    trainswitch::construct(const sim_parameters::ptr& params)
    {
      /**
        sstkeyword {
            gui=800mhz;
            docstring=$train_freq;
        }
      */
      double freq_hz = params->get_freq_param("network_train_switch_freq");
      /**
        sstkeyword {
            gui=30;
            docstring=$train_latency;
        }
      */
      uint64_t switch_latency_cycles = params->get_long_param(
          "network_train_switch_latency_cycles");
      /**
        sstkeyword { gui=2; docstring=$bytes_per_cycle; }
      */
      double bytes_per_cycle = params->get_double_param(
          "network_train_switch_bytes_per_cycle");

      double payload_fraction = params->get_optional_double_param(
          "network_train_payload_fraction", 1.0);
      /**
        sstkeyword {
            docstring=[NOT YET IMPLEMENTED] ENDL The total
                      number of subswitch tiles.
                      Routers with many tiles ease congestion by offloading
                      arbitration logic to many different tiles.;
        }
      */
      int num_tiles = params->get_optional_int_param(
          "network_train_switch_num_tiles", 1);
      /**
        sstkeyword { docstring=$ignore_congestion; }
      */
      bool ignore_congestion = params->get_optional_bool_param(
          "network_train_switch_ignore_congestion", false);
      /**
        sstkeyword { docstring=$enable_multipath; }
      */
      bool enable_multipath = params->get_optional_bool_param(
          "network_train_switch_enable_multipath", false);

      bool vis = params->has_param("vis_engine");

      return new trainswitch(freq_hz, switch_latency_cycles,
          bytes_per_cycle, payload_fraction, num_tiles, enable_multipath,
          ignore_congestion, vis);
    }

    void
    trainswitch::ack(const timestamp &t, const packettrain::ptr &train)
    {
      //do nothing
    }

    void
    trainswitch::init(router::ptr route, nodeaddress::ptr addr)
    {
      my_addr_ = addr;
      router_ = route;
    }

    trainswitch::~trainswitch()
    {
    }

    void
    trainswitch::finish()
    {
      if (finished_)
        return;

      traincongestionpoint::finish();

      {
        boost::unordered_map<int, boost::unordered_map<int, trainlink::ptr> >::iterator
            it, end = outports_.end();
        for (it = outports_.begin(); it != end; it++)
        {
          boost::unordered_map<int, trainlink::ptr>::iterator it2, end2 =
              it->second.end();
          for (it2 = it->second.begin(); it2 != end2; it2++)
          {
            trainlink::ptr link = it2->second;
            if (link)
            {
              link->finish();
            }
            else
            {
              sst_throw(ssterror, "at trainswitch::finish(): null link at %d, %d", it->first, it2->first);
            }
          }
          it->second.clear();
        }
      }
      outports_.clear();

      {
        boost::unordered_map<nodeaddress::ptr, eventmodule::ptr>::iterator it,
            end = ejectors_.end();
        for (it = ejectors_.begin(); it != end; ++it)
        {
          it->second->finish();
        }
      }
      ejectors_.clear();

      router_ = 0;

    }

    void
    trainswitch::connect(int dim, int dir, trainswitch::ptr sw,
        const timestamp& link_latency, double bytes_per_cycle,
        double link_freq_hz, double payload_fraction, bool enable_multipath,
        bool ignore_congestion, const output_bw_arbitrator::ptr& arb,
        const std::string& link_type, const std::string& link_name)
    {
      outports_[dim][dir] = trainlink::construct(sw, link_latency,
          bytes_per_cycle, link_freq_hz, payload_fraction, enable_multipath,
          ignore_congestion, arb, link_type, link_name);
    }

    void
    trainswitch::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);
      boost::unordered_map<int, boost::unordered_map<int, trainlink::ptr> >::iterator
          it, end = outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        boost::unordered_map<int, trainlink::ptr>::iterator it2, end2 =
            it->second.end();
        for (it2 = it->second.begin(); it2 != end2; it2++)
        {
          trainlink::ptr link = it2->second;
          if (link)
          {
            link->set_eventmanager(m);
          }
        }
      }

    }

    void
    trainswitch::add_ejector(const nodeaddress::ptr& addr,
        const trainpath::ptr& dest)
    {
      SSTMAC_DEBUG2 << sstprintf("Adding ejector at node %s to switch at %s\n",
            addr->to_string().c_str(), my_addr_->to_string().c_str());
      ejectors_[addr] = dest;
    }

    void
    trainswitch::update_router_cong()
    {
      boost::unordered_map<int, boost::unordered_map<int, trainlink::ptr> >::iterator
          it, end = outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        boost::unordered_map<int, trainlink::ptr>::iterator it2, end2 =
            it->second.end();
        for (it2 = it->second.begin(); it2 != end2; it2++)
        {
          router_->set_congestion(
              router_->convert_to_outport(it->first, it2->first),
              it2->second->next_free());
        }
      }
    }

    void
    trainswitch::do_send(const timestamp &t_start,
        const packettrain::ptr& train)
    {
      train->throttle_bw(bw_bytes_);

      if (router_->needs_congestion_info())
      {
        update_router_cong();
      }

      int dim;
      int dir;
      router_->route(train->rinfo(), dim, dir);
      timestamp arrive = now() + latency_;
      if (dim == networkswitch::EJECT_PORT)
      {
        schedule(arrive, ejectors_[train->toaddr()], train);
      }
      else
      {
        if (vis_)
        {
          if (linkvis_[dim][dir])
          {

            double act = outports_[dim][dir]->get_util_bw();
            linkvis_[dim][dir]->set_activity(act);
            vis_parent_->vis_update(now());
          }

        }

        schedule(arrive, outports_[dim][dir], train);
      }
    }

    void
    trainswitch::handle(const sst_message::ptr &msg)
    {
      packettrain::ptr train = safe_cast<packettrain> (msg);
      do_send(now(), train);
    }

    std::string
    trainswitch::to_string() const
    {
      return "train switch";
    }

  }
}

#endif // TRAINSWITCH_CC
