#ifndef TRAINSWITCH_H
#define TRAINSWITCH_H

#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/hardware/network/topology/router.h>
#include <sstmac/hardware/packettrain/trainstation.h>
#include <sstmac/hardware/packettrain/trainlink.h>
#include <sstmac/hardware/packettrain/congestion.h>

#include <sstmac/common/vis/vis.h>

namespace sstmac
{
  namespace hw
  {
    class train_nic;
    typedef boost::intrusive_ptr<train_nic> train_nicPtr;

    /**
     @class trainswitch
     A switch in the network that arbitrates/routes packettrains
     to the next link in the network
     */
    class trainswitch :
            public traincongestionpoint,
            public factory_type,
            public vis::vis_comp
    {

    protected:
      trainswitch(double freq_hz, uint64_t switch_latency_cycles,
          double bytes_per_cycle, double payload_fraction, int num_tiles,
          bool enable_multipath, bool ignore_congestion, bool vis);

      nodeaddress::ptr my_addr_;
      router::ptr router_;
      timestamp latency_;
      double bytes_per_cycle_;
      uint64_t switch_latency_cycles_;
      int num_tiles_;
      double tile_bw_;
      bool vis_;

      boost::unordered_map<nodeaddress::ptr, eventmodule::ptr> ejectors_;

      boost::unordered_map<int, boost::unordered_map<int, trainlink::ptr> >
          outports_;

      //      dim            dir    vis obj
      nested_type<int, int, vis::vis_obj::ptr>::unordered_map linkvis_;

      /**
       Implementation of virtual method from parent class.  This performs
       the routing operation and sends the packet train off to the next
       link.
       @param t_start The (absolute) simulation time in seconds to begin the send
       @param train The train to send
       */
      void
      do_send(const timestamp& t_start, const packettrain::ptr& train);

      /**
       Loop through all outport and collect congestion info.
       Pass that info along to the router for adaptive routing.
       */
      void
      update_router_cong();

    public:
      typedef boost::intrusive_ptr<trainswitch> ptr;

      vis::vis_display::ptr vis_parent_;

      static ptr
      construct(const sim_parameters::ptr& params);

      /**
       Following the creation of this switch via #clone,
       initialize the router and its network node address.
       @param route Switch router
       @param addr Switch network node address
       */
      virtual void
      init(router::ptr route, nodeaddress::ptr addr);

      /**
       Create an exact copy of this switch, but with different router and network node address
       @param The router for the cloned switch
       @param The network node address for the cloned switch
       @return The new trainswitch
       */
      virtual trainswitch::ptr
      clone(router::ptr route, nodeaddress::ptr addr) const
      {
        ptr ret = new trainswitch(freq_hz_, switch_latency_cycles_,
            bytes_per_cycle_, payload_fraction_, num_tiles_, enable_multipath_,
            ignore_congestion_, vis_);
        ret->init(route, addr);
        return ret;
      }

      /**
       Create a new link connection on a given outport for this switch
       @param dim The dimension (x,y,z,etc) the link will go
       @param dir The direction (+/-) the link will go
       @param link_latency The hop latency for the link
       @param bytes_per_cycle The number of bytes the link can send each clock cycle
       @param link_freq_hz The number of cycles per second (in GHz).  Bandwidth is
       bytes_per_cycle * link_freq
       @param payload_fraction The fraction (0.0-1.0) of bytes per packet
       that comprise the payload as opposed to the
       routing metadata overhead
       @param enable_multipath Whether the link should allow simultaneous packet trains
       from the same message.  See #inorder_vc_allocator
       vs #uncongested_vc_allocator.
       @param ignore_congestion Whether the link should monitor congestion
       and monitor bandwidth accordingly
       @param arbtype           A string identifying the type of output_bw_arbitrator.
       See network_train_link_arbitrator (and related) in keywords
       reference for allowed values.
       @param link_type         A string identifying the type of link.  Common values are
       network, memory, injection.
       @param link_name         A unique name for the link, e.g. for network link in 3D torus
       name might be "[1 0 3]->[1 0 4]" to identify +Z link for
       switch at x=1,y=0,z=3.

       */
      virtual void
      connect(int dim, int dir, trainswitch::ptr sw,
          const timestamp& link_latency, double bytes_per_cycle,
          double link_freq_hz, double payload_fraction, bool enable_multipath,
          bool ignore_congestion, const output_bw_arbitrator::ptr& arb_template,
          const std::string& link_type, const std::string& link_name = "");

      /**
       Cast message and pass to #send
       @param msg Incoming message (should cast to packettrain)
       */
      void
      handle(const sst_message::ptr &msg);

      /**
       @return The network node address
       */
      nodeaddress::ptr
      get_addr() const
      {
        return my_addr_;
      }

      router::ptr
      get_router() const
      {
        return router_;
      }

      virtual void
      set_link_vis(vis::vis_obj::ptr vo, int dim, int dir)
      {

        linkvis_[dim][dir] = vo;
      }

      /**
       Set the link to use when ejecting packets at their endpoint.  A trainswitch
       can have any number of ejectors, corresponding to the number of nodes
       per switch.
       @param addr The compute node address of the endpoint to eject to
       @param link The link to the compute node for ejection
       */
      void
      add_ejector(const nodeaddress::ptr& addr, const trainpath::ptr& link);

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      /**
       Send a network ack when the train leaves.  Currently, switch
       does nothing.
       @param t The time the ack can be sent
       @param train The train being sent that can be acked
       */
      void
      ack(const timestamp &t, const packettrain::ptr &train);

      virtual void
      finish();

      virtual std::string
      to_string() const;

      virtual
      ~trainswitch();

    };

    DeclareFactory(trainswitch);
  }
}

#endif // TRAINSWITCH_H
