/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac.h>

#include <mpi.h>

#include <sstmac/common/fileio.h>
#include <sys/time.h>
#include <unistd.h>

using namespace sstmac;
//using namespace sstmac::sw;
using namespace sstmac::hw;
using namespace sstmac::native;


std::ofstream multilog_;

//                   pid   run#
boost::unordered_map<int, int> pidruns;

//                   pid  proc#
boost::unordered_map<int, int> pidprocs;


static inline double
get_real_time()
{
  timeval t_st;
  gettimeofday(&t_st, 0);
  double t = t_st.tv_sec + 1e-6 * t_st.tv_usec;
  return t;
}

void
child_done(int snum)
{
  int pid;
  int status;

  pid = wait(&status);

  std::cout << "---------- Run " << pidruns[pid]
      << " has finished ------------ \n";

  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (rank != 0)
  {
    int b = pidruns[pid];
    MPI_Send(&b, 1, MPI_INT, 0, pidprocs[pid], MPI_COMM_WORLD);
  }
  pidprocs.erase(pid);
  pidruns.erase(pid);

}


sstmac::sim_parameters::ptr
fork_a_run(int runid, int procid, sstmac::sim_parameters::ptr params,
     std::map<int, std::string> &keys,
     std::vector<std::string> &run, bool &goforit)
{
  //set signal handler
  // signal(SIGCHLD, child_done);

  int pid;
  switch (pid = fork())
  {
  case 0: /* a fork returns 0 to the child */
  {
    sim_parameters::ptr clone = params->clone();
    for (int i = 0; i < (int) run.size(); i++)
    {
      clone->add_param_override(keys[i], run[i]);
    }

    std::stringstream ss;
    ss << "sim_output-" << runid << ".out";

    //std::ofstream myfile;

    multilog_.open(ss.str().c_str(), std::ios::out);

    if (!multilog_.is_open())
    {
      sst_throw(ioerror, "statlogger: cannot open file %s for writing", ss.str().c_str());
    }

    //_multisim_logs_[it2->first] = myfile;

    //debug::debug_stream_ = &_multisim_logs_[it2->first];
    debug::debug_stream_ = &multilog_;
    goforit = true;
    return clone;
  }

  default: /* a fork returns a pid to the parent */
    pidruns[pid] = runid;
    pidprocs[pid] = procid;
    break;

  case -1: /* something went wrong */
    perror("fork");
    exit(1);
  }

  goforit = false;
  return sstmac::sim_parameters::ptr();
}


sstmac::sim_parameters::ptr
do_multisim_mpi(sstmac::sim_parameters::ptr params,
    std::map<int, std::string> &keys,
    std::map<int, std::vector<std::string> > &runs, int mpirank)
{
 // int num_free_procs = params->get_optional_int_param("multisim_nproc", 1);
  int num_free_procs = 1;

  if (mpirank == 0)
  {
    std::cout << "------------------------------------- \n";
    std::cout << "-> About to run " << num_free_procs
        << " simulations at a time on each rank, press enter to continue \n";
  }
  if (!params->has_param("multisim_nproc"))
  {
    if (mpirank == 0)
    {
      std::cout
          << "   (Warning: you have not set the multisim_nproc parameter, 1 is default)\n";
    }
  }

  if (mpirank == 0)
  {
    std::cin.ignore();
  }

  MPI_Barrier(MPI_COMM_WORLD);

  int mpisize;
  MPI_Comm_size(MPI_COMM_WORLD, &mpisize);


  int masterruncount = 0;
  int rundummy;

  if (mpirank == 0)
  {

    //first, fork my own and get them going
    for (int z = 0; z < num_free_procs && masterruncount < runs.size(); z++)
    {
      bool go;
      sstmac::sim_parameters::ptr p = fork_a_run(masterruncount, z, params,
          keys, runs[masterruncount], go);

      if (go)
      { //this is the forked child
        return p;
      }
      masterruncount++;
    }

    //                   rank                      proc    req
    MPI_Request reqs[(mpisize - 1) * num_free_procs];

    //initialize correctly
    for (int z = 0; z < (mpisize - 1) * num_free_procs; z++)
    {
      reqs[z] = MPI_REQUEST_NULL;
    }

    int numwaiting = 0;

    std::cout << "master: spawning out, comm size is  " << mpisize << "\n";

    for (int i = 1; i < mpisize && masterruncount < runs.size(); i++)
    {
      std::cout << "master: considering rank " << i << "\n";

      for (int j = 0; j < num_free_procs && masterruncount < runs.size(); j++)
      {
        std::cout << "master: giving run " << masterruncount << " to rank " << i << " on proc " << j << "\n";
        MPI_Request req;
        MPI_Isend(&masterruncount, 1, MPI_INT, i, j, MPI_COMM_WORLD, &req);
        masterruncount++;

        MPI_Irecv(&rundummy, 1, MPI_INT, i, j, MPI_COMM_WORLD, &req);

        reqs[(i - 1) * num_free_procs + j] = req;
        numwaiting++;
      }
    }

    //keep spawning more when necessary
    while (masterruncount < runs.size() && numwaiting > 0)
    {
      int ind;
      MPI_Status stat;
      MPI_Waitany((mpisize - 1) * num_free_procs, reqs, &ind, &stat);


      int donerank = (ind / num_free_procs) + 1;
      int doneproc = ind % num_free_procs;

      std::cout << "master: got back index " << ind << ", which is donerank " << donerank  <<
              " and doneproc " << doneproc << "\n";

      MPI_Request req;
      MPI_Isend(&masterruncount, 1, MPI_INT, donerank, doneproc, MPI_COMM_WORLD, &req);
      masterruncount++;

      MPI_Irecv(&rundummy, 1, MPI_INT, donerank, doneproc, MPI_COMM_WORLD, &req);

      reqs[ind] = req;
    }

    //finish by sending all the ranks an exit code
    std::cout << "master: all done, cleaning up \n";
    int exit = -1;
    for (int i = 1; i < mpisize; i++)
    {
      MPI_Request req;

      MPI_Isend(&exit, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &req);
    }

  }
  else
  {
    int* runtodo = (int*) malloc(sizeof(int) * num_free_procs);
    MPI_Request reqs[num_free_procs];

    for (int j = 0; j < num_free_procs; j++)
    {
      MPI_Request req;
      MPI_Irecv(&runtodo[j], 1, MPI_INT, 0, j, MPI_COMM_WORLD, &req);
      reqs[j] = req;
    }

    int activerun;
    do
    {
      MPI_Status stat;
      int index;
      MPI_Waitany(num_free_procs, reqs, &index, &stat);
      activerun = runtodo[index];
      if (activerun >= 0)
      {
        std::cout << "--- rank " << mpirank << " is running #" << activerun << " \n";

        bool go;
        sstmac::sim_parameters::ptr p = fork_a_run(activerun, index, params,
            keys, runs[activerun], go);

        if (go)
        { //this is the forked child
          return p;
        }

        child_done(0);

        //i'm the parent, wait for another
        MPI_Request req;
        MPI_Irecv(&runtodo[index], 1, MPI_INT, 0, index, MPI_COMM_WORLD, &req);
        reqs[index] = req;
      }
    }
    while (activerun >= 0);

    std::cout << "--- rank " << mpirank << " got the exit code, exiting \n";

  }

  MPI_Finalize();
  exit(0);
}

static int
try_mpisweepmain(int argc, char **argv, int mpirank)
{
  double start = get_real_time();

  //set up the search path
  SSTFileIO::init();

  bool doingmultisim = false;

  opts oo;
  int parse_status = parse_opts(argc, argv, oo);
  if (parse_status == PARSE_OPT_SUCCESS)
  {
    if (oo.print_apps)
    {
      SSTAppFactory::print_apps();
      return 0;
    }

    if (oo.debug != "")
      logger::set_user_param(oo.debug);

    if (oo.dumpimetafile != "")
    {
      std::cout << "Running parsedumpi on metafile " << oo.dumpimetafile
          << std::endl;
      if (oo.network == "")
      {
        std::cerr
            << "ERROR: parsedumpi needs a network (e.g. hopper, franklin) specified with the -n option"
            << std::endl;
        return 1;
      }
      oo.configfile = "parsedumpi.ini";
    }

    if (oo.color)
      logger::set_debug_coloring(true);

    if (oo.configfile == "")
    {
      std::cerr << "ERROR: need a configuration file" << std::endl;
      return 1;
    }

    //use the config file to set up file search paths
    size_t pos = oo.configfile.find_last_of('/');
    if (pos != std::string::npos)
    {
      std::string dir = oo.configfile.substr(0, pos + 1);
      SSTFileIO::add_path(dir);
    }

    file_parameters::ptr fparams = file_parameters::construct(oo.configfile);
    if (oo.network != "")
    {
      std::string machinefile = oo.network + ".ini";
      fparams->append_params(machinefile);
    }

    sim_parameters::ptr params = fparams;



    if (oo.params)
    { // there were command-line overrides
      oo.params->combine_into(params);
    }

    if (params->is_multi_sim())
    {

      std::map<int, std::string> keys;
      std::map<int, std::vector<std::string> > runs;
      multisim_init(params, keys, runs, (mpirank == 0));
      params = do_multisim_mpi(params, keys, runs, mpirank);
      doingmultisim = true;
    }
    else if (mpirank != 0)
    {
      std::cout
          << "WARNING: mpisweep is being used, but multiple parameter combinations were not found. Rank "
          << mpirank << " is exiting without doing anything \n";
      return 0;
    }

    //at this point, we have read in parameters - init malloc system
    //set the global parameters object
    sstmac_env::params = params;

    sstmac_init_cxx_heap();

    if (oo.dumpimetafile != "")
    {
      params->add_param_override("launch_dumpi_metaname", oo.dumpimetafile);
      //count the number of applications to launch
      std::ifstream in;
      in.open(oo.dumpimetafile.c_str());

      if (!in.is_open()) //see if the file is included in the include path
      {
        std::cerr << "ERROR: " << oo.dumpimetafile
            << " is not a valid dumpi meta file" << std::endl;
        return 1;
      }

      std::string line;
      while (in.good())
      {
        std::getline(in, line);
        if (line.find("=") != std::string::npos)
        { //an assignment
          std::pair<std::string, std::string> keyval =
              sim_parameters::parse_param(line);
          if (keyval.first == "numprocs")
            params->add_param_override("launch_app1_size", keyval.second);
        }
      }
    }

    /** initialize static variables */
    nodeid::init_statics();

    long timescale = params->get_optional_long_param("timestamp_resolution", 1);
    timestamp::init_stamps(timescale);

    if (params->has_param("logger_params"))
    {
      std::string log_params = params->get_param("logger_params");
      logger::set_user_param(log_params);
    }

    if (params->has_param("stat_time_interval"))
    {
      stat_time_window::set_time_interval(
          params->get_time_param("stat_time_interval"));
    }

    if (oo.mode == "normal" || oo.mode == "notime")
    {
      try
      {
        run(params, oo.stoptime, (oo.mode != "notime"));
      }
      catch (const std::exception &e)
      {
        std::cout << argv[0]
            << ": caught exception while running simulation:\n" << e.what()
            << "\n";
        return 1;
      }
      catch (...)
      {
        std::cout << argv[0]
            << ": caught unknown exception while running simulation\n";
        return 1;
      }

      // now that we finished running, print the parameters that we used
      if (oo.print_params)
        params->print_params();
    }
    else if (oo.mode == "debug")
    {
      run_debugger(params, oo);
    }
  }
  else if (parse_status == PARSE_OPT_EXIT_SUCCESS)
  {
    return 0;
  }
  else if (parse_status == PARSE_OPT_EXIT_FAIL)
  {
    return 1;
  }

  double stop = get_real_time();
  double wall_time = stop - start;
  if (oo.print_walltime)
    std::cout << sstprintf("SST/macro ran for %12.4f seconds\n", wall_time);

  clear_statics();

  sstmac_finalize_cxx_heap();

  if(doingmultisim){
    exit(0);
  }

  return 0;
}

int
main(int argc, char **argv)
{

  MPI_Init(&argc, &argv);
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  int rc;

  try
  {
    rc = try_mpisweepmain(argc, argv, rank);
  }
  catch (const std::exception &e)
  {
    std::cout << argv[0] << ": caught exception while setting up simulation:\n"
        << e.what() << "\n";
    return 1;
  }
  catch (...)
  {
    std::cout << argv[0]
        << ": caught unknown exception while setting up simulation\n";
    return 1;
  }

  MPI_Finalize();

  return rc;
}

