/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac.h>

#include <sstmac/common/fileio.h>
#include <sstmac/common/malloc.h>

#include <sys/types.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/wait.h>

using namespace sstmac;
using namespace sstmac::hw;
using namespace sstmac::native;

int _multisim_nprocs_ = 0;
//type<int, std::ostream>::map _multisim_logs_;
std::ofstream multilog_;

static inline double
get_real_time()
{
  timeval t_st;
  gettimeofday(&t_st, 0);
  double t = t_st.tv_sec + 1e-6 * t_st.tv_usec;
  return t;
}

void
child_done(int snum)
{
  int pid;
  int status;

  pid = wait(&status);
  std::cout << "---------- Child simulation has finished ------------ \n";
  _multisim_nprocs_++;
}

sstmac::sim_parameters::ptr
do_multisim_fork(sstmac::sim_parameters::ptr params,
    std::map<int, std::string> &keys,
    std::map<int, std::vector<std::string> > &runs)
{
  _multisim_nprocs_ = params->get_optional_int_param("multisim_nproc", 1);
  std::cout << "------------------------------------- \n";
  std::cout << "-> About to run " << _multisim_nprocs_
      << " simulations at a time, press enter to continue \n";
  if (!params->has_param("multisim_nproc"))
  {
    std::cout
        << "   (Warning: you have not set the multisim_nproc parameter, 1 is default)\n";
  }

  std::cin.ignore();

  //get busy
  signal(SIGCHLD, child_done);

  for (int r = 0; r < runs.size(); r++)
  {
    int pid;

    _multisim_nprocs_--;

    switch (pid = fork())
    {
    case 0: /* a fork returns 0 to the child */
    {
      sim_parameters::ptr clone = params->clone();
      for (int i = 0; i < (int) runs[r].size(); i++)
      {
        clone->add_param_override(keys[i], runs[r][i]);
      }

      std::stringstream ss;
      ss << "sim_output-" << r << ".out";

      //std::ofstream myfile;

      multilog_.open(ss.str().c_str(), std::ios::out);

      if (!multilog_.is_open())
      {
        sst_throw(ioerror, "statlogger: cannot open file %s for writing",
            ss.str().c_str());
      }

      //_multisim_logs_[it2->first] = myfile;

      //debug::debug_stream_ = &_multisim_logs_[it2->first];
      debug::debug_stream_ = &multilog_;

      return clone;
    }

    default: /* a fork returns a pid to the parent */

      break;

    case -1: /* something went wrong */
      perror("fork");
      exit(1);
    }

    while (_multisim_nprocs_ <= 0)
    {
      sleep(60);
    }
  }
  exit(0);
}

static int
try_sstmacmain(int argc, char **argv)
{
  double start = get_real_time();

  //set up the search path
  SSTFileIO::init();

  opts oo;
  int parse_status = parse_opts(argc, argv, oo);
  if (parse_status == PARSE_OPT_SUCCESS)
    {
      if (oo.print_apps)
        {
          sstmac::sw::app_factory::print_apps();
          return 0;
        }

      if (oo.debug != "")
        logger::set_user_param(oo.debug);

      if (oo.dumpimetafile != "")
        {
          std::cout << "Running parsedumpi on metafile " << oo.dumpimetafile
              << std::endl;
          if (oo.network == "")
            {
              std::cerr
                  << "ERROR: parsedumpi needs a network (e.g. hopper, franklin) specified with the -n option"
                  << std::endl;
              return 1;
            }
          oo.configfile = "parsedumpi.ini";
        }

      if (oo.color)
        logger::set_debug_coloring(true);

      if (oo.configfile == "")
        {
          std::cerr << "ERROR: need a configuration file" << std::endl;
          return 1;
        }

      //use the config file to set up file search paths
      size_t pos = oo.configfile.find_last_of('/');
      if (pos != std::string::npos)
        {
          std::string dir = oo.configfile.substr(0, pos + 1);
          SSTFileIO::add_path(dir);
        }

      file_parameters::ptr fparams = file_parameters::construct(oo.configfile);
      if (oo.network != "")
        {
          std::string machinefile = oo.network + ".ini";
          fparams->append_params(machinefile);
        }

      sim_parameters::ptr params = fparams;

      if (oo.params)
        { // there were command-line overrides
          oo.params->combine_into(params);
        }

      if (params->is_multi_sim())
        {
          std::map<int, std::string> keys;
          std::map<int, std::vector<std::string> > runs;
          multisim_init(params, keys, runs, true);
          params = do_multisim_fork(params, keys, runs);
        }

      //at this point, we have read in parameters - init malloc system
      //set the global parameters object
      sstmac_env::params = params;

      sstmac_init_cxx_heap();

      if (oo.dumpimetafile != "")
        {
          params->add_param_override("launch_dumpi_metaname", oo.dumpimetafile);
          //count the number of applications to launch
          std::ifstream in;
          in.open(oo.dumpimetafile.c_str());

          if (!in.is_open()) //see if the file is included in the include path
            {
              std::cerr << "ERROR: " << oo.dumpimetafile
                  << " is not a valid dumpi meta file" << std::endl;
              return 1;
            }

          std::string line;
          while (in.good())
            {
              std::getline(in, line);
              if (line.find("=") != std::string::npos)
                { //an assignment
                  std::pair<std::string, std::string> keyval;
                  sim_parameters::ptr scope = params->parse_param(line, keyval);
                  if (keyval.first == "numprocs")
                    scope->add_param_override("launch_app1_size", keyval.second);
                }
            }
        }

      /** initialize static variables */
      nodeid::init_statics();

      long timescale = params->get_optional_long_param("timestamp_resolution",
          1);
      timestamp::init_stamps(timescale);

      if (params->has_param("logger_params"))
        {
          std::string log_params = params->get_param("logger_params");
          logger::set_user_param(log_params);
        }

      if (params->has_param("stat_time_interval"))
        {
          stat_time_window::set_time_interval(
              params->get_time_param("stat_time_interval"));
        }

      if (oo.mode == "normal" || oo.mode == "notime")
        {
          try
            {
              run(params, oo.stoptime, (oo.mode != "notime"));
            }
          catch (const std::exception &e)
            {
              std::cout << argv[0]
                  << ": caught exception while running simulation:\n"
                  << e.what() << "\n";
              return 1;
            }
          catch (...)
            {
              std::cout << argv[0]
                  << ": caught unknown exception while running simulation\n";
              return 1;
            }

          // now that we finished running, print the parameters that we used
          if (oo.print_params)
            params->print_params();
        }
      else if (oo.mode == "debug")
        {
          run_debugger(params, oo);
        }
    }
  else if (parse_status == PARSE_OPT_EXIT_SUCCESS)
    {
      return 0;
    }
  else if (parse_status == PARSE_OPT_EXIT_FAIL)
    {
      return 1;
    }

  double stop = get_real_time();
  double wall_time = stop - start;
  if (oo.print_walltime)
    std::cout << sstprintf("SST/macro ran for %12.4f seconds\n", wall_time);

  clear_statics();

  sstmac_finalize_cxx_heap();

  return 0;
}

int
main(int argc, char **argv)
{
  int rc;

  try
    {
      rc = try_sstmacmain(argc, argv);
    }
  catch (const std::exception &e)
    {
      std::cout << argv[0]
          << ": caught exception while setting up simulation:\n" << e.what()
          << "\n";
      return 1;
    }
  catch (...)
    {
      std::cout << argv[0]
          << ": caught unknown exception while setting up simulation\n";
      return 1;
    }

  return rc;
}

