/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac.h>

#include <iostream>
#include <sstream>
#include <iterator>
#include <getopt.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <boost/lexical_cast.hpp>

#include <sstmac/common/fileio.h>
#include <sstmac/common/malloc.h>

#include <sstmac/common/basicstringtokenizer.h>

#include <sstmac/common/types.h>

#include <sys/types.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/wait.h>

using namespace sstmac;
//using namespace sstmac::sw;
using namespace sstmac::hw;
using namespace sstmac::native;

using boost::lexical_cast;

static std::ostream&
operator<<(std::ostream &os, const opts &oo)
{
  if (oo.help)
    os << "  --help \\\n";
  if (oo.color)
    os << "  --color \\\n";
  os << "  --debug=\"" << oo.debug << "\" \\\n";
  os << "  --configfile=\"" << oo.configfile << "\" \\\n";
  os << "  --mode=\"" << oo.mode << "\" \\\n";
  os << "  --stoptime=" << oo.stoptime.sec();
  return os;
}





int
parse_opts(int argc, char **argv, opts &oo)
{
  int dompitest = 0;
  int printnodes = 0;
  int no_wall_time = 0;
  int print_params = 0;
  option gopt[] =
  {
  { "help", no_argument, NULL, 'h' },
  { "include", required_argument, NULL, 'i' },
  { "debug", required_argument, NULL, 'd' },
  { "stoptime", required_argument, NULL, 't' },
  { "configfile", required_argument, NULL, 'f' },
  { "param", required_argument, NULL, 'p' },
  { "srun", required_argument, NULL, 's' },
  { "mode", required_argument, NULL, 'm' },
  { "color", no_argument, NULL, 'c' },
  { "print-apps", no_argument, NULL, 'P' },
  { "parse", required_argument, NULL, 'u' }, //u = undumpi
      { "network", required_argument, NULL, 'n' },
      { "mpitest", no_argument, &dompitest, 1 },
      { "print-nodes", no_argument, &printnodes, 1 },
      { "print-params", no_argument, &print_params, 1 },
      { "no-wall-time", no_argument, &no_wall_time, 1 },
      { NULL, 0, NULL, '\0' } };
  int ch;
  bool errorflag = false;
  std::list < std::pair<std::string, std::string> > paramlist;
  oo.params = command_parameters::construct();
  while ((ch = getopt_long(argc, argv, "Phcd:f:t:p:m:n:u:i:", gopt, NULL))
      != -1)
  {
    switch (ch)
    {
    case 0:
      //this set an input flag
      break;
    case 'h':
      oo.help = 1;
      break;
    case 'd':
      oo.debug = optarg;
      break;
    case 'f':
      oo.configfile = optarg;
      break;
    case 'u':
      oo.dumpimetafile = optarg;
      break;
    case 't':
      oo.stoptime = timestamp(atof(optarg));
      break;
    case 'n':
      oo.network = optarg;
      break;
    case 'P':
      oo.print_apps = true;
      break;
    case 'i':
    {
      file_parameters::ptr params = file_parameters::construct(optarg);
      params->combine_into(oo.params);
    }
      break;
    case 'p':
    {
      std::pair < std::string, std::string > keyval;
      sim_parameters::ptr scope = oo.params->parse_param(optarg, keyval);
      paramlist.push_back(keyval);
      scope->add_param(keyval.first, keyval.second);
      break;
    }
    case 'm':
      oo.mode = optarg;
      break;

    case 'c':
      oo.color = true;
      break;

    default:
      std::cerr << "Unhandled input flag.\n";
      errorflag = true;
      std::cout << "optarg: " << optarg << std::endl;
      break;
    }
    if (oo.help)
    {
      print_help(argc, argv);
      return PARSE_OPT_EXIT_SUCCESS;
    }
    if (errorflag)
    {
      std::cerr << "Error return\n";
      return PARSE_OPT_EXIT_FAIL;
    }
  }

  if (print_params)
  {
    oo.print_params = true;
  }

  /** check to see if we should do an mpitestall */
  if (dompitest)
  {
    if (oo.configfile != "")
    {
      std::cerr
          << "Cannot set config file with --configfile and also request --mpitest"
          << std::endl;
      return PARSE_OPT_EXIT_FAIL;
    }
    oo.configfile = "mpitestall.ini";

    if (oo.network == "")
    {
      std::cerr
          << "For --mpitest, you must specify a network and node model with --network"
          << std::endl;
      return PARSE_OPT_EXIT_FAIL;
    }
  }
  if (printnodes)
  {
    oo.params->add_param_override("launch_app1", "mpi_printnodes");
  }

  /** double negative, sorry about that */
  oo.print_walltime = !no_wall_time;

  /** Set sensible defaults for some of the paramters if not set*/
  if (oo.configfile == "")
    oo.configfile = "parameters.ini";

  std::cerr << argv[0] << " \\\n" << oo;
  if (paramlist.size())
  {
    for (std::list<std::pair<std::string, std::string> >::iterator i =
        paramlist.begin(); i != paramlist.end(); i++)
    {
      std::cerr << " \\\n  --param \"" << i->first << "=" << i->second << "\"";
    }
  }
  std::cerr << "\n";

  return PARSE_OPT_SUCCESS;
}

void
run_debugger(sim_parameters::ptr params, opts oo)
{
  std::cout << "------------------------------------------ \n";
  std::cout << "--- Welcome to the SST/macro debugger ---- \n";
  std::cout << "------------------------------------------- \n";
  std::cout << "Type \"help\" for more information. \n";

  timestamp stop = oo.stoptime;
  std::string dbg_str = "timestamps | " + oo.debug;
  logger::set_user_param(dbg_str);
  bool exit = false;
  bool hasrunyet = false;
  eventmanager::ptr eventman;

  while (!exit)
  {
    std::cout << logger::red << "sstmac-debug >> " << logger::black;
    std::string command;
    std::cin >> command;

    if (command == "help" || command == "h")
    {
      std::cout << "Commands are: \n";
      std::cout << "  help (h): print this help list \n";
      std::cout << "  start (st): start the simulation \n";
      std::cout
          << "  stopat (sa) <value>: set a breakpoint at the given time \n";
      std::cout << "  exit: exit the debugger \n";
      std::cout << "  silly: print something silly \n";
      std::cout << "  debug (d) <value>: add <value> to the debug list \n";
      std::cout
          << "  debug-list (dl): see the list of things being debugged \n";
      std::cout << "  debug-clear (cl): clear the debug list \n";
      std::cout << "  now: print the current time of the simulation \n";
      std::cout << "  step|next (n): execute one event \n";
      std::cout << "  reset: reset the simulation so it can be started again\n";
    }
    else if (command == "start" || command == "st")
    {
      if (!hasrunyet)
        eventman = run(params, stop, false);
      else
        eventman->run();
      hasrunyet = true;
    }
    else if (command == "stopat" || command == "sa")
    {
      double time;
      std::cin >> time;
      stop = timestamp(time);
      std::cout << logger::green << "debug: should stop at " << stop << "\n";
      if (hasrunyet)
      {
        eventman->schedule(stop, manager::stophandler::construct(),
            sst_message::ptr());
      }
    }
    else if (command == "exit")
    {
      exit = true;
    }
    else if (command == "silly")
    {
      std::cout << "firetruck lightbulb purple-monkey-dishwasher \n";
    }
    else if (command == "debug" || command == "d")
    {
      std::string val;
      std::cin >> val;
      if (dbg_str == "")
      {
        dbg_str = "<debug> " + val;
      }
      else
      {
        dbg_str = dbg_str + " | <debug> " + val;
      }

      logger::set_user_param(dbg_str);
    }
    else if (command == "debug-list" || command == "dl")
    {
      std::cout << "debug list: " << dbg_str << "\n";
    }
    else if (command == "debug-clear" || command == "cl")
    {
      dbg_str = "timestamps";
      logger::set_user_param(dbg_str);
    }
    else if (command == "next" || command == "step" || command == "n")
    {
      if (!hasrunyet)
      {
        std::cout << "haven't started yet, have to start first \n";
      }
      else
      {
        eventman->step();
      }
    }
    else if (command == "now")
    {
      if (!hasrunyet)
        std::cout << "0 \n";
      else
        std::cout << eventman->now() << "\n";
    }
    else if (command == "reset")
    {

      stop = oo.stoptime;
      dbg_str = "timestamps | " + oo.debug;
      hasrunyet = false;
      logger::set_user_param(dbg_str);

    }
    else
    {

      std::cout << "unknown command: " << command << "\n";
    }
  }

  std::cout << "---------------------------------------------------------- \n";
  std::cout << "--- Exiting the SST/macro debugger normally, so long! ---- \n";
  std::cout << "---------------------------------------------------------- \n";

}

void
print_help(int, char **argv)
{
  char *name = argv[0];
  char *tail = name + strlen(name);
  while (tail != name && *tail != '/')
    --tail;
  if (tail != name)
    ++tail;

  std::cerr << "usage: " << tail << " -f <config file> [options] \n" << " \n"
      << "Options are: \n" << "\t[(--help|-h)]                          \n"
      << "\t[(--print-apps|-P)]                    \n"
      << "\t[(--color|-c)]                         \n"
      << "\t[(--debug|-d)             <value> ]    \n"
      << "\t[(--configfile|-f)        <value> ]    \n"
      << "\t[(--param|-p)       <key>=<value> ]    \n"
      << "\t[(--stoptime|-t)          <value> ]    \n"
      << "\t[(--mode|-m)              <value> ]    \n"
      << "\t[(--include|-i)           <value> ]    \n" << "\n"
      << "Configuration file is not optional. See parameters.ini for \n"
      << "a list of parameters. \n" << "\n"
      << "Turn ANSI console output coloring on with --color or -c, which is \n"
      << "particularly useful when debugging, but should not be used for make check \n"
      << "or redirecting to a file. \n" << "\n"
      << "--param (-p) is used to override any parameters in the \n"
      << "configuration file, for example: \n" << "\t-p network_name=buffer \n"
      << "\n--include (-i) is used to include a pre-made parameter file \n"
      << "from the configurations folder into your parameter set\n" << "\n"
      << "--stoptime (-t) stops the simulation at the given time, \n"
      << "which is useful for debugging \n" << "\n"
      << "--mode (-m) can be used to open the debugger by saying -m debug, \n"
      << "and can turn off printing the simulation at the end with -m notime \n"
      << "\n" << "Valid arguments to --debug (-d) are strings of the form \n"
      << "\"<(debug|stats)> (name1) | (name2) | ... \" \n"
      << "\t- examples: \n"
      << "\t\t- \"<debug> mpi\" turns on all mpi output \n"
      << "\t\t- \"<debug> app | mpicollective\" turns on all \n"
      << "\t\tskeleton and mpicollective output \n"
      << "\t\t- \"<stats> all\" turns on all stat collection \n"
      << "\t- some current useful debug tags are: \n"
      << "\t\tall sim app mpi mpiapi mpicollective mpiqueue os processor \n"
      << "\t\tnode network manager eventmanager allocation indexing \n"
      << "\t\tlaunch lib_compute lib_sleep flower \n"
      << "\t- current stat tags are: \n" << "\t\tall processor network \n";

}

void
make_run(std::map<int, std::deque<std::string> > &vars,
    std::map<int, std::vector<std::string> > &runs, std::vector<int> &indexes)
{
  // std::cout << "making run...";
  int next = runs.size();

  std::vector < std::string > vec(vars.size());
  for (int i = 0; i < (int) indexes.size(); i++)
  {
    std::deque<std::string>::const_iterator it, end = vars[i].end();
    int cnt = 0;
    for (it = vars[i].begin(); it != end && cnt < indexes[i]; it++)
    {
      cnt++;
    }

    vec[i] = *it;
  }

  runs[next] = vec;
  //std::cout << "done \n";
}

void
rec_run(std::map<int, std::deque<std::string> > &vars,
    std::map<int, std::vector<std::string> > &runs, std::vector<int> &indexes,
    int var)
{

  if (var == ((int) vars.size() - 1)) // at the end, make the run
  {
    for (int i = 0; i < (int) vars[var].size(); i++)
    {

      make_run(vars, runs, indexes);
      indexes[var]++;
    }
    indexes[var] = 0;
  }
  else
  {
    for (int i = 0; i < (int) vars[var].size(); i++)
    {

      rec_run(vars, runs, indexes, var + 1);
      indexes[var]++;
    }
    indexes[var] = 0;

  }

}

void
multisim_init(sstmac::sim_parameters::ptr params,
    std::map<int, std::string> &keys,
    std::map<int, std::vector<std::string> > &runs, bool print)
{
  if (print)
  {
    std::cout << "-------------------------------------- \n";
    std::cout << "----- SST/macro Multi-Sim Manager ---- \n";
    std::cout << "-------------------------------------- \n \n";

    std::cout << "We found the following multi-parameters: \n";
    printf("%-40s%-100s\n", "Name", "Value");
    //std::cout << "Name\t\t\t\t\tValue \n";
    std::cout
        << "-------------------------------------------------------------------- \n";
  }

  sstmac::sim_parameters::multisim_maptype p = params->get_multi_params();
  sstmac::sim_parameters::multisim_maptype::iterator it, end = p.end();
  std::map<int, std::deque<std::string> > parsed;

  int count = 0;
  for (it = p.begin(); it != end; it++)
  {
    if (print)
      printf("%-40s%-100s\n", it->first.c_str(), it->second.c_str());
    //std::cout << it->first << "\t\t" << it->second << "\n";
    std::string val = trim_str(it->second, " \r\n{}");
    std::deque < std::string > tok;

    std::string space = "|";
    pst::BasicStringTokenizer::tokenize(val, tok, space);

    parsed[count] = tok;
    keys[count] = it->first;
    count++;
  }

  std::vector<int> ind(count, 0);
  rec_run(parsed, runs, ind, 0);

  if (print)
  {
    std::cout << "------------------------------------ \n\n";
    std::cout << "Parameter combinations will be as follows: \n";
    std::cout << "  (values are in order as they appear above) \n";
    printf("%-10s%-100s\n", "Run#", "Values");
    std::cout << "------------------------------------------------ \n";
  }
  std::map<int, std::vector<std::string> >::iterator it2, end2 = runs.end();
  for (it2 = runs.begin(); it2 != end2; it2++)
  {
    std::string vecstr = "";
    for (int i = 0; i < (int) it2->second.size(); i++)
    {
      vecstr += it2->second[i] + " ";
    }
    if (print)
      printf("%-10i%-100s\n", it2->first, vecstr.c_str());
  }
}

