/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <math.h>

#include <sstmac/software/libraries/compute/lib_compute_loops.h>

#include <sstmac/software/process/key.h>

#include <stdint.h>

namespace sstmac
{
  namespace sw
  {

    int lib_compute_loops::mem_op_size_ = 16;

    double lib_compute_loops::mem_line_ratio_ = -1;
    double lib_compute_loops::flop_line_ratio_ = -1;
    bool lib_compute_loops::do_loops_ = true;

    void
    lib_compute_loops::check_init()
    {
      if (mem_line_ratio_ < 0 && sstmac_env::params)
        {
          mem_line_ratio_ = sstmac_env::params->get_optional_double_param(
              "lib_compute_loops_mem_ratio", 0.8);
          flop_line_ratio_ = sstmac_env::params->get_optional_double_param(
              "lib_compute_loops_flop_ratio", 0.8);
          do_loops_ = sstmac_env::params->get_optional_bool_param(
              "lib_compute_loops_enable", true);
        }
    }

    void
    lib_compute_loops::compute_loop_work(long long loop, double numlines, long tid) 
    {
        SSTMACBacktrace("Compute Loops");
          if (loop < 0)
            {
              loop = 0;
            }

          if (numlines < 0)
            {
              numlines = 0;
            }

          doing_memory_ = true;
          long long bytes = std::max((long long) 1,
              (long long) (loop * mem_line_ratio_ * numlines * mem_op_size_));
          lib_compute_memmove::read(bytes, tid);
          doing_memory_ = false;

          eventdata::ptr inst = eventdata::construct(get_lib_name());
          long long flops = std::max((long long) 1,
              (long long) (loop * flop_line_ratio_ * numlines));
          inst->set_event_value(eventdata::INST_FLOP, flops);
          lib_compute_inst::compute_inst(inst);
    }


    void
    lib_compute_loops::compute_loop(int from, int to, int nl, long tid)
    {
      compute_loop(from, to, double(nl),tid);
    }

    void
    lib_compute_loops::compute_loop(int from, int to, double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop = to - from + 1; //+1 for fortran loops

          compute_loop_work(loop, numlines,tid);

        }
    }

    void
    lib_compute_loops::compute_loop2(int from1, int to1, int from2, int to2,
        int nl, long tid)
    {
      compute_loop2(from1,to1,from2,to2,(double)nl,tid);
    }

    void
    lib_compute_loops::compute_loop2(int from1, int to1, int from2, int to2,
        double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop = loop1 * loop2;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_loop3(int from1, int to1, int from2, int to2,
        int from3, int to3, int nl, long tid)
    {
      compute_loop3(from1,to1,from2,to2,from3,to3,(double)nl,tid);
    }

    void
    lib_compute_loops::compute_loop3(int from1, int to1, int from2, int to2,
        int from3, int to3, double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop3 = to3 - from3 + 1;
          long long loop = loop1 * loop2 * loop3;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_loop4(int from1, int to1, int from2, int to2,
        int from3, int to3, int from4, int to4, int nl, long tid)
    {
      compute_loop4(from1,to1,from2,to2,from3,to3,from4,to4,(double)nl,tid);
    }

    void
    lib_compute_loops::compute_loop4(int from1, int to1, int from2, int to2,
        int from3, int to3, int from4, int to4, double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop3 = to3 - from3 + 1;
          long long loop4 = to4 - from4 + 1;
          long long loop = loop1 * loop2 * loop3 * loop4;;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_loop5(int from1, int to1, int from2, int to2,
        int from3, int to3, int from4, int to4, int from5, int to5, 
        double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop3 = to3 - from3 + 1;
          long long loop4 = to4 - from4 + 1;
          long long loop5 = to5 - from5 + 1;
          long long loop = loop1 * loop2 * loop3 * loop4 * loop5;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_loop6(int from1, int to1, int from2, int to2,
        int from3, int to3, int from4, int to4, int from5, int to5, int from6,
        int to6, double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop3 = to3 - from3 + 1;
          long long loop4 = to4 - from4 + 1;
          long long loop5 = to5 - from5 + 1;
          long long loop6 = to6 - from6 + 1;
          long loop = loop1 * loop2 * loop3 * loop4 *loop5 *loop6;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_loop7(int from1, int to1, int from2, int to2,
        int from3, int to3, int from4, int to4, int from5, int to5, int from6,
        int to6, int from7, int to7, double numlines, long tid)
    {
      check_init();
      if (do_loops_)
        {
          long long loop1 = to1 - from1 + 1; //+1 for fortran loops
          long long loop2 = to2 - from2 + 1;
          long long loop3 = to3 - from3 + 1;
          long long loop4 = to4 - from4 + 1;
          long long loop5 = to5 - from5 + 1;
          long long loop6 = to6 - from6 + 1;
          long long loop7 = to7 - from7 + 1;
          long long loop = loop1 * loop2 * loop3 * loop4 *loop5 *loop6 *loop7;

          compute_loop_work(loop, numlines,tid);
        }
    }

    void
    lib_compute_loops::compute_fft()
    {
      check_init();
      if (do_loops_)
        {
          eventdata::ptr inst = eventdata::construct(get_lib_name());
          inst->set_event_value(eventdata::INST_FLOP, 500);
          lib_compute_inst::compute_inst(inst);
        }
    }

  }
} //end of namespace sstmac
