/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/library.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace sw
  {

    static std::map<std::string, library_builder*>* library_builders_ = 0;

    void
    library::register_builder(const std::string &libname, library_builder *builder)
    {
        if (!library_builders_){
            library_builders_ = new std::map<std::string, library_builder*>;
        }
        (*library_builders_)[libname] = builder;
    }

    library::ptr
    library::construct_lib(const std::string& libname)
    {
        library_builder* builder = (*library_builders_)[libname];
        if (!builder){
            sst_throw(valueerror, "name %s is not known for constructing library", libname.c_str());
        }
        library::ptr lib = builder->construct(libname);
        return lib;
    }

    library::library() :
          blocking_(true),
        ev_type_(key::general),
        handler_(eventhandler::ptr()),
        waiter_(waiter::ptr()),
        self_initialized_(false)
    {
    }

    library::library(key::event_type_t ty) :
        blocking_(true),
      ev_type_(ty),
      handler_(eventhandler::ptr()),
      waiter_(waiter::ptr()),
      self_initialized_(false)
    {

    }

    library::~library()
    {
    }

    sst_message::ptr
    library::wait_any()
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_any::construct();
      waiter_->wait();
      return current_any_->recent_;
    }

    sst_message::ptr
    library::wait_this(const sst_message::ptr &msg)
    {
        if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }

        if (waiter_->is_callback()){
            if (!msg->has_callback()){
                sst_throw(illformederror,
                          "Callback waiter for library %s received msg without callback",
                          get_lib_name().c_str());
            }
            //no wait machinery needs to happen
            return msg;
        }


      thises_[msg] = wait_this::construct(msg);
      cancelled_[msg] = false;
      waiter_->wait();

      return msg;
    }

    void
    library::unregister_all_libs()
    {
      std::list<library::ptr>::iterator it, end = otherlibs_.end();
      for (it = otherlibs_.begin(); it != end; it++)
        {
          (*it)->unregister_all_libs();
          os_->unregister_lib(*it);
        }
      otherlibs_.clear();
    }

    void
    library::register_child_libs()
    {
    }

    std::string
    library::waiter_name() const
    {
        if (waiter_)
        {
            return waiter_->to_string();
        }
        else
        {
            return "NULL";
        }
    }


    void
    library::cancel_wait_this(const sst_message::ptr &msg)
    {
      //if (thises_.find(msg) != thises_.end())
      //  {
      //    thises_.erase(msg);
      //  }
      cancelled_[msg] = true;
    }

    sst_message::ptr
    library::wait_type(sst_message::MESSAGE_TYPES t)
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_type::construct(t);
      waiter_->wait();
      return current_any_->recent_;
    }

    sst_message::ptr
    library::wait_num_type(sst_message::MESSAGE_TYPES t, long num)
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_num_type::construct(t, num);
      waiter_->wait();
      return current_any_->recent_;
    }

    void
    library::register_lib(const library::ptr &lib)
    {
      if (os_)
        os_->register_lib(lib);
      else
        throw valueerror("library::register_lib - os not initialized yet");

      otherlibs_.push_back(lib);
    }

    operatingsystem::ptr
    library::get_full_os() const
    {
        return fullos_;
    }

    void
    library::init_os(const OS_API::ptr& os)
    {
      os_ = os;
      fullos_ = boost::dynamic_pointer_cast<operatingsystem>(os);

      if (!handler_)
        {
          key::ptr k = key::construct(ev_type_);
          handler_ = unblockhandler::construct(os_, k);
          waiter_ = blockwaiter::construct(os_, k);
          blocking_ = true;
          self_initialized_ = true;
        }

      if (sstmac_env::params)
        {
          consume_params(sstmac_env::params);
        }
    }

    void
    library::consume_params(const sim_parameters::ptr& params)
    {
      //do nothing by default
    }

    OS_API::ptr
    library::get_os() const
    {
      return os_;
    }

    void
    library::incoming_message(const sst_message::ptr &msg)
    {
        if (!handler_)
          {
            throw libraryerror(
                "library::wait - handler has not been initialized. If"
                  "the default library constructor was called, make sure that init_os was also called");
          }

        if (waiter_->is_callback()){
            if (!msg->has_callback())
                sst_throw(illformederror, "incoming message to library with callback handler does not have callback");
            handler_->handle(msg);
            return; //our work here is done
        }

      bool handle = false;
      if (thises_.find(msg) != thises_.end())
        {
          if (!cancelled_[msg])
            {
              wait::ptr w = thises_[msg];
              handle = w->incoming_message(msg);
            }
          cancelled_.erase(msg);
          thises_.erase(msg);
        }
      else
        {

          if (!current_any_)
            {
              std::cerr << "library " << to_string()
                  << " incoming message with no current wait \n";

              throw ssterror("library::incoming message");

            }
          handle = current_any_->incoming_message(msg);
        }



      if (handle)
        {
          handler_->handle(msg);

        }

    }

    sst_message::ptr
    library::get_arrived_msg()
    {
      if (!self_initialized_)
        {
          throw libraryerror(
              "library::get_arrived_msg - can't use this function unless the"
                "library base class auto-filled the handler");
        }
      unblockhandler::ptr ubh = boost::dynamic_pointer_cast<unblockhandler>(
          handler_);
      return ubh->saved_msg_;
    }

  }
} //end namespace

