/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_LIBRARY_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_LIBRARY_H_INCLUDED

#include <sstmac/software/process/os_api.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/software/process/waiter.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/process/blockwaiter.h>
#include <sstmac/software/libraries/unblockhandler.h>

#include <sstmac/software/libraries/wait.h>

namespace sstmac
{
  namespace sw
  {

    class operatingsystem;
    typedef boost::intrusive_ptr<operatingsystem> operatingsystemPtr;

    class library_builder;

    class library :
        virtual public ptr_type
    {
    public:
      typedef boost::intrusive_ptr<library> ptr;


    protected:

      library();

      library(key::event_type_t ty);

      virtual ~library();

      /**
       * wait will unblock on any message that arrives
       */
      sst_message::ptr
      wait_any();

      /**
       * wait_this will unblock only when it receives the given message back
       */
      sst_message::ptr
      wait_this(const sst_message::ptr &msg);

      void
      cancel_wait_this(const sst_message::ptr &msg);

      /**
       * wait_type will unblock when any message of the given type arrives
       */
      sst_message::ptr
      wait_type(sst_message::MESSAGE_TYPES t);

      /**
       * wait_num_type will unblock when num messages of the given type have arrived. returns the last one that arrived
       */
      sst_message::ptr
      wait_num_type(sst_message::MESSAGE_TYPES t, long num);

      /**
       * This function is provided so that libraries can instantiate, register, and use other libraries.
       * @param lib the library to register
       */
      void
      register_lib(const library::ptr &lib);

      /**
       * This is usually convenient to define to check if the node was set up correctly to support
       * what this library wants to do. This function is not currently a strict part of an interface.
       * @return true if the hardware supports this libraries functions
       */
      virtual bool supported() const
      {
        return true;
      }

      std::list<library::ptr> otherlibs_;

    public:
      static library::ptr
      construct_lib(const std::string& libname);

      static void register_builder(const std::string& libname, library_builder* builder);

      virtual void
      unregister_all_libs();

      /**
       * Optional function that is called when this library gets register.
       * Override this to register child libs.
       */
      virtual void
      register_child_libs();

      void
      init_handler(const eventhandler::ptr& env)
      {
        handler_ = env;
      }

      void
      init_waiter(const waiter::ptr& wa)
      {
        waiter_ = wa;
        blocking_ = wa->is_blocking();
      }

      virtual std::string
      to_string() const
      {
        return libname_;
      }

      virtual std::string
      get_lib_name() const
      {
        return libname_;
      }

      virtual void
      init_os(const OS_API::ptr& os);

      virtual void
      consume_params(const sim_parameters::ptr& params);


      virtual void
      incoming_message(const sst_message::ptr &msg);

      std::string
      waiter_name() const;

      OS_API::ptr get_os() const;

      operatingsystemPtr get_full_os() const;

    protected:

      sst_message::ptr
      get_arrived_msg();

      OS_API::ptr os_;
      operatingsystemPtr fullos_;

      std::string libname_;

      bool blocking_;

    private:
      key::event_type_t ev_type_;
      eventhandler::ptr handler_;
      waiter::ptr waiter_;

      std::map<sst_message::ptr, wait::ptr> thises_;
      std::map<sst_message::ptr, bool> cancelled_;
      wait::ptr current_any_;


      bool self_initialized_;
    };

    class library_builder
    {
        public:
            virtual library::ptr construct(const std::string& libname) = 0;

    };

    template <class T>
    class library_builder_tmpl :
        public library_builder
    {

        public:
            library_builder_tmpl(const std::string& libname)
            {
                library::register_builder(libname, this);
            }

            library::ptr construct(const std::string& libname)
            {
                return T::construct(libname);
            }

    };



  }
} //end of namespace sstmac

#endif
