#ifndef sstmac_software_libraries_mpi_MPIPROTOCOL_H
#define sstmac_software_libraries_mpi_MPIPROTOCOL_H

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-recvrequest.h>
#include <sstmac/software/process/graphviz.h>

namespace sstmac
{
  namespace sw
  {

    class mpiprotocol : public ptr_type
    {
    public:
      virtual void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg) = 0;

      virtual void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg) = 0;

      virtual void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg) = 0;

      virtual void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req) = 0;

      virtual void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req) = 0;

      virtual bool
      handshake_only() const = 0;

      virtual bool
      send_needs_completion_ack() const = 0;

      virtual bool
      send_needs_rendezvous_ack() const = 0;

      virtual bool
      send_needs_nic_ack() const = 0;

      virtual bool
      send_needs_eager_ack() const = 0;

      virtual void
      handle_nic_ack(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      typedef boost::intrusive_ptr<mpiprotocol> ptr;

      static mpiprotocol::ptr eager_ssend_protocol;
      static mpiprotocol::ptr eager0_mmap_protocol;
      static mpiprotocol::ptr eager0_socket_protocol;
      static mpiprotocol::ptr eager0_rdma_protocol;
      static mpiprotocol::ptr eager1_rdma_singlecpy_protocol;
      static mpiprotocol::ptr eager1_rdma_doublecpy_protocol;
      static mpiprotocol::ptr rendezvous_mmap_protocol;
      static mpiprotocol::ptr rendezvous_socket_protocol;
      static mpiprotocol::ptr rendezvous_rdma_protocol;
      static mpiprotocol::ptr creme_protocol;
      static mpiprotocol::ptr rendezvous_rma_protocol;

      static void
      init();
      static bool statics_done_;

    protected:
      static debug dbg_;

    };

    class eager_ssend : public mpiprotocol
    {
    public:

      typedef boost::intrusive_ptr<eager_ssend> ptr;

      bool
      handshake_only() const;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;
      bool
      send_needs_nic_ack() const;
      bool
      send_needs_eager_ack() const;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager ssend";
      }
    };

    class eager0_protocol : public mpiprotocol
    {
    public:

      typedef boost::intrusive_ptr<eager0_protocol> ptr;

      bool
      handshake_only() const;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;
      bool
      send_needs_nic_ack() const;
      bool
      send_needs_eager_ack() const;

      void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);
      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);
      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

    };

    class eager0_mmap : public eager0_protocol
    {
    public:
      typedef boost::intrusive_ptr<eager0_mmap> ptr;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager0_mmap";
      }
    };

    class eager0_socket : public eager0_protocol
    {
    public:
      typedef boost::intrusive_ptr<eager0_socket> ptr;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager0_socket";
      }
    };

    class eager0_rdma : public eager0_protocol
    {
    public:
      typedef boost::intrusive_ptr<eager0_rdma> ptr;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager0_rdma";
      }

    };

    class eager1_rdma : public mpiprotocol
    {
    public:
      typedef boost::intrusive_ptr<eager1_rdma> ptr;

      bool
      handshake_only() const;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;
      bool
      send_needs_nic_ack() const;
      bool
      send_needs_eager_ack() const;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      std::string
      to_string() const
      {
        return "eager1_rdma";
      }

    };

    class eager1_rdma_singlecpy : public eager1_rdma
    {
    public:
      typedef boost::intrusive_ptr<eager1_rdma_singlecpy> ptr;

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager1_rdma_singlecpy";
      }
    };

    class eager1_rdma_doublecpy : public eager1_rdma
    {
    public:
      typedef boost::intrusive_ptr<eager1_rdma_doublecpy> ptr;

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "eager1_rdma_doublecpy";
      }
    };

    class rendezvous_protocol : public mpiprotocol
    {
    public:
      typedef boost::intrusive_ptr<rendezvous_protocol> ptr;

      bool
      handshake_only() const;

      bool
      send_needs_nic_ack() const;
      bool
      send_needs_eager_ack() const;

      virtual void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

    };

    class rendezvous_socket : public rendezvous_protocol
    {
    public:
      typedef boost::intrusive_ptr<rendezvous_socket> ptr;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);
      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "rendezvous protocol socket";
      }
    };

    class rendezvous_rdma : public rendezvous_protocol
    {
    public:
      typedef boost::intrusive_ptr<rendezvous_rdma> ptr;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "rendezvous protocol rdma";
      }

      void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      init_system_recv(long nbytes, const mpiqueue::ptr& queue,
          const event::ptr& ev);

    };

    class rendezvous_rma : public rendezvous_protocol
    {
    public:
      typedef boost::intrusive_ptr<rendezvous_rma> ptr;

      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "rendezvous protocol rdma rma";
      }

      void
      recv_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

      void
      init_system_recv(long nbytes, const mpiqueue::ptr& queue,
          const event::ptr& ev);

    };

    class rendezvous_mmap : public rendezvous_protocol
    {
    public:
      bool
      send_needs_completion_ack() const;
      bool
      send_needs_rendezvous_ack() const;

      typedef boost::intrusive_ptr<rendezvous_mmap> ptr;

      void
      send_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_header(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      void
      handle_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg);

      std::string
      to_string() const
      {
        return "rendezvous protocol mmap";
      }

      void
      recv_payload(const mpiqueue::ptr& queue, const mpimessage::ptr& msg,
          const mpiqueue::recvrequest::ptr& req);

    };

  }
}

#endif // MPIPROTOCOL_H
