/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SHMEM_SHMEM_H_
#define SSTMAC_SHMEM_SHMEM_H_

#include <sstmac/sstmacro.h>

#include <cstddef>

#ifdef __cplusplus
extern "C"
{
#endif

  void
  start_pes(int npes);

  // ---- PE stuff ------
  int
  _num_pes();
  int
  shmem_n_pes();

  int
  _my_pe();
  int
  shmem_my_pe();

  int
  shmem_pe_accessible(int pe);

  int
  shmem_addr_accessible(void* addr, int pe);

  // ------ Heap and memory stuff ------
  void*
  shmalloc(size_t size);

  void*
  shmemalign(size_t align, size_t size);

  void*
  shrealloc(void* ptr, size_t size);

  void
  shfree(void* ptr);

  void*
  shmem_ptr(void* target, int pe);

  // ----- Synchornization stuff -----

  void
  shmem_barrier_all();

  void
  shmem_fence();

  void
  shmem_quiet();

  // ----- Access stuff ------

  void
  shmem_char_p(char *addr, char value, int pe);

  void
  shmem_short_p(short *addr, short value, int pe);

  void
  shmem_int_p(int *addr, int value, int pe);

  void
  shmem_long_p(long *addr, long value, int pe);

  void
  shmem_float_p(float *addr, float value, int pe);

  void
  shmem_double_p(double *addr, double value, int pe);

  void
  shmem_longlong_p(long long *addr, long long value, int pe);

  void
  shmem_longdouble_p(long double *addr, long double value, int pe);

  void
  shmem_char_put(char *addr, const char *value, size_t nelems, int pe);

  void
  shmem_short_put(short *addr, const short *value, size_t nelems, int pe);

  void
  shmem_int_put(int *addr, const int *value, size_t nelems, int pe);

  void
  shmem_long_put(long *addr, const long *value, size_t nelems, int pe);

  void
  shmem_float_put(float *addr, const float *value, size_t nelems, int pe);

  void
  shmem_double_put(double *addr, const double *value, size_t nelems, int pe);

  void
  shmem_longlong_put(long long *addr, const long long *value, size_t nelems,
      int pe);

  void
  shmem_longdouble_put(long double *addr, const long double *value,
      size_t nelems, int pe);

  short
  shmem_short_g(short *addr, int pe);
  int
  shmem_int_g(int *addr, int pe);
  long
  shmem_long_g(long *addr, int pe);
  float
  shmem_float_g(float *addr, int pe);
  double
  shmem_double_g(double *addr, int pe);
  long long
  shmem_longlong_g(long long *addr, int pe);
  long double
  shmem_longdouble_g(long double *addr, int pe);

  void
  shmem_char_get(char *target, const char *source, size_t nelems, int pe);
  void
  shmem_short_get(short *target, const short *source, size_t nelems, int pe);
  void
  shmem_int_get(int *target, const int *source, size_t nelems, int pe);
  void
  shmem_long_get(long *target, const long *source, size_t nelems, int pe);
  void
  shmem_float_get(float *target, const float *source, size_t nelems, int pe);
  void
  shmem_double_get(double *target, const double *source, size_t nelems, int pe);
  void
  shmem_longlong_get(long long *target, const long long *source, size_t nelems,
      int pe);
  void
  shmem_longdouble_get(long double *target, const long double *source,
      size_t nelems, int pe);

  // ---------- Atomic operations --------------- /
  void
  shmem_int_inc(int *target, int pe);
  void
  shmem_long_inc(long *target, int pe);
  void
  shmem_longlong_inc(long long *target, int pe);

  // ------------ Collectives -------------------- /

  void
  shmem_short_sum_to_all(short *target, short *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
  void
  shmem_int_sum_to_all(int *target, int *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, int *pWrk, long *pSync);
  void
  shmem_long_sum_to_all(long *target, long *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, long *pWrk, long *pSync);
  void
  shmem_float_sum_to_all(float *target, float *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
  void
  shmem_double_sum_to_all(double *target, double *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
  void
      shmem_longlong_sum_to_all(long long *target, long long *source,
          int nreduce, int PE_start, int logPE_stride, int PE_size,
          long long *pWrk, long *pSync);
  void
  shmem_longdouble_sum_to_all(long double *target, long double *source,
      int nreduce, int PE_start, int logPE_stride, int PE_size,
      long double *pWrk, long *pSync);

  void
  shmem_short_min_to_all(short *target, short *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
  void
  shmem_int_min_to_all(int *target, int *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, int *pWrk, long *pSync);
  void
  shmem_long_min_to_all(long *target, long *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, long *pWrk, long *pSync);
  void
  shmem_float_min_to_all(float *target, float *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
  void
  shmem_double_min_to_all(double *target, double *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
  void
      shmem_longlong_min_to_all(long long *target, long long *source,
          int nreduce, int PE_start, int logPE_stride, int PE_size,
          long long *pWrk, long *pSync);
  void
  shmem_longdouble_min_to_all(long double *target, long double *source,
      int nreduce, int PE_start, int logPE_stride, int PE_size,
      long double *pWrk, long *pSync);

  void
  shmem_short_max_to_all(short *target, short *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
  void
  shmem_int_max_to_all(int *target, int *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, int *pWrk, long *pSync);
  void
  shmem_long_max_to_all(long *target, long *source, int nreduce, int PE_start,
      int logPE_stride, int PE_size, long *pWrk, long *pSync);
  void
  shmem_float_max_to_all(float *target, float *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
  void
  shmem_double_max_to_all(double *target, double *source, int nreduce,
      int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
  void
      shmem_longlong_max_to_all(long long *target, long long *source,
          int nreduce, int PE_start, int logPE_stride, int PE_size,
          long long *pWrk, long *pSync);
  void
  shmem_longdouble_max_to_all(long double *target, long double *source,
      int nreduce, int PE_start, int logPE_stride, int PE_size,
      long double *pWrk, long *pSync);

#ifdef __cplusplus
}
#endif

#endif
