/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

/**
 * This file provides the conversion from C into the SST/macro
 * implementations of the SHMEM functions
 *
 */

#include "shmem.h"

#include <sstmac/software/process/os_api.h>
#include "shmem/app.h"
#include "shmem/api.h"

using namespace sstmac;
using namespace sstmac::sw;

/*
 * ---------------------------------------------------------------------------
 * --------------- Utility functions to convert to sstmac --------------------
 * ---------------------------------------------------------------------------
 */
shmem::api::ptr
get_current_shmem()
{
  thread::ptr t = OS_API::get_current_thread();
  shmem::app::ptr app;
  try
  {
    app = boost::dynamic_pointer_cast<shmem::app>(t);
  }
  catch (std::exception& e)
  {
    std::cerr << e.what() << std::endl;
    return shmem::api::ptr();
  }

  if (!app)
  {
    sst_throw(ssterror,
        "sstcpp::get_current_shmem - app didn't cast, make sure app is a shmemapp");
  }

  return app->get_shmem();
}

/*
 * ---------------------------------------------------------------------------
 * --------------- Now for SHMEM implementation functions --------------------
 * ---------------------------------------------------------------------------
 */

extern "C" void
start_pes(int npes)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->start_pes(npes);
}

extern "C" int
_num_pes()
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->_num_pes();
}

extern "C" int
shmem_n_pes()
{
  return _num_pes();
}

extern "C" int
_my_pe()
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->_my_pe();
}

extern "C" int
shmem_my_pe()
{
  return _my_pe();
}

extern "C" int
shmem_pe_accessible(int pe)
{
  return pe > 0 && pe < _num_pes();
}

extern "C" int
shmem_addr_accessible(void* addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_addr_accessible(addr, pe);
}

extern "C" void*
shmalloc(size_t size)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmalloc(size);
}

extern "C" void*
shmemalign(size_t align, size_t size)
{
  sst_throw(unimplementederror, "shmem::shmemalign is not implemented yet, alignment is not supported");
}

extern "C" void*
shrealloc(void* ptr, size_t size)
{
  sst_throw(unimplementederror, "shmem::shrealloc is not implemented yet");
}

extern "C" void
shfree(void* ptr)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shfree(ptr);
}

extern "C" void*
shmem_ptr(void* target, int pe)
{
  sst_throw(unimplementederror, "shmem::shmem_ptr is not implemented yet");
}

extern "C" void
shmem_barrier_all()
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_barrier_all();
}

/**
 * This is not implemented, it just calls quiet
 */
extern "C" void
shmem_fence()
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_quiet();
}

extern "C" void
shmem_quiet()
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_quiet();
}

extern "C" void
shmem_char_p(char *addr, char value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<char> (addr, value, pe);
}

extern "C" void
shmem_short_p(short *addr, short value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<short> (addr, value, pe);
}

extern "C" void
shmem_int_p(int *addr, int value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<int> (addr, value, pe);
}

extern "C" void
shmem_long_p(long *addr, long value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<long> (addr, value, pe);
}

extern "C" void
shmem_float_p(float *addr, float value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<float> (addr, value, pe);
}

extern "C" void
shmem_double_p(double *addr, double value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<double> (addr, value, pe);
}

extern "C" void
shmem_longlong_p(long long *addr, long long value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<long long> (addr, value, pe);
}

extern "C" void
shmem_longdouble_p(long double *addr, long double value, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_p<long double> (addr, value, pe);
}

extern "C" void
shmem_char_put(char *addr, const char *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<char> (addr, value, nelems, pe);
}

extern "C" void
shmem_short_put(short *addr, const short *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<short> (addr, value, nelems, pe);
}

extern "C" void
shmem_int_put(int *addr, const int *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<int> (addr, value, nelems, pe);
}

extern "C" void
shmem_long_put(long *addr, const long *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<long> (addr, value, nelems, pe);
}

extern "C" void
shmem_float_put(float *addr, const float *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<float> (addr, value, nelems, pe);
}

extern "C" void
shmem_double_put(double *addr, const double *value, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<double> (addr, value, nelems, pe);
}

extern "C" void
shmem_longlong_put(long long *addr, const long long *value, size_t nelems,
    int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<long long> (addr, value, nelems, pe);
}

extern "C" void
shmem_longdouble_put(long double *addr, const long double *value,
    size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_put<long double> (addr, value, nelems, pe);
}

short
shmem_short_g(short *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<short> (addr, pe);
}
int
shmem_int_g(int *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<int> (addr, pe);
}
long
shmem_long_g(long *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<long> (addr, pe);
}
float
shmem_float_g(float *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<float> (addr, pe);
}
double
shmem_double_g(double *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<double> (addr, pe);
}
long long
shmem_longlong_g(long long *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<long long> (addr, pe);
}
long double
shmem_longdouble_g(long double *addr, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  return sh->shmem_g<long double> (addr, pe);
}

extern "C" void
shmem_char_get(char *target, const char *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<char> (target, source, nelems, pe);
}
extern "C" void
shmem_short_get(short *target, const short *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<short> (target, source, nelems, pe);
}
extern "C" void
shmem_int_get(int *target, const int *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<int> (target, source, nelems, pe);
}
extern "C" void
shmem_long_get(long *target, const long *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<long> (target, source, nelems, pe);
}
extern "C" void
shmem_float_get(float *target, const float *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<float> (target, source, nelems, pe);
}
extern "C" void
shmem_double_get(double *target, const double *source, size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<double> (target, source, nelems, pe);
}
extern "C" void
shmem_longlong_get(long long *target, const long long *source, size_t nelems,
    int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<long long> (target, source, nelems, pe);
}
extern "C" void
shmem_longdouble_get(long double *target, const long double *source,
    size_t nelems, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_get<long double> (target, source, nelems, pe);
}

extern "C" void
shmem_int_inc(int *target, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_int_inc(target, pe);
}
extern "C" void
shmem_long_inc(long *target, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_long_inc(target, pe);
}
extern "C" void
shmem_longlong_inc(long long *target, int pe)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_longlong_inc(target, pe);
}

extern "C" void
shmem_short_sum_to_all(short *target, short *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, short *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<short> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_int_sum_to_all(int *target, int *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, int *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<int> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_long_sum_to_all(long *target, long *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<long> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_float_sum_to_all(float *target, float *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, float *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<float> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_double_sum_to_all(double *target, double *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longlong_sum_to_all(long long *target, long long *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<long long> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longdouble_sum_to_all(long double *target, long double *source,
    int nreduce, int PE_start, int logPE_stride, int PE_size,
    long double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_sum_to_all<long double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}

extern "C" void
shmem_short_min_to_all(short *target, short *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, short *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<short> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_int_min_to_all(int *target, int *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, int *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<int> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_long_min_to_all(long *target, long *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<long> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_float_min_to_all(float *target, float *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, float *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<float> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_double_min_to_all(double *target, double *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longlong_min_to_all(long long *target, long long *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<long long> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longdouble_min_to_all(long double *target, long double *source,
    int nreduce, int PE_start, int logPE_stride, int PE_size,
    long double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_min_to_all<long double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}

extern "C" void
shmem_short_max_to_all(short *target, short *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, short *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<short> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_int_max_to_all(int *target, int *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, int *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<int> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_long_max_to_all(long *target, long *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<long> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_float_max_to_all(float *target, float *source, int nreduce, int PE_start,
    int logPE_stride, int PE_size, float *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<float> (target, source, nreduce, PE_start, logPE_stride,
      PE_size, pWrk, pSync);
}
extern "C" void
shmem_double_max_to_all(double *target, double *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longlong_max_to_all(long long *target, long long *source, int nreduce,
    int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<long long> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
extern "C" void
shmem_longdouble_max_to_all(long double *target, long double *source,
    int nreduce, int PE_start, int logPE_stride, int PE_size,
    long double *pWrk, long *pSync)
{
  shmem::api::ptr sh = get_current_shmem();

  sh->shmem_max_to_all<long double> (target, source, nreduce, PE_start,
      logPE_stride, PE_size, pWrk, pSync);
}
