#include <sstmac/software/libraries/sockets/lib_sockets.h>
#include <sstmac/software/process/operatingsystem.h>

#define enumcase(x) case x: return #x;

namespace sstmac {
    namespace sw {

    debug lib_socket::dbg_;
    debug socket::dbg_;
    taskid socket_message::nic_task(-1);
    static library_builder_tmpl<lib_socket> lib_socket_builder("libsocket");

    socket_message::socket_message(int number, long bytes, const std::string &libname,
                                   const nodeaddress::ptr &toaddr, const nodeaddress::ptr &fromaddr)
        : network_message(toaddr, fromaddr), library_message(libname), number_(number), nbytes_(bytes)
    {

    }

    long
    socket_message::nbytes() const
    {
        return nbytes_;
    }

    int
    socket_message::number() const
    {
        return number_;
    }

    long
    socket_message::get_byte_length() const
    {
        return nbytes_;
    }

    sst_message::ptr
    socket_message::clone(MESSAGE_TYPES t) const
    {
        socket_message::ptr newmsg = socket_message::construct(number_, nbytes_, lib_name_, toaddr_, fromaddr_);
        newmsg->msgtype_ = t;
        return newmsg;
    }

    socket_message::ptr
    socket_message::construct(int socket, long bytes, const std::string &libname,
                              const nodeaddress::ptr &toaddr, const nodeaddress::ptr &fromaddr)
    {
        return new socket_message(socket, bytes, libname, toaddr, fromaddr);
    }

    socket::socket(int fd, const nodeaddress::ptr &my_addr,
                   socket_domain_t domain, socket_type_t typ, socket_protocol_t protocol,
                   const operatingsystem::ptr& os)
        :  number_(unbound), fd_(fd), domain_(domain), protocol_(protocol),
          type_(typ), my_addr_(my_addr),
          key_(key::construct()), os_(os),
          state_(UNBLOCKED), bytes_available(0)
    {
        if (!dbg_.is_real()){
            dbg_ = debug("<debug> socket");
        }
    }

    int
    socket::number() const
    {
        return number_;
    }

    void
    socket::bind(int number, const nodeaddress::ptr& addr)
    {
        if (number_ != unbound){
            sst_throw(valueerror, "socket %d is already bound.  cannot bind again", number_);
        }
        number_ = number;
    }

    void
    socket::disconnect()
    {
        remote_addr_ = 0;
        number_ = unbound;
    }

    void
    socket::connect(int number, const nodeaddress::ptr& addr)
    {
        if (remote_addr_){
            sst_throw(valueerror, "socket %d is already connected", number_);
        }
        remote_addr_ = addr;
        number_ = number;
    }

    void
    socket::unaccept()
    {
        bytes_available = 0;
        remote_addr_ = 0;
        pending_recv_.clear();
    }

    void
    socket::shutdown()
    {
        number_ = unbound;
        bytes_available = 0;
        remote_addr_ = 0;
    }

    std::string
    socket::to_string() const
    {
        return sstprintf("socket fd=%d on node %s", number_, my_addr_->to_string().c_str());
    }

    void
    socket::add_pending(const socket_message::ptr& msg)
    {
        bytes_available += msg->nbytes();
        pending_recv_.push_back(msg);
    }

    void
    socket::accept(const nodeaddress::ptr& addr)
    {
        remote_addr_ = addr;
    }

    nodeaddress::ptr
    socket::send_addr() const
    {
        return remote_addr_;
    }

    nodeaddress::ptr
    socket::my_addr() const
    {
        return my_addr_;
    }

    long
    socket::recv(long max_bytes)
    {
        long bytes_recved = 0;
        std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        while (it != end){
            std::list<socket_message::ptr>::iterator tmp = it++;
            socket_message::ptr msg = *tmp;
            long next_bytes = bytes_recved + msg->nbytes();
            if (next_bytes > max_bytes){ //can't go over!
                return bytes_recved;
            }
            bytes_recved += msg->nbytes();
            pending_recv_.erase(tmp);
            bytes_available -= msg->nbytes();
        }
        return bytes_recved;
    }

    socket::state_t
    socket::state() const
    {
        return state_;
    }

    int
    socket::fd() const
    {
        return fd_;
    }

    void
    socket::rebind()
    {
        remote_addr_ = 0;
    }

    nodeaddress::ptr
    socket::recv_addr() const
    {
        return remote_addr_;
    }

    void
    socket::block(state_t state)
    {
        state_ = state;
        os_->block(key_);
    }

    void
    socket::unblock()
    {
        state_ = UNBLOCKED;
        os_->unblock(key_);
    }

    const char*
    socket::tostr(state_t state)
    {
        switch (state){
            enumcase(UNBLOCKED);
            enumcase(BLOCK_ON_ACCEPT);
            enumcase(BLOCK_ON_SEND);
            enumcase(BLOCK_ON_RECV);
        }
        return "";
    }

    lib_socket::lib_socket(const std::string &id)
        : lib_compute_memmove(id),
          next_free_socket_(0),
          in_blocking_send_(false)
    {
        if (!dbg_.is_real()){
            dbg_ = debug("<debug> socket | lib_socket");
        }
    }

    lib_socket::~lib_socket()
    {
    }

    socket::ptr
    lib_socket::get_socket(int fd) const
    {
        std::map<int, socket::ptr>::const_iterator it = all_sockets_.find(fd);
        if (it == all_sockets_.end()){
            sst_throw(valueerror, "socket %d does not exist yet in lib_socket::recv");
        }
        return it->second;
    }

    lib_socket::ptr
    lib_socket::construct(const std::string &str)
    {
        return new lib_socket(str);
    }

    socket::ptr
    lib_socket::get_connected_socket(int number) const
    {
        std::map<int, socket::ptr>::const_iterator it = connected_sockets_.find(number);
        if (it == connected_sockets_.end()){
            return socket::ptr();
        }
        return it->second;
    }

    socket::ptr
    lib_socket::get_bound_socket(int number) const
    {
        std::map<int, socket::ptr>::const_iterator it = bound_sockets_.find(number);
        if (it == bound_sockets_.end()){
            return socket::ptr();
        }
        return it->second;
    }

    long
    lib_socket::recv(int fd, long max_bytes, bool blocking)
    {
        SSTMAC_DEBUG << sstprintf("Starting %s receive of %ld bytes on socket %d\n",
                                  blocking ? "blocking" : "non-blocking",
                                  max_bytes, fd);
        socket::ptr sock = get_socket(fd);
        if (blocking && sock->bytes_available == 0){ //gotta recv something!
            SSTMAC_DEBUG << sstprintf("Blocking on socket %d until %ld bytes are available.  Only have %ld\n",
                                      fd, max_bytes, sock->bytes_available);
            sock->block(socket::BLOCK_ON_RECV);
            SSTMAC_DEBUG << sstprintf("Unblocking on socket %d now that %ld bytes are available\n",
                                      fd, sock->bytes_available);
        }
        long bytes = std::min(sock->bytes_available, max_bytes);

        long dstid = fullos_->get_current_threadid();
        sock->rebind();
        lib_compute_memmove::read(bytes, thread::nic_thread, dstid);
        return bytes;
    }

    long
    lib_socket::send(int fd, long bytes, bool wait_on_ack)
    {
        SSTMAC_DEBUG << sstprintf("Starting send of %ld bytes on socket %d\n", bytes, fd);
        socket::ptr sock = get_socket(fd);

        socket_message::ptr msg = socket_message::construct(
                                    sock->number(), bytes, libname_,
                                    sock->send_addr(),
                                    sock->my_addr());
        msg->set_needs_ack(wait_on_ack);
        long srcid = fullos_->get_current_threadid();
        lib_compute_memmove::write(bytes, srcid, thread::nic_thread);
        os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
        if (wait_on_ack){
            SSTMAC_DEBUG << sstprintf("Waiting for ack on socket %d\n", fd);
            sock->block(socket::BLOCK_ON_SEND);
            SSTMAC_DEBUG << sstprintf("Got ack on socket %d\n", fd);
        }
        return bytes;
    }

    void
    lib_socket::init_os(const OS_API::ptr &os)
    {
        library::init_os(os);
    }

    void
    lib_socket::close(int fd)
    {
        std::map<int, socket::ptr>::iterator it = all_sockets_.find(fd);
        if (it == all_sockets_.end()){
            sst_throw(valueerror, "cannot close socket %d. no such file descriptor could be found", fd);
        }
        all_sockets_.erase(it);
        free_sockets_.push_back(fd);
    }

    int
    lib_socket::allocate_socket(socket_domain_t domain, socket_type_t typ, socket_protocol_t prot)
    {
        int fd;
        if (free_sockets_.empty()){
            fd = next_free_socket_++;

        }
        else {
            fd = free_sockets_.back();
            free_sockets_.pop_back();
        }
        nodeaddress::ptr addr = operatingsystem::get_current_node();
        socket::ptr newsock = new socket(fd, addr, domain, typ, prot, fullos_);
        all_sockets_[fd] = newsock;
        return fd;
    }

    void
    lib_socket::incoming_message(const sst_message::ptr &msg)
    {
        socket_message::ptr sockmsg = boost::dynamic_pointer_cast<socket_message>(msg);
        int number = sockmsg->number();
        if (!sockmsg){
            sst_throw(illformederror, "lib_socket received a payload from a non-socket message");
        }

        SSTMAC_DEBUG << sstprintf("Incoming message of type=%s size=%ld from addr=%ld on socket number=%d\n",
                                  sst_message::tostr(msg->get_type()), sockmsg->nbytes(), msg->fromaddr()->unique_id(), number);

        if (msg->get_type() == sst_message::SEND_ACK){
            socket::ptr sock = get_connected_socket(number);
            if (!sock){
                sst_throw(illformederror, "received ack for unconnected send on socket number=%d", number);
            }
            else if (sock->state() != socket::BLOCK_ON_SEND){
                sst_throw(illformederror, "received ack for non-blocking send on socket number=%d", number);
            }
            else {
                sock->unblock();
            }
        }
        else {
            socket::ptr sock = get_bound_socket(number);
            if (!sock){
                SSTMAC_DEBUG << sstprintf("No bound socket found for number=%d\n", number);
                pending_recv_.push_back(sockmsg);
                return;
            }
            else if (sock->recv_addr()->equals(msg->fromaddr())){
                sock->add_pending(sockmsg);
                SSTMAC_DEBUG << sstprintf("Socket number=%d now has %ld bytes available\n",
                                          sock->number(), sock->bytes_available);
            }
            else {
                pending_recv_.push_back(sockmsg);
            }

            SSTMAC_DEBUG << sstprintf("Checking socket on fd=%d number=%d in state=%s\n",
                                      sock->fd(), sock->number(), socket::tostr(sock->state()));


            if (sock->state() == socket::BLOCK_ON_ACCEPT){
                if (sock->recv_addr()->equals(msg->fromaddr()))
                {
                    SSTMAC_DEBUG << sstprintf("Unblocking accept on fd=%d number=%d on node=%ld\n",
                                             sock->fd(), sock->number(), msg->fromaddr()->unique_id());
                    sock->accept(msg->fromaddr());
                    sock->unblock();
                }
                else {
                    SSTMAC_DEBUG << sstprintf("Failing to unblock on fd=%d number=%d for needed addr=%ld on msg addr=%ld\n",
                                              sock->fd(), sock->number(), sock->recv_addr()->unique_id(), msg->fromaddr()->unique_id());
                }
            }
            else if (sock->state() == socket::BLOCK_ON_RECV)
            {
                SSTMAC_DEBUG << sstprintf("Socket fd=%d number=%d has %ld bytes available and needs %ld to unblock\n",
                                         sock->fd(), sock->number(),
                                         sock->bytes_available, sock->bytes_needed_to_unblock);
                sock->unblock();
            }

        }
    }

    nodeaddress::ptr
    lib_socket::accept(int fd, const nodeaddress::ptr& addr, bool blocking)
    {
        socket::ptr sock = get_socket(fd);

        SSTMAC_DEBUG << sstprintf("%s accept on socket fd=%d number=%d for addr %ld\n",
                                  blocking ? "Blocking" : "Non-blocking",
                                  fd, sock->number(), addr->unique_id());

        { std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        for ( ; it != end; ++it){
            socket_message::ptr msg = *it;
            if (msg->number() == sock->number() && addr->equals(msg->fromaddr())){
                sock->accept(msg->fromaddr());
                break;
            }
        } }

        if (!sock->recv_addr()){
            if (blocking){
                sock->accept(addr);
                sock->block(socket::BLOCK_ON_ACCEPT);
            }
            else {
                return nodeaddress::ptr();
            }
        }

        { std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        for ( ; it != end; ++it){
            //socket_message::ptr msg = *it;
        } }

        SSTMAC_DEBUG << sstprintf("Accept on socket fd=%d number=%d found addr %ld\n",
                                  fd, sock->number(), sock->recv_addr()->unique_id());

        return sock->recv_addr();
    }

    void
    lib_socket::unaccept(int fd)
    {
         socket::ptr sock = get_socket(fd);
         sock->unaccept();
    }

    void
    lib_socket::connect(int fd, int number, const nodeaddress::ptr& addr)
    {
        SSTMAC_DEBUG << sstprintf("Connected socket fd=%d number=%d\n",
                                  fd, number);
        socket::ptr sock = get_socket(fd);
        sock->connect(number,addr);
        connected_sockets_[number] = sock;
    }

    void
    lib_socket::disconnect(int fd)
    {
        socket::ptr sock = get_socket(fd);
        SSTMAC_DEBUG << sstprintf("Disconnected socket fd=%d number=%d\n",
                                  sock->fd(), sock->number());
        sock->disconnect();
        connected_sockets_.erase(sock->number());
    }

    void
    lib_socket::bind(int fd, int number, const nodeaddress::ptr& addr)
    {
        SSTMAC_DEBUG << sstprintf("Binding socket on fd %d number %d\n",
                                  fd, number);
        socket::ptr sock = get_socket(fd);
        sock->bind(number, addr);
        bound_sockets_[number] = sock;
    }

    bool
    lib_socket::supported() const
    {
        return lib_compute_memmove::supported() && os_->kernel_supported(AMI::AMI_COMM_SEND);
    }



    }
}
