#include <sstmac/software/process/ftq.h>

namespace sstmac {
    namespace sw {


    FTQCalendar::FTQCalendar()
        : stat_logger("<stats> ftq | intensity", "ftq.csv"),
          num_epochs_(0), min_event_size_(0), num_ticks_epoch_(0), show_vis_(false),
          show_vis_thread_chart_(false), show_vis_thread_chart_type_("bar"),
          show_vis_thread_chart_separate_(true), show_vis_summary_chart_(false)
    {
    }

    void
    FTQCalendar::init(long num_epochs, long nticks_per_epoch, long min_event_size, bool show_vis,
    		bool show_vis_thread_chart, std::string show_vis_thread_chart_type,
       		bool show_vis_thread_chart_separate, bool show_vis_summary_chart)
    {
        num_epochs_ = num_epochs;
        min_event_size_ = min_event_size;
        num_ticks_epoch_ = nticks_per_epoch;
        show_vis_ = show_vis;
        show_vis_thread_chart_ = show_vis_thread_chart;
        show_vis_thread_chart_type_ = show_vis_thread_chart_type;
        show_vis_thread_chart_separate_ = show_vis_thread_chart_separate;
        show_vis_summary_chart_ = show_vis_summary_chart;
    }

    FTQCalendar::ptr
    FTQCalendar::construct()
    {
        return new FTQCalendar();
    }

    void
    FTQCalendar::register_app(int aid, const std::string& appname)
    {
        calendars_[aid] = new AppFTQCalendar(aid, appname, num_epochs_, num_ticks_epoch_, min_event_size_, show_vis_,
        		show_vis_thread_chart_, show_vis_thread_chart_type_, show_vis_thread_chart_separate_, show_vis_summary_chart_);
    }

    void
    FTQCalendar::simulation_finished(timestamp t_end)
    {
        if (!active_)
            return;

        boost::unordered_map<int, AppFTQCalendar::ptr>::iterator it = calendars_.begin(), end = calendars_.end();
        for ( ; it != end; ++it){
            it->second->dump();
        }
    }

    void
    FTQCalendar::collect(key::event_type_t ty, int aid, int tid, long ticks_begin, long num_ticks)
    {
        if (num_ticks < min_event_size_)
            return; //don't record it

        calendars_[aid]->collect(ty, tid, ticks_begin, num_ticks);
    }

    AppFTQCalendar::ptr
    FTQCalendar::get_calendar(int aid) const
    {
        boost::unordered_map<int, AppFTQCalendar::ptr>::const_iterator it = calendars_.find(aid);
        if (it == calendars_.end()){
            sst_throw(valueerror, "no FTQ calendar found for app %d", aid);
        }
        return it->second;
    }

    AppFTQCalendar::AppFTQCalendar(int aid, const std::string& appname, long num_epochs, long nticks_epoch, long min_event_size,
    		bool show_vis, bool show_vis_thread_chart, std::string show_vis_thread_chart_type,
       		bool show_vis_thread_chart_separate, bool show_vis_summary_chart)
        : epochs_(0), aid_(aid), max_tid_(0), num_epochs_(num_epochs), max_epoch_(0),
          min_event_size_(min_event_size), num_ticks_epoch_(nticks_epoch), appname_(appname), show_vis_(show_vis),
          show_vis_thread_chart_(show_vis_thread_chart), show_vis_thread_chart_type_(show_vis_thread_chart_type),
          show_vis_thread_chart_separate_(show_vis_thread_chart_separate), show_vis_summary_chart_(show_vis_summary_chart)
    {
        epochs_ = new FTQEpoch[num_epochs];

#ifdef HAVE_VTK
        if (show_vis_ && view == 0 && chart == 0) {
        	view = vtkSmartPointer<vtkContextView>::New();
        	chart = vtkSmartPointer<vtkChartXY>::New();
        	view->GetRenderWindow()->SetSize(500, 500);
        	view->GetScene()->AddItem(chart);

        	chart->SetTitle("CPU and MPI Time for App: " + appname_);
        	chart->GetAxis(vtkAxis::LEFT)->SetTitle("Thread ID");
        	chart->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch (us)");

        	chart->SetShowLegend(true);
        	chart->GetLegend()->SetInline(false);
        	chart->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
        	chart->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);

        	view->GetRenderWindow()->SetMultiSamples(0);
        	if (!show_vis_summary_chart_)
        		view->GetInteractor()->Initialize();
        }
#endif
    }

    void
    AppFTQCalendar::collect(key::event_type_t ty, int tid, long ticks_begin, long num_ticks)
    {
        TaskFTQCalendar::ptr& cal = calendars_[tid];
        if (!cal){
            cal = new TaskFTQCalendar();
            if (tid > max_tid_){
                max_tid_ = tid;
            }
        }
        FTQEpoch*& t_epoch = thread_epochs_[tid];
        if (!t_epoch){
        	t_epoch = new FTQEpoch[num_epochs_];
            if (tid > max_tid_){
                max_tid_ = tid;
            }
        }
        /** aggregate for all time intervals and all threads for given event type */
        totals_[ty] += num_ticks;
        cal->collect(ty, ticks_begin, num_ticks);

        long epoch = ticks_begin / num_ticks_epoch_;
        long nticks_next_epoch = num_ticks_epoch_ - num_ticks % num_ticks_epoch_;
        nticks_next_epoch = std::min(nticks_next_epoch, num_ticks);
        /** aggregate for all threads for given time interval and event type */
        epochs_[epoch].collect(ty, nticks_next_epoch);
        t_epoch[epoch].collect(ty, nticks_next_epoch);
        num_ticks -= nticks_next_epoch;
        while (num_ticks > 0){
            ++epoch;
            nticks_next_epoch = std::min(num_ticks, num_ticks_epoch_);
            epochs_[epoch].collect(ty, nticks_next_epoch);
            t_epoch[epoch].collect(ty, nticks_next_epoch);
            num_ticks -= nticks_next_epoch;
        }
        long old_max_epoch = max_epoch_;
        max_epoch_ = std::max(epoch, max_epoch_);
        // Update charts only if max_epoch_ has advanced
        if (show_vis_ && !show_vis_summary_chart_ && max_epoch_ && (max_epoch_ > old_max_epoch))
        	display_charts(tid, max_epoch_);
    }

    void
    AppFTQCalendar::display_charts(int tid, long max_epoch){

#ifdef HAVE_VTK
    	// Data structure for main chart with all threads showing
    	std::map<int, vtkTable*> table_cpu_map;
    	std::map<int, vtkLongArray*> X_cpu_map;
    	std::map<int, vtkLongArray*> Y_cpu_map;
    	std::map<int, vtkPlot *> line_cpu_map;

    	// Data structure for per thread chart on separate window
    	std::map<int, vtkTable*> table_thread_cpu_map;
    	std::map<int, vtkLongArray*> X_thread_cpu_map;
    	std::map<int, vtkLongArray*> Y_thread_cpu_map;
    	std::map<int, vtkPlot *> thread_cpu_plot_map;
    	long cpu_max[max_tid_];

    	// Data structure for main chart with all threads showing
    	std::map<int, vtkTable*> table_mpi_map;
    	std::map<int, vtkLongArray*> X_mpi_map;
    	std::map<int, vtkLongArray*> Y_mpi_map;
    	std::map<int, vtkPlot *> line_mpi_map;
    	long mpi_max[max_tid_];

    	long epoch_max[max_tid_];
    	long maximum_epoch = 0;

    	// Data structure for per thread chart on separate window
    	std::map<int, vtkTable*> table_thread_mpi_map;
    	std::map<int, vtkLongArray*> X_thread_mpi_map;
    	std::map<int, vtkLongArray*> Y_thread_mpi_map;
    	std::map<int, vtkPlot *> thread_mpi_plot_map;

    	// Data structure for per thread chart upto 9 threads per window
     	std::map <int, vtkTable*> table_both_map;
    	std::map <int, vtkFloatArray*> X_both_map;
    	std::map <int, vtkFloatArray*> Y_both_cpu_map;
    	std::map <int, vtkFloatArray*> Y_both_mpi_map;
    	std::map <int, vtkPlotStacked *> line_both_map;

    	chart->ClearPlots();

    	for (int i = 0; i <= max_tid_; ++i) {
    		FTQEpoch* t_epoch = thread_epochs_[i];
    		if (t_epoch) {
    			table_cpu_map.insert(std::make_pair(i, vtkTable::New()));
    			X_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
    			Y_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
    			// CPU will be shown points with CROSS characteristics (X mark)
    			line_cpu_map.insert(
    					std::make_pair(i, chart->AddPlot(vtkChart::POINTS)));
    			vtkPlotPoints::SafeDownCast(line_cpu_map[i])->SetMarkerStyle(vtkPlotPoints::CROSS);
    			X_cpu_map[i]->SetName("X");
    			table_cpu_map[i]->AddColumn(X_cpu_map[i]);
    			Y_cpu_map[i]->SetName("Compute Time");
    			table_cpu_map[i]->AddColumn(Y_cpu_map[i]);

    			table_mpi_map.insert(std::make_pair(i, vtkTable::New()));
    			X_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
    			Y_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
    			// MPI will be shown solid points (dot)
    			line_mpi_map.insert(
    					std::make_pair(i, chart->AddPlot(vtkChart::POINTS)));
    			X_mpi_map[i]->SetName("X");
    			table_mpi_map[i]->AddColumn(X_mpi_map[i]);
    			Y_mpi_map[i]->SetName("MPI Time");
    			table_mpi_map[i]->AddColumn(Y_mpi_map[i]);

    			cpu_max[i] = 0;
    			mpi_max[i] = 0;
    			epoch_max[i] = 0;

    			if (show_vis_thread_chart_) {
    				if (show_vis_thread_chart_separate_) {
						if (view_thread.find(i) == view_thread.end()) { // If it is first time being called, init data structures
							view_thread.insert(std::make_pair(i, vtkContextView::New()));
							chart_thread.insert(std::make_pair(i, vtkChartXY::New()));
							view_thread[i]->GetRenderWindow()->SetSize(500, 500);
							view_thread[i]->GetScene()->AddItem(chart_thread[i]);

							chart_thread[i]->SetTitle(
									"CPU and MPI Time for Thread: " + boost::lexical_cast<std::string>(i));
							chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetTitle("Number of Ticks");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch (us)");

							chart_thread[i]->SetShowLegend(true);
							// Show legend outside the chart box
							chart_thread[i]->GetLegend()->SetInline(false);
							chart_thread[i]->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
							chart_thread[i]->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);

							view_thread[i]->GetRenderWindow()->SetMultiSamples(0);
							view_thread[i]->GetInteractor()->Initialize();
						}
						else
							chart_thread[i]->ClearPlots();
    				}
    				else { // show_vis_thread_chart_separate_ == false. Show upto 9 threads in one chart
    					if (view_thread.find(i/9) == view_thread.end()) { // If it is first time being called, init data structures
							view_thread.insert(std::make_pair(i/9, vtkContextView::New()));
							view_thread[i/9]->GetRenderWindow()->SetSize(900, 900);

							view_thread[i/9]->GetRenderWindow()->SetMultiSamples(0);
							view_thread[i/9]->GetInteractor()->Initialize();
    					}
						if (chart_thread.find(i) == chart_thread.end()) {
							chart_thread.insert(std::make_pair(i, vtkChartXY::New()));
							view_thread[i/9]->GetScene()->AddItem(chart_thread[i]);

							chart_thread[i]->SetTitle(
									"Thread: " + boost::lexical_cast<std::string>(i));
							chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetTitle("Number of Ticks");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch (us)");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetBehavior(vtkAxis::FIXED);
							chart_thread[i]->SetAutoSize(false);
							chart_thread[i]->SetSize(vtkRectf((i%3)*300, ((8-(i%9))/3)*300, 300.0, 300.0));

							chart_thread[i]->SetShowLegend(true);
							chart_thread[i]->GetLegend()->SetInline(true);
							chart_thread[i]->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
		 					chart_thread[i]->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);
						}
    				}

    				if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
						chart_thread[i]->ClearPlots();
						if (show_vis_thread_chart_type_ == "area") {
							table_both_map.insert(std::make_pair(i, vtkTable::New()));
							X_both_map.insert(std::make_pair(i, vtkFloatArray::New()));
							Y_both_cpu_map.insert(std::make_pair(i, vtkFloatArray::New()));
							Y_both_mpi_map.insert(std::make_pair(i, vtkFloatArray::New()));
							line_both_map.insert(std::make_pair(i, vtkPlotStacked::SafeDownCast(chart_thread[i]->AddPlot(vtkChart::STACKED))));

							X_both_map[i]->SetName("X");
							table_both_map[i]->AddColumn(X_both_map[i]);
							Y_both_cpu_map[i]->SetName("CPU");
							table_both_map[i]->AddColumn(Y_both_cpu_map[i]);
							Y_both_mpi_map[i]->SetName("MPI");
							table_both_map[i]->AddColumn(Y_both_mpi_map[i]);
						} else { // if chart type is bar or line
							table_thread_cpu_map.insert(std::make_pair(i, vtkTable::New()));
							X_thread_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
							Y_thread_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
							if (show_vis_thread_chart_type_ == "bar") {
								thread_cpu_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::BAR)));
							} else if (show_vis_thread_chart_type_ == "line") {
								thread_cpu_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::LINE)));
							}
							X_thread_cpu_map[i]->SetName("X");
							table_thread_cpu_map[i]->AddColumn(X_thread_cpu_map[i]);
							Y_thread_cpu_map[i]->SetName("Compute Time");
							table_thread_cpu_map[i]->AddColumn(Y_thread_cpu_map[i]);

							table_thread_mpi_map.insert(std::make_pair(i, vtkTable::New()));
							X_thread_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
							Y_thread_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
							if (show_vis_thread_chart_type_ == "bar") {
								thread_mpi_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::BAR)));
							} else if (show_vis_thread_chart_type_ == "line") {
								thread_mpi_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::LINE)));
							}
							X_thread_mpi_map[i]->SetName("X");
							table_thread_mpi_map[i]->AddColumn(X_thread_mpi_map[i]);
							Y_thread_mpi_map[i]->SetName("MPI Time");
							table_thread_mpi_map[i]->AddColumn(Y_thread_mpi_map[i]);
						}
    				}
    			}
    		}
    	}

    	// Fill in the data that will be plotted in the chart
    	for (int i = 0; i < max_epoch; i++) {
    		for (int j = 0; j <= max_tid_; ++j) {
    			FTQEpoch* t_epoch = thread_epochs_[j];
    			if (t_epoch) {
    				long nticks_cpu = t_epoch[i].get_event_time(key::compute);
    				long nticks_mpi = t_epoch[i].get_event_time(key::mpi);
    				if (nticks_cpu > nticks_mpi && nticks_cpu > 0) {
    					table_cpu_map[j]->InsertNextBlankRow();
					    table_cpu_map[j]->SetValue(table_cpu_map[j]->GetNumberOfRows()-1, 0, i);
    					table_cpu_map[j]->SetValue(table_cpu_map[j]->GetNumberOfRows()-1, 1, j);
    					if (i > maximum_epoch)
    						maximum_epoch = i;
    				}
    				if (nticks_mpi > nticks_cpu && nticks_mpi > 0) {
    					table_mpi_map[j]->InsertNextBlankRow();
					    table_mpi_map[j]->SetValue(table_mpi_map[j]->GetNumberOfRows()-1, 0, i);
    					table_mpi_map[j]->SetValue(table_mpi_map[j]->GetNumberOfRows()-1, 1, j);
    					if (i > maximum_epoch)
    						maximum_epoch = i;
    				}

    				if (show_vis_thread_chart_) {
    					if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
    						if (show_vis_thread_chart_type_ == "area") {
    				        	table_both_map[j]->InsertNextBlankRow();
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 0, table_both_map[j]->GetNumberOfRows()-1);
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 1, nticks_cpu);
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 2, nticks_mpi);
								if (cpu_max[j] < nticks_cpu)
									cpu_max[j] = nticks_cpu;
								if (mpi_max[j] < nticks_mpi)
									mpi_max[j] = nticks_mpi;
								if (epoch_max[j] < i)
									epoch_max[j] = i;

    						} else { // chart type is bar or line
								if (nticks_cpu > 0) {
									table_thread_cpu_map[j]->InsertNextBlankRow();
									table_thread_cpu_map[j]->SetValue(table_thread_cpu_map[j]->GetNumberOfRows()-1, 0, i);
									table_thread_cpu_map[j]->SetValue(table_thread_cpu_map[j]->GetNumberOfRows()-1, 1, nticks_cpu);
									if (cpu_max[j] < nticks_cpu)
										cpu_max[j] = nticks_cpu;
									if (epoch_max[j] < i)
										epoch_max[j] = i;
								}
								if (nticks_mpi > 0) {
									table_thread_mpi_map[j]->InsertNextBlankRow();
									table_thread_mpi_map[j]->SetValue(table_thread_mpi_map[j]->GetNumberOfRows()-1, 0, i);
									table_thread_mpi_map[j]->SetValue(table_thread_mpi_map[j]->GetNumberOfRows()-1, 1, nticks_mpi);
									if (mpi_max[j] < nticks_mpi)
										mpi_max[j] = nticks_mpi;
									if (epoch_max[j] < i)
										epoch_max[j] = i;
								}
    						}
    					}
					}
    			}
    		}
    	}

    	// Display the chart(s)
    	bool labeled = false;
    	for (int i = 0; i <= max_tid_; ++i) {
    		FTQEpoch* t_epoch = thread_epochs_[i];
    		if (t_epoch) {
    			  line_cpu_map[i]->SetInput(table_cpu_map[i], 0, 1);
    			  line_cpu_map[i]->SetColor(255, 0, 0, 255);
    			  line_cpu_map[i]->SetWidth(1.0);
    			  // Show the legend for the first thread since, same thing will be repeated per thread
    			  if (!labeled)
    				  line_cpu_map[i]->SetLabel("CPU");
    			  else
    				  line_cpu_map[i]->SetLabel("");

    			  line_mpi_map[i]->SetInput(table_mpi_map[i], 0, 1);
    			  line_mpi_map[i]->SetColor(0, 255, 0, 255);
    			  line_mpi_map[i]->SetWidth(1.0);
    			  // Show the legend for the first thread since, same thing will be repeated per thread
    			  if (!labeled) {
    				  line_mpi_map[i]->SetLabel("MPI");
    				  labeled = true;
    			  } else
    				  line_mpi_map[i]->SetLabel("");

    			  if (show_vis_thread_chart_) {
    				  if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
    					  if (show_vis_thread_chart_type_ == "area") {
    					  	  line_both_map[i]->SetUseIndexForXSeries(true);
    					  	  line_both_map[i]->SetInput(table_both_map[i]);
    					  	  line_both_map[i]->SetInputArray(1, "CPU");
    					  	  line_both_map[i]->SetInputArray(2, "MPI");

    					  	  // Due to bug in area plot with more than one Legend, show the last legend only
    					  	  line_both_map[i]->SetLabel("MPI");

    					  	  VTK_CREATE(vtkColorSeries,colorSeries);
    					  	  //colorSeries->SetColorScheme(vtkColorSeries::COOL);
    					  	  colorSeries->SetColorScheme(vtkColorSeries::WILD_FLOWER);
    					  	  line_both_map[i]->SetColorSeries(colorSeries);
    					  } else {
							  if (table_thread_cpu_map[i]->GetNumberOfRows() > 1) {
								  if (show_vis_thread_chart_separate_ || ((!show_vis_thread_chart_separate_) && (i <= 8))) {
									  thread_cpu_plot_map[i]->SetInput(table_thread_cpu_map[i], 0, 1);
									  thread_cpu_plot_map[i]->SetColor(255, 0, 0, 255);
									  thread_cpu_plot_map[i]->SetWidth(0.5);
									  thread_cpu_plot_map[i]->SetLabel(boost::lexical_cast<std::string>(i) + " CPU");
								  }
							  }

							  if (table_thread_mpi_map[i]->GetNumberOfRows() > 1) {
								  if (show_vis_thread_chart_separate_ || ((!show_vis_thread_chart_separate_) && (i <= 8))) {
									  thread_mpi_plot_map[i]->SetInput(table_thread_mpi_map[i], 0, 1);
									  thread_mpi_plot_map[i]->SetColor(0, 255, 0, 255);
									  thread_mpi_plot_map[i]->SetWidth(0.5);
									  thread_mpi_plot_map[i]->SetLabel(boost::lexical_cast<std::string>(i) + " MPI");
								  }
							  }

    					  }
						  chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetRange(0, std::max(cpu_max[i], mpi_max[i])+1);
						  chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetRange(0, epoch_max[i]+1);

						  if (show_vis_thread_chart_separate_) { // Show one thread per window
							  view_thread[i]->GetRenderWindow()->SetMultiSamples(0);
							  chart_thread[i]->Update();
							  view_thread[i]->GetInteractor()->Initialize();

							  view_thread[i]->GetInteractor()->Render();
							  view_thread[i]->GetInteractor()->Enable();
						  }
						  else { // Show only charts applicable for 9 threads per chart case
							  view_thread[i/9]->GetRenderWindow()->SetMultiSamples(0);
							  chart_thread[i]->Update();
							  view_thread[i/9]->GetInteractor()->Initialize();

							  view_thread[i/9]->GetInteractor()->Render();
							  view_thread[i/9]->GetInteractor()->Enable();

						  }
					  }
    			  }
    		}
    	}

      chart->GetAxis(vtkAxis::LEFT)->SetRange(0, max_tid_+2);
	  chart->GetAxis(vtkAxis::BOTTOM)->SetRange(0, maximum_epoch+1);
   	  view->GetRenderWindow()->SetMultiSamples(0);
   	  chart->Update();
   	  view->GetInteractor()->Initialize();

   	  view->GetInteractor()->Render();
   	  view->GetInteractor()->Enable();
#endif
    }

    void
    AppFTQCalendar::dump()
    {
        std::string fname = sstprintf("ftq_app%d.csv", aid_);
        std::ofstream out(fname.c_str());
        //print the first line header
        out << sstprintf("%12s", "Epoch (us)");
        for (int i=0; i < KEY_NUM_EVENT_TYPES; ++i){
            out << sstprintf(",%12s", key::tostr((key::event_type_t) i));
        }
        out << "\n";
        timestamp one_us(1e-6);
        int64_t ticks_us = one_us.ticks_int64();
        for (int ep=0; ep < max_epoch_; ++ep){
            //figure out how many us
            int64_t num_us = (ep * num_ticks_epoch_) / ticks_us;
            out << sstprintf("%12ld", num_us);
            for (int i=0; i < KEY_NUM_EVENT_TYPES; ++i){
                out << sstprintf(",%12ld", epochs_[ep].get_event_time((key::event_type_t) i));
            }
            out << "\n";
        }
        out.close();

        fname = sstprintf("nodes_app%d.out", aid_);
        std::ofstream node_out(fname.c_str());
        for (int i=0; i <= max_tid_; ++i){
            TaskFTQCalendar::ptr cal = calendars_[i];
            if (cal){
                cal->dump(node_out);
            }
            else
                node_out << "\n";
        }
        node_out.close();



         timestamp stamp_sec = timestamp(1);
         int64_t ticks_s = stamp_sec.ticks_int64();
         std::cout << sstprintf("Average time stats for application %s: \n", appname_.c_str());
         /** print the stat totals */
         int ntasks = max_tid_ + 1;
         for (int i=0; i < KEY_NUM_EVENT_TYPES; ++i){
             key::event_type_t ty = (key::event_type_t) i;
             double av_per_app = (double) totals_[i] / ntasks;
             double t_sec = av_per_app / (double) ticks_s;
             std::cout << sstprintf("%12s: %16.8f s\n", key::tostr(ty), t_sec);
         }
#ifdef HAVE_VTK
         if (show_vis_) {
        	 display_charts(max_tid_, max_epoch_);
        	 std::cout << "--- Please close the chart window to exit \n";
			 view->GetInteractor()->Render();
			 view->GetInteractor()->Enable();
			 view->GetInteractor()->Start();
         }
#endif
     }

    TaskFTQCalendar::TaskFTQCalendar()
        : head_(0), tail_(0)
    {
    }

    void
    TaskFTQCalendar::dump(std::ofstream &os)
    {
        char ev_chars[KEY_NUM_EVENT_TYPES];
        ev_chars[key::mpi] = 'M';
        ev_chars[key::compute] = 'C';
        ev_chars[key::sleep] = 'S';
        ev_chars[key::general] = 'G';
        ev_chars[key::buffer] = 'B';

        long num_ticks_per_char = 1000 * 1000 * 100;
        long num_chars = max_tick_ / num_ticks_per_char;

        char* newline = new char[num_chars+1];
        newline[num_chars] = '\0';
        ::memset(newline, ' ' , num_chars);
        EventNode* node = head_;
        while (node){
            long idx_begin = node->ticks_begin / num_ticks_per_char;
            long idx_end = node->ticks_end / num_ticks_per_char;
            char next_char = ev_chars[node->ev_type];
            for (long idx=idx_begin; idx < idx_end; ++idx){
                newline[idx] = next_char;
            }
            node = node->next;
        }
        os << newline;
        os << "\n";
    }

    void
    TaskFTQCalendar::collect(key::event_type_t ty, long ticks_begin, long ticks)
    {
        if (tail_){
            if (tail_->ev_type == ty){
                //combine the two events
                tail_->ticks_end = ticks_begin + ticks;
            }
            else {
                EventNode* node = new EventNode;
                node->ticks_begin = ticks_begin;
                node->ticks_end = ticks_begin + ticks;
                node->next = 0;
                node->ev_type = ty;
                tail_->next = node;
                tail_ = node;
            }
        }
        else {
            EventNode* node = new EventNode;
            node->ticks_begin = ticks_begin;
            node->ticks_end = ticks_begin + ticks;
            node->next = 0;
            node->ev_type = ty;
            tail_ = head_ = node;
        }
        max_tick_ = ticks_begin + ticks;
    }

    FTQEpoch::FTQEpoch()
    {
        for (int i=0; i < KEY_NUM_EVENT_TYPES; ++i){
            totals_[i] = 0;
        }
    }

    void
    FTQEpoch::collect(key::event_type_t ty, long ticks)
    {
        totals_[ty] += ticks;
    }

    long long
    FTQEpoch::get_event_time(key::event_type_t ty) const
    {
        return totals_[ty];
    }


    }
}
