/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_SSTMAC_GLOBAL_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_SSTMAC_GLOBAL_H_INCLUDED

#include <sstream>

#include <boost/unordered_map.hpp>
#include <sstmac/common/logger.h>
#include <sstmac/software/process/process_context.h>

#include <sstmac/software/process/os_api.h>
#include <sstmac/software/process/thread.h>

#define SSTMAC_INT sstmac::sw::sstmac_global_builtin<int>
#define SSTMAC_LONG sstmac::sw::sstmac_global_builtin<long>
#define SSTMAC_DOUBLE sstmac::sw::sstmac_global_builtin<double>
#define SSTMAC_BOOL sstmac::sw::sstmac_global_builtin<bool>

namespace sstmac
{
  namespace sw
  {

    class sstmac_global
    {
    protected:
      process_context::ptr
      get_current_context() const
      {
        thread::ptr t = OS_API::get_current_thread();
        if (t)
          return t->get_process_context();
        else
          return NULL;
      }

    public:
      virtual std::string
      to_string() const
      {
        return "sstmac_global base class";
      }
    };

    template<typename T>
      class sstmac_global_builtin : public sstmac_global
      {

      private:
        typedef typename boost::unordered_map<long, T> val_map;
        val_map vals_;
        T init_;

        bool
        check_init(process_context::ptr ptxt) const
        {
          val_map& vals = const_cast<val_map&> (vals_);
          if (ptxt)
          {
            typename val_map::iterator it = vals.find(ptxt->get_key());
            if (it == vals.end())
            {
              vals[ptxt->get_key()] = init_;
            }
            return true;
          }
          else
            return false;
        }

      public:
        typedef typename val_map::const_iterator const_iterator;

        explicit
        sstmac_global_builtin() :
          init_(0)
        {
        }

        explicit
        sstmac_global_builtin(T init) :
          init_(init)
        {
        }

        sstmac_global_builtin(const sstmac_global_builtin<T>& other)
        {
          sst_throw(illformederror, "copy constructor should never be called for primitive global");
        }

        virtual
        ~sstmac_global_builtin()
        {

        }

        void
        print_all() const
        {
          typename val_map::const_iterator it = vals_.begin();
          for (; it != vals_.end(); ++it)
          {
            std::cout << it->second << std::endl;
          }
        }

        virtual T&
        get_val() const
        {
          process_context::ptr ptxt = get_current_context();
          val_map& vals = const_cast<val_map&> (vals_);
          if (ptxt)
          {
            typename val_map::iterator it = vals.find(ptxt->get_key());
            if (it == vals.end())
            {
              T& ret = vals[ptxt->get_key()];
              ret = init_;
              return ret;
            }
            else
            {
              return it->second;
            }
          }
          return const_cast<T&> (init_);
        }

        virtual std::string
        to_string() const
        {
          std::stringstream ss;
          ss << get_val();
          return ss.str();
        }

        template<typename U>
          T&
          operator =(const U& b)
          {
            T& myval = get_val();
            myval = (T) b;
            return myval;
          }

        template<typename U>
          T&
          operator +=(const U& b)
          {
            T& myval = get_val();
            myval += (T) b;
            return myval;
          }

        template<typename U>
          T&
          operator -=(const U& b)
          {
            T& myval = get_val();
            myval -= (T) b;
            return myval;
          }

        template<typename U>
          bool
          operator ==(const U& b) const
          {
            T myval = get_val();
            T otherval = (T) b;
            return myval == otherval;
          }

        template<typename U>
          bool
          operator !=(const U& b) const
          {
            T myval = get_val();
            T otherval = (T) b;
            return myval != otherval;
          }

        T
        operator ++(int dummy)
        {
          T& myval = get_val();
          return myval++;
        }

        T
        operator --(int dummy)
        {
          T& myval = get_val();
          return myval--;
        }

        T
        operator ++()
        {
          T& myval = get_val();
          return ++myval;
        }

        T
        operator --()
        {
          T& myval = get_val();
          return --myval;
        }

        operator double() const
        {
          return (double) get_val();
        }

        operator int() const
        {
          return (int) get_val();
        }

        operator bool() const
        {
          return (bool) get_val();
        }

        operator size_t() const
        {
          return (size_t) get_val();
        }

        operator long() const
        {
          return (long) get_val();
        }

        operator long long() const
        {
          return get_val();
        }

        template<typename U>
          T
          operator +(const U& b) const
          {
            return (get_val() + (T) b);
          }

        template<typename U>
          T
          operator -(const U& b) const
          {
            return (get_val() - (T) b);
          }

        template<typename U>
          T
          operator /(const U& b) const
          {
            return (get_val() / ((T) b));
          }

        template<typename U>
          T
          operator <<(const U& b) const
          {
            return (get_val() << ((T) b));
          }

        template<typename U>
          T
          operator >>(const U& b) const
          {
            return (get_val() >> ((T) b));
          }

        template<typename U>
          T
          operator %(const U& b) const
          {
            return get_val() % ((T) b);
          }

        template<typename U>
          T
          operator *(const U& b) const
          {
            return (get_val() * ((T) b));
          }

        T*
        operator&() const
        {
          return &get_val();
        }

        template<typename U>
          bool
          operator >(const U& b)
          {
            T myval = get_val();
            T otherval = (T) b;
            return myval > otherval;
          }

        template<typename U>
          bool
          operator <(const U& b)
          {
            T myval = get_val();
            T otherval = (T) b;
            return myval < otherval;
          }

        void
        delete_vals()
        {
          if (init_)
          {
            delete init_;
            init_ = 0;
          }
          typedef typename boost::unordered_map<long, T>::iterator myiter;
          myiter it;
          for (it = vals_.begin(); it != vals_.end(); ++it)
            if (it->second)
              delete it->second;
          vals_.clear();
        }

        const_iterator
        begin() const
        {
          return vals_.begin();
        }
        const_iterator
        end() const
        {
          return vals_.end();
        }

      };

    inline std::ostream&
    operator<<(std::ostream &os, sstmac_global &g)
    {
      return (os << g.to_string());
    }

    inline logger&
    operator<<(logger &os, sstmac_global &g)
    {
      return (os << g.to_string());
    }

    template<typename T, typename U>
      U
      operator/(const sstmac_global_builtin<U>& b,
          const sstmac_global_builtin<T>& t)
      {
        return ((U) b) / ((T) t);
      }

    template<typename T, typename U>
      U
      operator+(const sstmac_global_builtin<U>& b,
          const sstmac_global_builtin<T>& t)
      {
        return (U) b + (T) t;
      }

    template<typename T, typename U>
      U
      operator-(const sstmac_global_builtin<U>& b,
          const sstmac_global_builtin<T>& t)
      {
        return (U) b - (T) t;
      }

    template<typename T, typename U>
      U
      operator*(const sstmac_global_builtin<U>& b,
          const sstmac_global_builtin<T>& t)
      {
        return ((U) b) * ((T) t);
      }

    template<typename T, typename U>
      U
      operator/(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b / ((T) t);
      }

    template<typename T, typename U>
      U
      operator+(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b + (T) t;
      }

    template<typename T, typename U>
      U
      operator-(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b - (T) t;
      }

    template<typename T, typename U>
      U
      operator*(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b * ((T) t);
      }

    template<typename T, typename U>
      U
      operator%(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b % ((T) t);
      }

    template<typename T, typename U>
      U
      operator+=(U& b, const sstmac_global_builtin<T>& t)
      {
        b += (T) t;
        return b;
      }

    template<typename T, typename U>
      U
      operator-=(const U& b, const sstmac_global_builtin<T>& t)
      {
        b -= (T) t;
        return b;
      }

    template<typename T, typename U>
      bool
      operator<(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b < (T) t;
      }

    template<typename T, typename U>
      bool
      operator>(const U& b, const sstmac_global_builtin<T>& t)
      {
        return b > (T) t;
      }

    template<typename T>
      class sstmac_global_builtin<T*> : public sstmac_global
      {

      protected:
        typedef typename boost::unordered_map<long, T*> val_map;
        typedef T* Tptr;
        val_map vals_;
        T* init_;

        bool
        check_init(process_context::ptr ptxt)
        {
          if (ptxt)
          {
            if (vals_.find(ptxt->get_key()) == vals_.end())
              vals_[ptxt->get_key()] = init_;
            return true;
          }
          else
            return false;
        }

      public:
        typedef typename val_map::const_iterator const_iterator;

        explicit
        sstmac_global_builtin() :
          init_(NULL)
        {
        }

        explicit
        sstmac_global_builtin(T* init) :
          init_(init)
        {
        }

        virtual
        ~sstmac_global_builtin()
        {
        }

        virtual T*&
        get_val() const
        {
          process_context::ptr ptxt = get_current_context();
          val_map& vals = const_cast<val_map&> (vals_);
          if (ptxt)
          {
            typename val_map::iterator it = vals.find(ptxt->get_key());
            if (it == vals.end())
            {
              T*& ret = vals[ptxt->get_key()] = init_;
              return ret;
            }
            else
            {
              return it->second;
            }
          }
          return const_cast<T*&> (init_);
        }

        virtual std::string
        to_string() const
        {
          std::stringstream ss;
          ss << get_val();
          return ss.str();
        }

        template<typename U>
          T*&
          operator=(const sstmac_global_builtin<U>& b)
          {
            T*& myval = get_val();
            myval = (T*) b;
            return myval;
          }

        template<typename U>
          T*&
          operator=(U* b)
          {
            T*& myval = get_val();
            myval = (T*) b;
            return myval;
          }

        T*
        operator++()
        {
          Tptr& p = get_val();
          return ++p;
        }

        T*
        operator--()
        {
          Tptr& p = get_val();
          return --p;
        }

        T*
        operator++(int dummy)
        {
          Tptr& p = get_val();
          return p++;
        }

        T*
        operator--(int dummy)
        {
          Tptr& p = get_val();
          return p--;
        }

        template<typename U>
          T*
          operator+(const U& b)
          {
            Tptr p = get_val();
            return p + b;
          }

        template<typename U>
          T*
          operator-(const U& b)
          {
            Tptr p = get_val();
            return p - b;
          }

        operator char*()
        {
          return (char*) get_val();
        }

        operator int*()
        {
          return (int*) get_val();
        }

        operator long*()
        {
          return (long*) get_val();
        }

        operator void*()
        {
          return get_val();
        }

        operator const char*() const
        {
          return (const char*) get_val();
        }

        operator const int*() const
        {
          return (const int*) get_val();
        }

        operator const long*() const
        {
          return (const long*) get_val();
        }

        operator const void*() const
        {
          return (const void*) get_val();
        }

        T&
        operator*()
        {
          return *(get_val());
        }

        template<typename U>
          bool
          operator ==(const U& b) const
          {
            T* myval = get_val();
            T* otherval = (T*) b;
            return myval == otherval;
          }

        T*
        operator->()
        {
          return get_val();
        }

        T&
        operator[](int idx)
        {
          T* arr = get_val();
          return arr[idx];
        }

        T&
        operator[](long idx)
        {
          T* arr = get_val();
          return arr[idx];
        }

        const_iterator
        begin() const
        {
          return vals_.begin();
        }
        const_iterator
        end() const
        {
          return vals_.end();
        }

      };

    template<typename T>
      class sstmac_global_struct
      {
      };

    template<typename T>
      class sstmac_global_struct<T*> : public sstmac_global_builtin<T*>
      {
        typedef sstmac_global_builtin<T*> parent;
      public:
        explicit
        sstmac_global_struct()
        {
        }

        explicit
        sstmac_global_struct(T* init) :
          parent(init)
        {
        }

        operator T*()
        {
          return parent::get_val();
        }

        operator const T*() const
        {
          return parent::get_val();
        }

        template<typename U>
          T*&
          operator=(const sstmac_global_struct<U>& b)
          {
            T*& myval = parent::get_val();
            myval = (T*) b;
            return myval;
          }

        T*&
        operator=(T* b)
        {
          T*& myval = parent::get_val();
          myval = b;
          return myval;
        }

        template<typename U>
          T*&
          operator=(U* b)
          {
            T*& myval = parent::get_val();
            myval = (T*) b;
            return myval;
          }

      };

  }
}

#ifdef SSTMAC_OFF
typedef double* global_double_ptr;
typedef double global_double;
typedef char* global_char_ptr;
typedef char global_char;
typedef long* global_long_ptr;
typedef long global_long;
typedef int* global_int_ptr;
typedef int global_int;
typedef long long global_long_long;
typedef bool* global_bool_ptr;
typedef bool global_bool;
#else
typedef sstmac::sw::sstmac_global_builtin<double*> global_double_ptr;
typedef sstmac::sw::sstmac_global_builtin<double> global_double;
typedef sstmac::sw::sstmac_global_builtin<char*> global_char_ptr;
typedef sstmac::sw::sstmac_global_builtin<char> global_char;
typedef sstmac::sw::sstmac_global_builtin<long*> global_long_ptr;
typedef sstmac::sw::sstmac_global_builtin<long> global_long;
typedef sstmac::sw::sstmac_global_builtin<int*> global_int_ptr;
typedef sstmac::sw::sstmac_global_builtin<int> global_int;
typedef sstmac::sw::sstmac_global_builtin<long long> global_long_long;
typedef sstmac::sw::sstmac_global_builtin<bool*> global_bool_ptr;
typedef sstmac::sw::sstmac_global_builtin<bool> global_bool;
#endif

#endif /* SSTMAC_GLOBAL_H_ */
