#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/allocation/cartallocation.h>
#include <sstmac/common/factories/factory.h>


using sstmac::hw::hdtorus;

namespace sstmac {
  namespace sw {
    
    SSTRegister("cart | cartesian", allocationstrategy, cartallocation);

    cartallocation::cartallocation(sim_parameters::ptr params) :
      allocationstrategy(params->get_long_param("node_num"))
    {
      dimension_ = 2;
      if( params->has_param("launch_zoff") ||
	  params->has_param("launch_zsize") ) {
	if( params->has_param("launch_zoff") &&
	    params->has_param("launch_zsize") ) {
	  dimension_ = 3;
	}
	else
	  throw valueerror(
	      "cartallocation::construct - 3D requires both zoff and zsize");
      }

      size_ = new long[dimension_];
      offset_ = new long[dimension_];

      size_[0] = params->get_long_param("launch_xsize");
      size_[1] = params->get_long_param("launch_ysize"); 
      offset_[0] = params->get_long_param("launch_xoff");
      offset_[1] = params->get_long_param("launch_yoff");       
      
      if( dimension_ == 3 ) {
	size_[2] = params->get_long_param("launch_zsize");
	offset_[2] = params->get_long_param("launch_zoff");
      }
    }

    cartallocation::ptr
    cartallocation::construct(sim_parameters::ptr params)
    {
      cartallocation::ptr retval(new cartallocation(params));
      return retval;
    }
    
    void
    cartallocation::set_topology(const sstmac::hw::topology::ptr& top)
    {
      allocationstrategy::set_topology(top);

      sstmac::hw::hdtorus::ptr network = boost::dynamic_pointer_cast<
          sstmac::hw::hdtorus>(top);
      if (!network)
        {
          throw valueerror(
	    "cartallocation: error in casting the topology to hdtorus");
        }

      network_ = network;

      if (network.get() == 0) {
	throw sstmac::valueerror("cartallocation: hdtorus network is null");
      }
      int netdim = network->dimensions().size();
      if (netdim != 2 && netdim != 3) {
	throw sstmac::valueerror("cartallocation: need 2D or 3D hdtorus");
      }
      for(int i=0; i<dimension_; ++i) {
	if( size_[i] < 0 )
	  throw sstmac::valueerror("cartallocation: size < 0");
      }
      if (size_[0] > network->dimensions()[0] ||
          size_[1] > network->dimensions()[1] ) {
	throw sstmac::valueerror("cartallocation: bad size");
      }
      if( dimension_ == 3 ) {
	  if (size_[2] > network->dimensions()[2] )
	    throw sstmac::valueerror("cartallocation: bad size");
      }

    }

    void
    cartallocation::insert_nodeid (node_set &allocation,
				   sstmac::hw::vertex &xyz) 
    {
      // convert into nodeid
      nodeaddress::ptr nid =
	nodeaddress::convert_from_uniqueid(network_->node_to_index(xyz));
      std::set<nodeaddress::ptr>::iterator niter = available_.find(nid);
      if (niter == available_.end()) {
	throw sstmac::valueerror("cartallocation: node not available");
      }
      allocation.insert(nid);      
    }
      
    void
    cartallocation::allocate(int nnode, node_set &allocation)
    {

      if (nnode > (int) available_.size()) {
	sst_throw(sstmac::valueerror,
		  "cartallocation: requested %d nodes, only %ld are available",
		  nnode, available_.size());
      }

      if( dimension_ == 2 ) {
        sstmac::hw::vertex xyz(3);
	xyz[2] = 0;
	for (int x = offset_[0]; x < offset_[0] + size_[0]; x++) {
	  for (int y = offset_[1]; y < offset_[1] + size_[1]; y++) {
	    xyz[0] = x;
	    xyz[1] = y;
	    insert_nodeid( allocation, xyz); 
	  }
	}
      }
      else {
        sstmac::hw::vertex xyz(4);
        xyz[3] = 0;
	for (int x = offset_[0]; x < offset_[0] + size_[0]; x++) {
	  for (int y = offset_[1]; y < offset_[1] + size_[1]; y++) {
	    for(int z = offset_[2]; z < offset_[2] + size_[2]; z++) {
	    xyz[0] = x;
	    xyz[1] = y;
	    xyz[2] = z;
	    insert_nodeid( allocation, xyz); 
	    }
	  }
	}
      }
			  
      for (std::set<sstmac::nodeaddress::ptr>::iterator i =
	     allocation.begin(); i != allocation.end(); i++) {
	available_.erase(*i);
	allocated_.insert(*i);
      }

    }
			 
  } // end namespace sw
} // end namespace sstmac
