/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/memory/numa_memorymodel.h>
#include <sstmac-eiger/eigernode.h>
#include <sstmac/hardware/processor/migratingcomputescheduler.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <sstmac/software/libraries/filesystem/nppfs_server.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("eiger | eigernode", node, eigernode,
                "eigernode contains an eiger processor capable of handling compute calls to eiger models" );

    using namespace sstmac::sw;

    eigernode::eigernode(const sim_parameters::ptr &params)
        : simplenode(params)
    {
    }

    eigernode::eigernode(const sim_parameters::ptr &params,
        nodeaddress::ptr my_id) : simplenode(params, my_id)
    {
    }

    void
    eigernode::initialize(eventmanager::ptr man)
    {

			simplenode::initialize(man);

      SSTMAC_DEBUG << "starting eigernode initialization \n";

      double tcycle = 1.0 / (freq_);
      double mop_freq = params_->get_optional_freq_param("node_memop_frequency", freq_);
      double memop_cycle = 1.0 / mop_freq;

			sstmac::hw::instructionprocessor_config::ptr pcfg;
			pcfg = sstmac::hw::instructionprocessor_config::construct(ncores_,
                    eventdata::INST_FLOP, tcycle, eventdata::INST_MEMOP, memop_cycle,
                    eventdata::INST_MEMOP_LOCAL, memop_cycle, eventdata::INST_MEMOP_RAND,
					memop_cycle);

			proc_eiger_ = sstmac::hw::eigerprocessor::construct(pcfg, msg_handler_, compute_sched_);

      SSTMAC_DEBUG << "eigernode initialization done, starting things \n";
			/* ???
      os_->start_apps();

      launcher_->start();
			*/

    }

    node::ptr
    eigernode::clone(nodeaddress::ptr my_id) const
    {
      node::ptr ret = construct(params_, my_id);

      return ret;
    }

    // --- sstmac::sw::AMI pure virtuals ---- //

    void
    eigernode::execute_kernel(AMI::AMI_COMM_FUNC func, const boost::intrusive_ptr<sst_message>& data)
    {
      bool failed = false;
      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        {

          if (data->get_type() == sst_message::FILEMETA)
            {
              data->toaddr() = nppfs_server_;

              if (!nppfs_server_)
                throw ssterror(
                    "eigernode: I don't have an address for the nppfs server");
            }

          nodeaddress::ptr nid = boost::dynamic_pointer_cast<nodeaddress>(
              data->toaddr());
          data->fromaddr() = nodeid_;

          SSTMAC_DEBUG << "Node(" << nodeid_->to_string() << "): sending to "
                << data->toaddr()->to_string() << "\n";

          if (data->toaddr()->equals(nodeid_))
            {
              schedule(now(), msg_handler_, data);
              // msg_handler_->handle(data);
            }
          else
            nic_->send(nid, data);

          failed = nic_->has_failed() || mem_model_->has_failed();
          break;
        }
      default:
        sst_throw(ssterror, "eigernode: cannot process kernel %s", AMI::tostr(func));

        }
    }

    void
    eigernode::execute_kernel(AMI::AMI_COMP_FUNC func, const boost::intrusive_ptr<sst_message>& data)
    {
      bool failed = false;

      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMP_TIME:
        {

          proc_model_->compute(data, os_->get_current_threadid());

          failed = compute_sched_->has_failed();

          break;
        }

		case sstmac::sw::AMI::AMI_COMP_INSTR_TIME:
        {
            data->set_event_time(proc_inst_->compute_time(data, os_->get_current_threadid()));
            break;
        }
      case sstmac::sw::AMI::AMI_COMP_INSTR:
        {

          proc_inst_->compute(data, os_->get_current_threadid());

          failed = compute_sched_->has_failed();

          break;
        }

			case sstmac::sw::AMI::AMI_COMP_EIGER:
				{

          proc_eiger_->compute(data, os_->get_current_threadid());

          failed = compute_sched_->has_failed();

          break;
				}

      case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
        {
          sstmac::sw::disk_message::ptr dm = boost::dynamic_pointer_cast<
              sstmac::sw::disk_message>(data);
          disk_->request_access(data);

          failed = disk_->has_failed();
          break;
        }

	  case sstmac::sw::AMI::AMI_COMP_SCHEDULE:
	    {
		  compute_sched_->reserve(data, os_->get_current_threadid());
		  break;
	    }

	  case sstmac::sw::AMI::AMI_COMP_RELEASE:
	    {
		  compute_sched_->release(data, os_->get_current_threadid());
		  break;
	    }

      case sstmac::sw::AMI::AMI_COMP_MEMREAD:
        {
          mem_model_->read(data);
          mem_model_->has_failed();
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
        {
          mem_model_->write(data);
          mem_model_->has_failed();
          break;
        }

	  case sstmac::sw::AMI::AMI_COMP_NIC_READ:
        {
            mem_model_->nic_read(data);
            failed = mem_model_->has_failed();
            break;
        }
		case sstmac::sw::AMI::AMI_COMP_NIC_WRITE:
        {
            mem_model_->nic_write(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_RDMA_GET:
        {
            mem_model_->rdma_get(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_RDMA_PUT:
        {
            mem_model_->rdma_put(data);
            failed = mem_model_->has_failed();
            break;
        }
 
      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        {
          timestamp delay = now() + safe_cast<timed_message>(data)->get_time();
          SSTMAC_DEBUG << "node (" << nodeid_->to_string()
                << "): sleeping until " << delay << "\n";
          schedule(delay, ptr(this), data);
          break;
        }

      default:
        sst_throw(ssterror, "eigernode: cannot process kernel %s", AMI::tostr(func));

        }

    }

		bool
		eigernode::kernel_supported(AMI::AMI_COMP_FUNC func) const
		{
			switch(func)
			{
				case sstmac::sw::AMI::AMI_COMP_TIME:
				case sstmac::sw::AMI::AMI_COMP_SLEEP:
				case sstmac::sw::AMI::AMI_COMP_INSTR:
				case sstmac::sw::AMI::AMI_COMP_MEMREAD:
				case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
				case sstmac::sw::AMI::AMI_COMP_EIGER:
					return true;
				case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
					return disk_;
                default:
                    return false;
			}
			return false;
		}

  }
} // end of namespace sstmac.
