/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "stearleyallocator.h"
#include <algorithm>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("stearley", allocationstrategy, stearleyallocator);

    stearleyallocator::stearleyallocator(long num) :
      allocationstrategy(num)
    {

    }

    void
    stearleyallocator::nodes_good(const node_set& good)
    {
      int init = suspects_.size();
      if (init > 0)
        {
          node_set::iterator it, end = good.end();
          for (it = good.begin(); it != end; it++)
            {

              suspects_.erase(*it);

            }
          SSTMAC_DEBUG << "stearleyallocator: we cut suspects by " << (init
                - suspects_.size()) << "\n";

          if (suspects_.size() == 0)
            {
              SSTMAC_DEBUG
                    << "stearleyallocator: it would seem that our problem has gone away! \n";
            }
        }

      check_nodes();
    }

    void
    stearleyallocator::nodes_suspect(const node_set& sus)
    {
      int init = suspects_.size();
      if (init != 0)
        {
          node_set::iterator it, end = suspects_.end();
          node_set diff;

          for (it = suspects_.begin(); it != end; it++)
            {
              if (sus.find(*it) == sus.end())
                {
                  diff.insert(*it);
                }
            }

          if (diff.size() == init)
            {
              throw ssterror(
                  "stearleyallocator: it seems we have more than one dead node, and this is not implemented yet");
            }
          else
            {
              SSTMAC_DEBUG << "stearleyallocator: cutting suspects by "
                    << diff.size() << "\n";
            }

          end = diff.end();
          for (it = diff.begin(); it != end; it++)
            {
              suspects_.erase(*it);
            }

        }
      else
        {
          SSTMAC_DEBUG << "stearleyallocator: starting to suspect \n";
          suspects_.insert(sus.begin(), sus.end());
        }

      check_nodes();
    }

    void
    stearleyallocator::check_nodes()
    {
      if (suspects_.size() == 1)
        {
          SSTMAC_DEBUG << "stearleyallocator: we found that SoB, it's "
                << *suspects_.begin() << ". removing from available. \n";
          available_.erase(*suspects_.begin());
          suspects_.clear();
        }
    }

    void
    stearleyallocator::allocate(int nnode_requested, node_set &allocation)
    {
      if (nnode_requested > available_.size())
        {
          return;
        }
      else
        {

          int left_to_alloc = nnode_requested;
          if (suspects_.size() > 0)
            {
              SSTMAC_DEBUG
                    << "stearleyallocator: we have some suspect nodes, so returning those first \n";

              node_set::iterator it, end = suspects_.end();

              for (it = suspects_.begin(); left_to_alloc > 0 && it != end; it++)
                {
                  if (available_.find(*it) != available_.end())
                    {
                      allocation.insert(*it);
                      allocated_.insert(*it);
                      available_.erase(*it);
                      left_to_alloc--;
                    }
                }


            }

          // allocation.resize(nnode_requested);
          for (int i = 0; i < left_to_alloc; i++)
            {
              SSTMAC_DEBUG << "firstavailable_: i=" << i << "\n";
              node_set::const_iterator nextnode = available_.begin();
              SSTMAC_DEBUG << "firstavailable_: nextnode= "
                    << (*nextnode)->to_string() << "\n";
              allocation.insert(*nextnode);
              allocated_.insert(*nextnode);
              available_.erase(nextnode);
            }
        }

    }

  }
}
