/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include "stearleylaunch.h"
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("stearley", launcher, stearleylaunch);
    // Pointer types.

    //
    // Protected constructor.
    //
    stearleylaunch::stearleylaunch() :
      reallaunch()
    {

    }

    service::ptr
    stearleylaunch::clone()
    {
      stearleylaunch::ptr ret = construct(empty_parameters::construct());
      ret->init(tolaunch_, alloc_, indexing_);
      return ret;
    }

    void
    stearleylaunch::init(std::list<launchinfo::ptr> &tolaunch,
        allocationstrategy::ptr &a, indexstrategy::ptr &ind)
    {
      reallaunch::init(tolaunch, a, ind);
      salloc_ = boost::dynamic_pointer_cast<stearleyallocator>(a);

      if (!salloc_)
        {
          throw ssterror(
              "stearleylaunch: needs a stearley allocator to operate correctly");
        }
    }

    //
    // Public constructor
    //
    stearleylaunch::ptr
    stearleylaunch::construct(sim_parameters::ptr params)
    {
      stearleylaunch::ptr retval(new stearleylaunch());

      return retval;
    }

    /**
     * This function is called when a node dies.  It is called at the scheduler node
     * when the kill message has arrived from the dead node.
     * @param msg
     */
    void
    stearleylaunch::msg_kill(kill_message::ptr msg)
    {
      reallaunch::msg_kill(msg);

      if (myid_ == master_->unique_id()) //make sure to send to everyone else
        {
          launchinfo::ptr linfo = msg->get_info();
          SSTMAC_DEBUG << "launcher[" << myid_
                << "]: some nodes are suspect \n";

          //type<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::set nodes;
          allocationstrategy::node_set nodes;
          for (long i = 0; i < envmap_[linfo->aid_]->get_app_count(); i++)
            {
              taskid tid(i);
              nodeaddress::ptr addr = envmap_[linfo->aid_]->get_node_address(
                  tid);
              nodes.insert(addr);

            }
          salloc_->nodes_suspect(nodes);
        }

    }

    /**
     * This function is called when an application completes normally.
     * @param cmsg
     */
    void
    stearleylaunch::msg_complete(complete_message::ptr cmsg)
    {
      reallaunch::msg_complete(cmsg);

      if (myid_ == master_->unique_id()) //make sure to send to everyone else
        {
          launchinfo::ptr linfo = cmsg->get_info();

          if (states_[linfo->aid_] == FINISHED)
            {
              //release allocation
              //type<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::set nodes;
              allocationstrategy::node_set nodes;
              for (long i = 0; i < envmap_[linfo->aid_]->get_app_count(); i++)
                {
                  taskid tid(i);
                  nodes.insert(envmap_[linfo->aid_]->get_node_address(tid));
                }

              salloc_->nodes_good(nodes);

            }
        }

    }

  }
} // end of namespace sstmac
