/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#include "ftpm.h"

#include <sstmac/software/libraries/mpi/sstmac_mpi.h>

namespace ftpm
{

  using namespace sstmac;
  using namespace sstmac::sw;

  void
  ftpm::skeleton_main()
  {
    int masters = sstmac_env::params->get_int_param("ftpm_num_masters");

    hw_temp_rate_ = sstmac_env::params->get_optional_double_param(
        "ftpm_hw_temp_rate", 0.0);
    hw_perm_rate_ = sstmac_env::params->get_optional_double_param(
        "ftpm_hw_perm_rate", 0.0);
    sw_rate_ = sstmac_env::params->get_optional_double_param("ftpm_sw_rate",
        0.0);
    avg_delay_ = sstmac_env::params->get_optional_time_param("ftpm_delay",
        timestamp(0));

    delay_rate_ = sstmac_env::params->get_optional_double_param(
        "ftpm_delay_rate", 0.0);

    request_timeout_ = sstmac_env::params->get_time_param(
        "ftpm_request_timeout");

    update_freq_ = sstmac_env::params->get_optional_int_param(
        "ftpm_update_freq", -1);

    int rank, size;

    /* initialize random seed: */
    srand(time(NULL));

    MPI_Init(NULL, NULL);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int masterdomain = size / masters; //size of domain, including master

    nwork_ = (size - masters); //number of workers in a domain

    int mdomsize = (size - masters) / masters;

    int mymastdom = rank / masterdomain;

    MPI_Comm mastercomm;
    if (masters > 0)
      {
        // --- create the masters group for collaborating --- //
        MPI_Group mastergroup;
        MPI_Group worldgroup;

        MPI_Comm_group(MPI_COMM_WORLD, &worldgroup);

        int masterranks[masters];

        for (int i = 0; i < masters; i++)
          {
            masterranks[i] = i * masterdomain;
          }

        MPI_Group_incl(worldgroup, masters, masterranks, &mastergroup);

        MPI_Comm_create(MPI_COMM_WORLD, mastergroup, &mastercomm);
        // ----------------------- //

      }

    if (rank % masterdomain == 0)
      { // I am a master
        SSTMAC_DEBUG << "ftpm[" << rank << "]: now I am the master \n";

        MPI_Request working[mdomsize];
        std::map<int, timestamp> send_times;
        std::map<int, int> node_timeouts;
        int good_nodes = mdomsize;

        m_generate_initial_data(rank, masters);

        if (update_freq_ > 0)
          {
            std::cout << "----- initial data ------ \n";
            std::cout << m_get_answer_string() << "\n";
          }

        for (int r = rank + 1; r < rank + masterdomain; r++)
          {
            //send args
            int w_id = r - (mymastdom + 1);
            m_isend_args(w_id, r);

            //recv result
            MPI_Request savereq = m_irecv_result(w_id, r);
            working[r - rank - 1] = savereq;
            send_times[r - rank - 1] = now();
            node_timeouts[r - rank - 1] = 0;

          }

        long update = 0;

        while (!m_done())
          {
            MPI_Status stat;
            int ind = -1;

            //----- cleanup old requests ----
            std::map<int, timestamp>::iterator it, end = send_times.end();
            for (it = send_times.begin(); it != end; it++)
              {
                if (node_timeouts[it->first] < ALLOWED_TIMEOUTS
                    && request_timeout_ + it->second < now())
                  {
                    SSTMAC_DEBUG << "ftpm[" << rank << "]: request from rank "
                          << it->first + rank + 1 << " has timed out \n";
                    MPI_Cancel(&working[it->first]);

                    //send new args back

                    m_isend_args(mymastdom * mdomsize + it->first,
                        it->first + rank + 1);

                    //recv result again
                    MPI_Request savereq = m_irecv_result(
                        mymastdom * mdomsize + it->first, it->first + rank + 1);
                    working[it->first] = savereq;

                    send_times[it->first] = now();

                    node_timeouts[it->first]++;

                    if (node_timeouts[it->first] >= ALLOWED_TIMEOUTS)
                      {
                        SSTMAC_DEBUG << "ftpm[" << rank
                              << "]: he's dead, Jim \n";
                        good_nodes--;
                      }
                  }
              }
            if (good_nodes <= 0)
              {
                SSTMAC_DEBUG << "ftpm[" << rank
                      << "]: master: I ran out of good nodes! They're all dead! \n";
                break;
              }
            //-------------------------------


            //------- master collaboration ------- //

            if (masters > 0 && m_collaborate())
              {
                /* double ret;
                 MPI_Allreduce(&error_, &ret, 1, MPI_DOUBLE, MPI_SUM, mastercomm);

                 SSTMAC_DEBUG << "ftpm[" << rank << "]: updating error to "
                 << ret / masters << "\n";
                 error_ = ret / masters;*/
              }

            //----------------------------------//


            //------- now we wait -----//
            MPI_Waitany_timeout(nwork_, working, &ind, &stat,
                request_timeout_.sec());
            //-------------------------//

            if (ind >= 0)
              {
                SSTMAC_DEBUG << "ftpm[" << rank
                      << "]: master received result from index " << ind
                      << ", which is rank " << ind + rank + 1 << "\n";

                double f_sw = rand() % 10000;
                m_unpack_result(mymastdom * mdomsize + ind, ind + rank + 1,
                    f_sw < sw_rate_ * 100);

                //send new args back
                m_isend_args(mymastdom * mdomsize + ind, ind + rank + 1);

                //recv result again
                MPI_Request savereq = m_irecv_result(
                    mymastdom * mdomsize + ind, ind + rank + 1);
                working[ind] = savereq;

                //update what the timeout should be
                timestamp diff = now() - send_times[ind];
                request_timeout_
                    = timestamp(
                        4.0 * (((request_timeout_.sec() / 4.0) + diff.sec())
                            / 2.0));

                SSTMAC_DEBUG << "request timeout is now " << request_timeout_
                      << "\n";

                send_times[ind] = now();

                update++;

                if (update_freq_ > 0 && update % update_freq_ == 0)
                  {
                    std::cout << "----- answer update ------ \n";
                    std::cout << m_get_answer_string() << "\n";
                  }
              }
            else
              {
                SSTMAC_DEBUG << "ftpm[" << rank
                      << "]: master got an invalid worker index back from mpi_waitany, which means we timed out \n";
                request_timeout_ = timestamp(request_timeout_.sec() * 1.2);
              }

          }
        SSTMAC_DEBUG << "ftpm[" << rank
              << "]: master is done, brutally suppressing the workers \n";
        //Send done signals to all my workers
        bool done = true;
        for (int r = rank + 1; r <= rank + mdomsize; r++)
          {
            MPI_Send(&done, 1, MPI_CHAR, r, DONE_TAG, MPI_COMM_WORLD);
          }

        SSTMAC_DEBUG << "ftpm[" << rank << "]: master reports the answer as: "
              << m_get_answer_string() << "\n";

        std::cout << "------- answer error = " << m_get_answer_err() << "\n";

      }
    else
      { // I am work slave

        int mymaster = mymastdom * masterdomain;
        work_id_ = rank - (mymastdom + 1);

        SSTMAC_DEBUG << "ftpm[" << rank << "]: worker " << work_id_
              << " with master " << mymaster << "\n";

        int done = 0;
        int ind = -1;

        MPI_Request waiting[2];
        MPI_Request done_req;
        MPI_Status stat;

        MPI_Irecv(&done, 1, MPI_CHAR, mymaster, DONE_TAG, MPI_COMM_WORLD,
            &done_req);
        MPI_Test(&done_req, &done, &stat);

        while (!done)
          {
            //------ model permanent hw failure ------
            double f_hw_perm = rand() % 10000;
            if (f_hw_perm < hw_perm_rate_ * 100)
              {
                SSTMAC_DEBUG << "ftpm[" << rank
                      << "]: worker node has permanently failed \n";

                //wait on the done req, so we call finalize when the master is ready
                MPI_Wait(&done_req, &stat);
                break;
              }
            //---------------------------------------
            SSTMAC_DEBUG << "ftpm[" << rank
                  << "]: worker starting work loop \n";

            MPI_Request argreq = w_irecv_args(mymaster);
            waiting[0] = done_req;
            waiting[1] = argreq;

            MPI_Waitany(2, waiting, &ind, &stat);

            if (ind == 0)
              {
                SSTMAC_DEBUG << "worker[" << rank
                      << "]: received done signal, I'm outta here \n";
                done = true;
                break;

              }
            else if (ind == 1)
              {
                //-------- model temp hw fault -----
                double f_hw_temp = rand() % 10000;
                if (f_hw_temp < hw_temp_rate_ * 100)
                  {
                    SSTMAC_DEBUG << "worker[" << rank
                          << "]: got some args, but the node has temporarily failed \n";

                  }
                else
                  {
                    //---------------------------------
                    SSTMAC_DEBUG << "worker[" << rank
                          << "]: got some args, lets work \n";

                    w_work(work_id_);

                    //----- model delay fault ------
                    double f_delay = rand() % 10000;
                    if (f_delay < delay_rate_ * 100)
                      {

                        SSTMAC_DEBUG << "ftpm[" << rank
                              << "]: worker result has been delayed by "
                              << avg_delay_ << "\n";
                        sleep(avg_delay_);
                      }
                    //-----------------------------


                    w_send_result(mymaster);
                  }
              }
            else
              {
                throw ssterror("ftpm -- something weird happened with a worker");
              }

          }

      }

    MPI_Finalize();
  }

}
