/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#include "solve_joint.h"

namespace ftpm
{
  SSTRegisterApp("solve_joint", solve_joint);

  void
  solve_joint::consume_params(const sim_parameters::ptr &params)
  {
    xmin_ = params->get_double_param("solve_joint_xmin");
    xmax_ = params->get_double_param("solve_joint_xmax");
    hh_ = params->get_double_param("solve_joint_hh");

    y0_ = params->get_double_param("solve_joint_y0");
    y1_ = params->get_double_param("solve_joint_y1");

    numsamp_ = params->get_int_param("solve_joint_num_samp");

    ymax_ = params->get_double_param("solve_joint_ymax");
    ymin_ = params->get_double_param("solve_joint_ymin");

    error_ = params->get_double_param("solve_joint_error");

    outlier_ = params->get_double_param("solve_joint_outlier");

    resample_freq_ = params->get_optional_int_param(
        "solve_joint_resample_freq", -1);

  }

  double
  solve_joint::solve_eq(double x, double *par, double a, double b)
  {
    double p1 = par[0];
    double p2 = par[1];

    double y = p1 + (p2 - p1) * (x - a + hh_) / (b - a + 2.0 * hh_);

    return y;
  }

  //---------------- worker functions ---------------- //

  void
  solve_joint::w_work(int dom)
  {

    double a = xmin_ + ((double) dom) * (xmax_ - xmin_) / (double) nworkers();
    double b = xmin_ + ((double) dom + 1.0) * (xmax_ - xmin_)
        / (double) nworkers();

    SSTMAC_DEBUG << "solve_joint[" << dom << "]: working with a=" << a
          << ", b=" << b << "\n";

    if (!w_res_buf_init_)
      {
        w_res_buf_ = (double*) malloc(sizeof(double) * numsamp_ * 2);
        w_res_buf_init_ = true;
      }

    for (int isam = 0; isam < numsamp_; isam++)
      {

        double s1 = solve_eq(a + hh_, &w_args_buf_[isam * 2], a, b);
        double s2 = solve_eq(b - hh_, &w_args_buf_[isam * 2], a, b);
        // SSTMAC_DEBUG << "solve_joint[" << dom << "]: row " << isam
        //       << ", par[0]=" << par[0] << ", par[1]=" << par[1] << ", s1= "
        //       << s1 << ", s2=" << s2 << "\n";
        w_res_buf_[isam] = s1;
        w_res_buf_[isam + numsamp_] = s2;
      }
    comp_lib_->compute_loop(0, numsamp_, 6);

  }

  MPI_Request
  solve_joint::w_irecv_args(int master_rank)
  {
    SSTMAC_DEBUG << "solve_joint -- w_recv_args at worker id " << work_id_
          << "\n";

    if (!w_arg_buf_init_)
      {
        w_args_buf_ = (double*) malloc(sizeof(double) * numsamp_ * 2);
        w_arg_buf_init_ = true;
      }

    MPI_Request req;
    MPI_Irecv(w_args_buf_, numsamp_ * 2, MPI_DOUBLE, master_rank, ARGS_TAG,
        MPI_COMM_WORLD, &req);

    return req;
  }

  void
  solve_joint::w_send_result(int master_rank)
  {
    SSTMAC_DEBUG << "solve_joint -- w_send_result \n";

    MPI_Send(w_res_buf_, numsamp_ * 2, MPI_DOUBLE, master_rank, RESULT_TAG,
        MPI_COMM_WORLD);
  }

  //---------------- master functions ---------------- //

  /**
   * This function should send the correct arguments to the given worker rank
   * @param worker_rank
   */
  void
  solve_joint::m_isend_args(int worker_id, int worker_rank)
  {
    SSTMAC_DEBUG << "solve_joint -- m_isend_args from wid " << worker_id
          << " at rank " << worker_rank << " \n";

    for (int s = 0; s < numsamp_; s++)
      {
        double v0;
        if (worker_id == 0)
          {
            v0 = y0_;
          }
        else
          {
            v0 = boundaries_[worker_id - 1]->data(s);
          }
        m_args_buf_[worker_id][s * 2] = v0;

        double v1;
        if (worker_id == nworkers() - 1)
          {
            v1 = y1_;
          }
        else
          {
            v1 = boundaries_[worker_id]->data(s);
          }

        m_args_buf_[worker_id][s * 2 + 1] = v1;
      }

    MPI_Request dumm;
    MPI_Isend(m_args_buf_[worker_id], numsamp_ * 2, MPI_DOUBLE, worker_rank,
        ARGS_TAG, MPI_COMM_WORLD, &dumm);
  }

  /**
   * This function should call MPI_Irecv, expecting the right amount of data for a result.  In real life,
   * a map of result buffers would be needed, but we should be safe here because the buffer is only filled when
   * MPI_Waitany returns.
   * @param worker_rank
   * @return the MPI_Request returned from MPI_Ircv
   */
  MPI_Request
  solve_joint::m_irecv_result(int worker_id, int worker_rank)
  {
    SSTMAC_DEBUG << "solve_joint -- m_irecv_result from wid " << worker_id
          << " at rank " << worker_rank << " \n";

    MPI_Status stat;
    MPI_Request ret;
    MPI_Irecv(m_res_buf_[worker_id], numsamp_ * 2, MPI_DOUBLE, worker_rank,
        RESULT_TAG, MPI_COMM_WORLD, &ret);

    return ret;
  }

  /**
   * This function is called after a recv has returned a result, and this function should
   * update the results
   * @return the right data structure holding the result
   */
  void
  solve_joint::m_unpack_result(int worker_id, int worker_rank, bool bitflip)
  {
    SSTMAC_DEBUG << "solve_joint -- m_unpack_result from wid " << worker_id
          << " and rank " << worker_rank << " \n";

    if (bitflip)
      {

        int num = rng_->realvalue(true, true) * 2 * (numsamp_ - 1);
        int pos = rng_->realvalue(true, true) * 62;
        uint64_t mask = 1;
        if(pos >= 31){
            pos -= 31;
            mask = mask << 31;
        }

        mask = mask << pos;

        //long *val = (long*) (&m_res_buf_[worker_id][num]);
        SSTMAC_DEBUG << "with pos: " << pos << " and mask " << mask << ", value before: " << m_res_buf_[worker_id][num]  << "\n";
        //m_res_buf_[worker_id][num] = (double)((*val) ^ mask);
        void* v = &m_res_buf_[worker_id][num];
        uint64_t* l = (uint64_t*)v;
        SSTMAC_DEBUG << "bit cast from " << (long)v << " is " << *l << "\n";
        uint64_t newval = (*l) ^ mask;
        SSTMAC_DEBUG << "after mask: " << newval << "\n";
        double* d = (double*)(&newval);
        m_res_buf_[worker_id][num] = *d;
        SSTMAC_DEBUG << "value after: " << m_res_buf_[worker_id][num]  << "\n";
      }

    //update left side
    if (worker_id != 0)
      {
        boundaries_[worker_id - 1]->update(m_res_buf_[worker_id],
            sample_data::RIGHT);

        if (resample_freq_ > 0 && (boundaries_[worker_id - 1]->updates()
            % resample_freq_ == 0))
          {
            boundaries_[worker_id - 1]->resample();
          }
      }

    //update right side
    if (worker_id != nworkers() - 1)
      {
        boundaries_[worker_id]->update(m_res_buf_[worker_id] + numsamp_,
            sample_data::LEFT);

        if (resample_freq_ > 0 && (boundaries_[worker_id]->updates()
            % resample_freq_ == 0))
          {
            boundaries_[worker_id]->resample();
          }
      }

  }

  /**
   * This function should return true if this master's domain has completed (error is within bounds).
   * @return
   */
  bool
  solve_joint::m_done()
  {
    SSTMAC_DEBUG << "solve_joint -- m_done() \n";
    bool done = true;
    for (int r = 0; r < nworkers() - 1; r++)
      {
        if (boundaries_[r]->stddev() > error_)
          {
            done = false;
            break;
          }
      }

    return done;
  }

  double
  solve_joint::m_get_answer_err()
  {
    //first, estimate the error

    double prev = boundaries_[0]->mean();
    double xdel = (xmax_ - xmin_) / nworkers();
    double realslop = (y1_ - y0_) / (xmax_ - xmin_);
    double slope = (boundaries_[0]->mean() - y0_) / xdel;
    double slopeerr = std::pow(slope - realslop, 2.0);
    for (int i = 1; i < nworkers() - 1; i++)
      {
        slope = (boundaries_[i]->mean() - prev) / xdel;

        slopeerr += std::pow(slope - realslop, 2.0);

        prev = boundaries_[i]->mean();
      }
    slope = (y1_ - prev) / xdel;
    slopeerr += std::pow(slope - realslop, 2.0);

    return slopeerr / (nworkers());
  }

  std::string
  solve_joint::m_get_answer_string()
  {
    std::stringstream ss;
    ss << "\n";
    double xdel = (xmax_ - xmin_) / nworkers();
    ss << xmin_ << "\t" << y0_ << "\n";
    double x = xmin_ + xdel;
    for (int i = 0; i < nworkers() - 1; i++)
      {
        ss << x << "\t" << boundaries_[i]->mean() << "\n";
        x += xdel;
      }
    ss << xmax_ << "\t" << y1_ << "\n";
    return ss.str();
  }

  bool
  solve_joint::m_collaborate()
  {
    return false;
  }

  void
  solve_joint::m_generate_initial_data(int rank, int masters)
  {
    int domsize = nworkers();

    m_args_buf_ = (double**) malloc(sizeof(double*) * domsize);
    m_res_buf_ = (double**) malloc(sizeof(double*) * domsize);

    for (int r = 0; r < domsize - 1; r++)
      {
        sample_data::ptr d = sample_data::construct(numsamp_, ymin_, ymax_,
            rng_, outlier_);
        boundaries_[r] = d;

        m_args_buf_[r] = (double*) malloc(sizeof(double) * numsamp_ * 2);
        m_res_buf_[r] = (double*) malloc(sizeof(double) * numsamp_ * 2);
      }

    m_args_buf_[domsize - 1] = (double*) malloc(sizeof(double) * numsamp_ * 2);
    m_res_buf_[domsize - 1] = (double*) malloc(sizeof(double) * numsamp_ * 2);

  }

}
