/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SOLVE_JOINT_H_
#define SOLVE_JOINT_H_

#include "ftpm.h"
#include "ftpm_data.h"
#include <fstream>
#include <sstream>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/rng.h>
#include <sstmac/software/libraries/compute/lib_compute_loops.h>
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>

namespace ftpm
{

  class solve_joint : public ftpm
  {
  protected:
    class sample_data;

  protected:
    solve_joint(sstmac::sw::softwareid id) :
      ftpm(id)
    {
      w_arg_buf_init_ = false;
      w_res_buf_init_ = false;

      comp_lib_ = sstmac::sw::lib_compute_loops::construct(id);
      register_lib(comp_lib_);
    }

    //------ parameters ------- //
    double xmin_;
    double xmax_;
    double ymax_;
    double ymin_;
    double y0_;
    double y1_;
    double hh_;
    double error_; //num of standard devations to solve within
    double outlier_; //num of standard deviations

    int numsamp_;

    //------ worker state ------- //
    double* w_args_buf_;
    bool w_arg_buf_init_;

    double*w_res_buf_;
    bool w_res_buf_init_;

    sstmac::sw::lib_compute_loops::ptr comp_lib_;

    // -------- master state ------- //

    std::map<int, boost::intrusive_ptr<sample_data> > boundaries_;

    double** m_res_buf_;
    double** m_args_buf_;

    int resample_freq_;
    //---------------------------- //
  public:
    typedef boost::intrusive_ptr<solve_joint> ptr;

    virtual
    ~solve_joint()
    {
    }

    virtual void
    consume_params(const sstmac::sim_parameters::ptr &params);

    static ptr
    construct(sstmac::sw::softwareid id)
    {
      return ptr(new solve_joint(id));
    }

    virtual std::string
    to_string() const
    {
      return "solve joint";
    }

    app::ptr
    clone(sstmac::sw::softwareid id)
    {
      return ptr(new solve_joint(id));
    }

  protected:

    double
    solve_eq(double x, double *par, double a, double b);

    //---------------- worker functions ---------------- //

    void
    w_work(int dom);

    MPI_Request
    w_irecv_args(int master_rank);

    void
    w_send_result(int master_rank);

    //---------------- master functions ---------------- //
    void
    m_isend_args(int worker_id, int worker_rank);

    MPI_Request
    m_irecv_result(int worker_id, int worker_rank);

    void
    m_unpack_result(int worker_id, int worker_rank, bool bitflip = false);

    bool
    m_done();

    double
    m_get_answer_err();

    std::string
    m_get_answer_string();

    bool
    m_collaborate();

    void
    m_generate_initial_data(int wid, int rank);

    //--------------------------------------------------- //
  protected:
    class sample_data : public ftpm_data
    {
    public:
      enum DIR
      {
        LEFT, RIGHT
      };
    protected:
      sample_data(int num, double lower, double upper,
          RNG::UniformInteger::ptr rng, double outlier) :
        num_(num), outlier_(outlier)
      {

        vals_ = (double*) malloc(sizeof(double) * num_);

        for (int s = 0; s < num_; s++)
          {
            double samp = rng->realvalue(true, true) * (upper - lower) + lower;
            vals_[s] = samp;

          }

        myrng_ = boost::mt19937(time(NULL));

        mean_valid_ = false;
        stddev_valid_ = false;

        right_pending_ = false;
        left_pending_ = false;

        rightbuf_ = (double*) malloc(sizeof(double) * num_);
        leftbuf_ = (double*) malloc(sizeof(double) * num_);

        updates_ = 0;
      }

      double* vals_;
      int num_;

      boost::mt19937 myrng_;

      double mean_;
      double stddev_;

      bool mean_valid_;
      bool stddev_valid_;

      double* rightbuf_;
      double* leftbuf_;

      bool right_pending_;
      bool left_pending_;

      double outlier_;

      int updates_;

    public:
      typedef boost::intrusive_ptr<sample_data> ptr;

      static ptr
      construct(int num, double lower, double upper,
          RNG::UniformInteger::ptr rng, double outlier)
      {
        return ptr(new sample_data(num, lower, upper, rng, outlier));
      }

      virtual std::string
      to_string() const
      {
        return "sample data";
      }

      virtual
      ~sample_data()
      {
        free(rightbuf_);
        free(leftbuf_);
        free(vals_);
      }

      long
      count() const
      {
        return num_;
      }

      long
      size() const
      {
        return sizeof(double) * count();
      }

      int
      updates() const
      {
        return updates_;
      }

      void
      resample()
      {


        boost::normal_distribution<> nd(mean(), stddev());

        boost::variate_generator<boost::mt19937&, boost::normal_distribution<> >
            norm(myrng_, nd);

        for (int i = 0; i < num_; i++)
          {
            vals_[i] = norm();
          }

        mean_valid_ = false;
        stddev_valid_ = false;

        left_pending_ = false;
        right_pending_ = false;
      }

      double
      data(int s)
      {
        if (s < 0 || s >= num_)
          {
            sst_throw(ssterror, "solve_joint::index %d was out of range (0, %d)", s, num_);
          }
        return vals_[s];
      }

      std::string
      print() const
      {
        std::stringstream ss;

        for (int i = 0; i < num_; i++)
          {
            ss << vals_[i] << " ";
          }
        ss << "\n";

        return ss.str();
      }

      void
      update(double *dat, DIR d)
      {
        if (d == LEFT)
          {
            for (int i = 0; i < num_; i++)
              {
                leftbuf_[i] = dat[i];
              }
            left_pending_ = true;
          }
        else if (d == RIGHT)
          {
            for (int i = 0; i < num_; i++)
              {
                rightbuf_[i] = dat[i];
              }
            right_pending_ = true;
          }

        if (right_pending_ && left_pending_)
          {
            double m = mean();
            double s = stddev();
            //take average of left and right
            int discarded = 0;
            for (int i = 0; i < num_; i++)
              {
                double a = (rightbuf_[i] + leftbuf_[i]) / 2.0;
                if (outlier_ <= 0 || std::abs(m - a) < outlier_
                    * s)
                  {
                    vals_[i] = a;
                  }else if(outlier_ > 0){
                      discarded++;
                  }
              }

            if(discarded >= num_ / 2.0){
                outlier_ *= 2;
            }

            right_pending_ = false;
            left_pending_ = false;

            mean_valid_ = false;
            stddev_valid_ = false;
          }

        updates_++;

      }

      double
      mean()
      {
        if (!mean_valid_)
          {
            double m = 0;
            for (int i = 0; i < num_; i++)
              {
                m += vals_[i];
              }
            mean_ = m / num_;
            mean_valid_ = true;
          }

        return mean_;
      }

      double
      stddev()
      {
        if (!stddev_valid_)
          {
            double dev = 0;
            double m = mean();
            for (int i = 0; i < num_; i++)
              {
                dev += std::pow((vals_[i] - m), 2.0);
              }
            stddev_ = std::sqrt(dev / num_);
            stddev_valid_ = true;
          }

        return stddev_;

      }
    };
  };

}

#endif /* SOLVE_JOINT_H_ */
