/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_PAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_PAYLOAD_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <typeinfo>

namespace sstmac
{

  class sst_serializer;

  /**
   * Pure abstract base for payload that can be associated with a network send.
   */
  class payload : public ptr_type
  {
  protected:
    static debug dbg_;

    /**
     * Empty protected constructor to enforce using construct()
     */
    payload()
    {
      if(!dbg_.is_real()){
          dbg_ = debug("<debug> payload");
          logger::register_logger(&dbg_);
      }
    }
  public:

    typedef boost::intrusive_ptr<payload> ptr;
    typedef boost::intrusive_ptr<const payload> const_ptr;

    /// Goodbye.
    virtual
    ~payload()
    {
    }

    /**
     * Create a copy of this payload
     * @return A smart pointer to a new payload object
     */
    virtual payload::const_ptr
    clone() const = 0;

    /**
     * Stringify
     * @return String description
     */
    virtual std::string
    to_string() const
    {
      return "payload";
    }

    /**
     *  Convenience function to get a null payload pointer.
     *  @return null smart pointer
     */
    static ptr
    null()
    {
      return ptr();
    }

    virtual long
    get_byte_length() const = 0;

    /**
     * Pure virtual add operator
     * @param other Payload object to add to this one
     * @return A smart pointer to a new object which is the sum of this and other
     */
    virtual const payload::const_ptr
    add(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual less than (<) operator
     * @param other Payload to compare to
     * @return true if this payload is considered less than other
     */
    virtual const payload::const_ptr
    less_than(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual greater than (>) operator
     * @param other Payload to compare to
     * @return true if this payload is considered greather than other
     */
    virtual const payload::const_ptr
    greater_than(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual equals operator
     * @param other Payload to compare to
     * @return true if this payload is considered equal to the other
     */
    virtual bool
    equals(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual logical or operator
     * @param other Payload to compare to
     * @return true if this either payload is considered true
     */
    virtual const payload::const_ptr
    logical_or(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual logical xor operator
     * @param other Payload to compare to
     * @return true if exactly one payload is considered true
     */
    virtual const payload::const_ptr
    logical_xor(const payload::const_ptr &other) const = 0;

    /**
     * Pure virtual logical and operator
     * @param other Payload to compare to
     * @return true if both payloads are considered true
     */
    virtual const payload::const_ptr
    logical_and(const payload::const_ptr &other) const = 0;

    /**
     * Serialize this paylod to an sst_serializer, basically
     * just packs the deserialize name (get_deser_name()).
     * Children of this class should override this function,
     * and call this one as well
     * @param ser the sst_serializer to use
     */
    virtual void
    serialize(sst_serializer* ser) const;

    /**
     * Dummy deserialize function in case there
     * was no payload given (returns a null pointer)
     * @param ser serializer to use
     * @return a null smart pointer
     */
    static const_ptr
    deserialize(sst_serializer* ser);

    /**
     * Returns the std::string that a deserializer can use to
     * figure out what kind of object it should make.
     * @return std::string describing that a null payload should be created
     */
    static std::string
    get_deser_name()
    {
      return "payload";
    }
  };
} // end of namespace sstmac
#endif
